/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.main;

import jp.pegasus.config.Config;
import jp.pegasus.log.Logger;
import jp.pegasus.pivot.factory.SSFrameFactory;
import jp.pegasus.pivot.factory.SSMenuBarFactory;
import jp.pegasus.pivot.factory.SSTreeNodeFactory;
import jp.pegasus.pivot.factory.tabpane.SSTabPaneMenuFactory;
import jp.pegasus.pivot.factory.tabpane.SSTabPanePropertyFactory;
import jp.pegasus.pivot.model.SSData;
import jp.pegasus.pivot.skin.*;
import jp.pegasus.pivot.ss.*;
import jp.pegasus.pivot.ss.tablepane.SSTablePane;
import jp.pegasus.pivot.ss.tableview.SSTableView;
import jp.pegasus.pivot.ss.tableview.SSTableViewHeader;
import jp.pegasus.util.JsonUtil;
import org.apache.pivot.collections.ArrayList;
import org.apache.pivot.collections.Map;
import org.apache.pivot.wtk.*;

/**
 * MainClass for creating the client UI
 */
public class Main implements Application {
    /**
     * Debug utility fields
     * for checking how long it takes to finish starting up.
     */
    private static long startTimeMillis;

    private SSFrame frame = null;
    private MenuBar menuBar = null;
    private SSTabPane leftTab = null;
    private SSTabPane centerTab = null;
    private SSTabPane rightTab = null;
    private SplitPane splitMain = null;
    private SplitPane splitSub = null;


    public Main() {
        super();
        Theme.getTheme().set(SSCardPane.class, SSCardPaneSkin.class);
        Theme.getTheme().set(SSCorner.class, SSCornerSkin.class);
        Theme.getTheme().set(SSFlowPane.class, SSFlowPaneSkin.class);
        Theme.getTheme().set(SSFrame.class, SSFrameSkin.class);
        Theme.getTheme().set(SSFrameSkin.SSFrameButton.class, SSFrameSkin.SSFrameButtonSkin.class);
        Theme.getTheme().set(SSLabel.class, SSLabelSkin.class);
        Theme.getTheme().set(SSListView.class, SSListViewSkin.class);
        Theme.getTheme().set(SSPalette.class, SSPaletteSkin.class);
        Theme.getTheme().set(SSPushButton.class, SSPushButtonSkin.class);
        Theme.getTheme().set(SSScrollBar.class, SSScrollBarSkin.class);
        Theme.getTheme().set(SSScrollPane.class, ScrollPaneSkin.class);
        Theme.getTheme().set(SSTablePane.class, SSTablePaneSkin.class);
        Theme.getTheme().set(SSTableView.class, SSTableViewSkin.class);
        Theme.getTheme().set(SSTableViewHeader.class, TableViewHeaderSkin.class);
        Theme.getTheme().set(SSTabPane.class, SSTabPaneSkin.class);
        Theme.getTheme().set(SSTextArea.class, SSTextAreaSkin.class);
        Theme.getTheme().set(SSTextInput.class, SSTextInputSkin.class);
        Theme.getTheme().set(SSTreeView.class, SSTreeViewSkin.class);
    }

    /**
     * Starting up.
     * <p/>
     * Fyi.
     * Abstract image of the Components hierarchy:
     * <p/>
     * Frame
     * |
     * SplitPane
     * |
     * |(Menu)            (Content)          (Property)
     * .------------------.------------------.
     * |                  |                  |
     * |1..*              |0..*              |3
     * TabPane            TabPane            TabPane
     * TabPane.TabButton  TabPane.TabButton  TabPane.TabButton
     * |                  |                  |
     * Border             Border             Border
     * |                  |                  |
     * ScrollPane         ScrollPane         ScrollPane
     * |                  |                  |
     * TreeView           Corner             TableHeaderView
     * |                  TableHeaderView    TableView
     * |                  TableView
     * |0..*
     * TreeBranch
     * TreeNode
     */
    @Override
    @SuppressWarnings("unchecked")
    public void startup(final Display display, Map<String, String> properties) throws Exception {
        frame = SSFrameFactory.createFrame();
        frame.setContent(new SplitPane(Orientation.HORIZONTAL, new SplitPane(), new SplitPane()));
        frame.open(display);


//        ArrayList<SSTreeNode> treeNodeList = SSTreeNodeFactory.constructTreeNodeList(data);
//        final int selectFileIndex = data.getSelectFileIndex();
//
//        /**
//         * Left Tab(Menu)
//         */
//        SSTabPane leftTab = SSTabPaneMenuFactory.constructTabPaneMenu(treeNodeList, selectFileIndex);
//
//        /**
//         * Center Tab(Content)
//         */
//        SSTabPane centerTab = null;
//        if (treeNodeList.getLength() != 0) {
//            SSTreeNode selectTreeNode = treeNodeList.get(selectFileIndex);
//            centerTab = selectTreeNode.getTabPane();
//        } else {
//            centerTab = new SSTabPane();
//        }
//

        /**
         * Menu Bar
         */
        ApplicationContext.queueCallback(new Runnable() {
            @Override
            public void run() {
                menuBar = SSMenuBarFactory.createMenuBar();
                frame.setMenuBar(menuBar);
            }
        });

        /**
         * Right Tab
         */
        ApplicationContext.queueCallback(new Runnable() {
            @Override
            public void run() {
                rightTab = SSTabPanePropertyFactory.constructTabPaneProperty();
                splitMain = (SplitPane) frame.getContent();
                splitMain.setOrientation(Orientation.HORIZONTAL);
                splitMain.setSplitRatio(Config.DEF_SPLIT_RATIO_RIGT);
                splitMain.setRight(rightTab);
            }
        });

        /**
         * Left Tab
         */
        ApplicationContext.queueCallback(new Runnable() {
            @Override
            public void run() {
                SSData data = (SSData) frame.getAttribute(SSData.DEFAULT_TYPE);
                if (data == null) {
                    data = JsonUtil.readGrandData(Config.APP_DATA_FILE_PATH);
                    frame.setAttribute(SSData.DEFAULT_TYPE, data);
                }
                ArrayList<SSTreeNode> treeNodeList = SSTreeNodeFactory.constructTreeNodeList(data);
                final int selectFileIndex = data.getSelectFileIndex();

                leftTab = SSTabPaneMenuFactory.constructTabPaneMenu(treeNodeList, selectFileIndex);

                splitMain = (SplitPane) frame.getContent();
                splitSub = (SplitPane) splitMain.getLeft();
                splitSub.setOrientation(Orientation.HORIZONTAL);
                splitSub.setSplitRatio(Config.DEF_SPLIT_RATIO_LEFT);
                splitSub.setLeft(leftTab);
            }
        });

        /**
         * Center Tab
         */
        ApplicationContext.queueCallback(new Runnable() {
            @Override
            public void run() {
                SSData data = (SSData) frame.getAttribute(SSData.DEFAULT_TYPE);
                if (data == null) {
                    data = JsonUtil.readGrandData(Config.APP_DATA_FILE_PATH);
                    frame.setAttribute(SSData.DEFAULT_TYPE, data);
                }
                ArrayList<SSTreeNode> treeNodeList = SSTreeNodeFactory.constructTreeNodeList(data);
                final int selectFileIndex = data.getSelectFileIndex();

                leftTab = SSTabPaneMenuFactory.constructTabPaneMenu(treeNodeList, selectFileIndex);

                /**
                 * Center Tab(Content)
                 */
                if (treeNodeList.getLength() != 0) {
                    SSTreeNode selectTreeNode = treeNodeList.get(selectFileIndex);
                    centerTab = selectTreeNode.getTabPane();
                } else {
                    centerTab = new SSTabPane();
                }

                splitMain = (SplitPane) frame.getContent();
                splitSub = (SplitPane) splitMain.getLeft();
                splitSub.setOrientation(Orientation.HORIZONTAL);
                splitSub.setSplitRatio(Config.DEF_SPLIT_RATIO_LEFT);
                splitSub.setRight(centerTab);
            }
        });


    }

    @Override
    public boolean shutdown(boolean optional) {
        if (frame != null) {
            frame.close();
        }
        return false;
    }

    @Override
    public void suspend() {
    }

    @Override
    public void resume() {
    }

    public static void main(String[] args) {
        startTimeMillis = System.currentTimeMillis();
        DesktopApplicationContext.main(Main.class, args);

        if (Logger.log.isDebugEnabled()) {
            final long executionTime = System.currentTimeMillis() - startTimeMillis;
            Logger.log.debug("Execution time: {} seconds", executionTime / 1000.0);
        }
    }
}
