/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.content;

import jp.pegasus.pivot.factory.SSLabelFactory;
import jp.pegasus.pivot.ss.SSLabel;
import jp.pegasus.pivot.ss.SSTreeBranch;
import jp.pegasus.pivot.ss.SSTreeNode;
import jp.pegasus.pivot.ss.SSTreeView;
import jp.pegasus.pivot.ss.tablepane.SSTablePane;
import jp.pegasus.util.ColorUtil;
import org.apache.pivot.wtk.*;
import org.apache.pivot.wtk.Insets;
import org.apache.pivot.wtk.media.Image;

import java.awt.*;

public class SSTreeViewNodeRenderer extends SSTablePane implements SSTreeView.NodeRenderer {
    protected ImageView imageView = new ImageView();
    protected SSLabel textLabel = SSLabelFactory.createLabel();
    protected SSLabel keyboardShortcutLabel = SSLabelFactory.createLabel();
    ;

    public SSTreeViewNodeRenderer() {
        getStyles().put("padding", new Insets(2));

        getColumns().add(new SSTablePane.Column(1, true));
        getColumns().add(new SSTablePane.Column());

        BoxPane boxPane = new BoxPane();
        boxPane.add(imageView);
        boxPane.add(textLabel);
        boxPane.getStyles().put("verticalAlignment", VerticalAlignment.CENTER);
        boxPane.getStyles().put("padding", new Insets(0, 0, 0, 6));

        SSTablePane.Row row = new SSTablePane.Row();
        row.add(boxPane);
        row.add(keyboardShortcutLabel);
        getRows().add(row);

        imageView.getStyles().put("backgroundColor", null);
    }

    @Override
    public void setSize(int width, int height) {
        super.setSize(width, height);

        // Since this component doesn't have a parent, it won't be validated
        // via layout; ensure that it is valid here
        validate();
    }

    @Override
    public void render(Object node, Tree.Path path, int rowIndex, SSTreeView treeView, boolean expanded,
                       boolean selected, SSTreeView.NodeCheckState checkState, boolean highlighted, boolean disabled) {
        if (node != null) {
            Image icon = null;
            String text = null;
            String keyboardShortCutText = null;
            boolean showShortCut = false;

            if (node instanceof SSTreeNode) {
                SSTreeNode treeNode = (SSTreeNode) node;

                if (expanded && treeNode instanceof SSTreeBranch) {
                    SSTreeBranch treeBranch = (SSTreeBranch) treeNode;
                    icon = treeBranch.getExpandedIcon();
                    if (icon == null) {
                        icon = treeBranch.getIcon();
                    }
                } else {
                    icon = treeNode.getIcon();
                }

                text = treeNode.getText();
                keyboardShortCutText = treeNode.getShortCutText();
//                Logger.log.debug("keyboardShortCutText:{}", keyboardShortCutText);
            } else if (node instanceof Image) {
                icon = (Image) node;
            } else {
                text = node.toString();
            }

            // Update the image view
            imageView.setImage(icon);
            imageView.getStyles().put("opacity",
            (treeView.isEnabled() && !disabled) ? 1.0f : 0.3f);

            // Update the label
            textLabel.setText(text);

            if (text == null) {
                textLabel.setVisible(false);
            } else {
                textLabel.setVisible(true);

                Font font = (Font) treeView.getStyles().get("font");
                textLabel.getStyles().put("font", font);

                Color color;
                if (treeView.isEnabled() && !disabled) {
                    if (selected) {
                        if (treeView.isFocused()) {
                            color = (Color) treeView.getStyles().get("selectColor");
                        } else {
                            color = (Color) treeView.getStyles().get("inactiveSelectColor");
                        }
                    } else if (highlighted) {
                        color = (Color) treeView.getStyles().get("highlightColor");
                    } else {
                        color = (Color) treeView.getStyles().get("color");
                    }
                } else {
                    color = (Color) treeView.getStyles().get("disabledColor");
                }

                textLabel.getStyles().put("color", color);
                if (keyboardShortCutText != null) {
                    keyboardShortcutLabel.setVisible(true);
                    keyboardShortcutLabel.setText(keyboardShortCutText);
                    Font shortCutFont = (Font) treeView.getStyles().get("shortCutFont");
                    keyboardShortcutLabel.getStyles().put("font", shortCutFont.deriveFont(Font.ITALIC));
                    keyboardShortcutLabel.getStyles().put("color", ColorUtil.blend(color, Color.BLUE));
                    keyboardShortcutLabel.getStyles().put("verticalAlignment", VerticalAlignment.CENTER);
                } else {
                    keyboardShortcutLabel.setVisible(false);
                }
            }
        }
    }

    public String toString(Object node) {
        String string = null;

        if (node instanceof SSTreeNode) {
            SSTreeNode treeNode = (SSTreeNode) node;
            if (treeNode.getShortCutText() != null) {
                string = String.format("%s %s", treeNode.getText(), treeNode.getShortCutText());
            } else {
                string = String.format("%s", treeNode.getText());
            }
        } else {
            if (node != null) {
                string = node.toString();
            }
        }

        return string;
    }

    public int getIconWidth() {
        return imageView.getPreferredWidth(-1);
    }

    public void setIconWidth(int iconWidth) {
        if (iconWidth == -1) {
            throw new IllegalArgumentException();
        }

        imageView.setPreferredWidth(iconWidth);
    }

    public int getIconHeight() {
        return imageView.getPreferredHeight(-1);
    }

    public void setIconHeight(int iconHeight) {
        if (iconHeight == -1) {
            throw new IllegalArgumentException();
        }

        imageView.setPreferredHeight(iconHeight);
    }

    public boolean getShowIcon() {
        return imageView.isVisible();
    }

    public void setShowIcon(boolean showIcon) {
        imageView.setVisible(showIcon);
    }

    public boolean getFillIcon() {
        return (Boolean) imageView.getStyles().get("fill");
    }

    public void setFillIcon(boolean fillIcon) {
        imageView.getStyles().put("fill", fillIcon);
    }

    /**
     * Gets the bounds of the text that is rendered by this renderer.
     *
     * @return The bounds of the rendered text,
     *         or <tt>null</tt> if this renderer did not render any text.
     */
    public Bounds getTextBounds() {
        return (textLabel.isVisible() ? (
        keyboardShortcutLabel.isVisible() ?
        textLabel.getBounds().union(keyboardShortcutLabel.getBounds()) : textLabel.getBounds()) : null);
    }
}
