/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.editor;

import jp.pegasus.config.Config;
import jp.pegasus.log.Logger;
import jp.pegasus.pivot.content.SSButtonDataRenderer;
import jp.pegasus.pivot.factory.SSTextInputFactory;
import jp.pegasus.pivot.model.SSData;
import jp.pegasus.pivot.model.SSFileData;
import jp.pegasus.pivot.model.SSGridData;
import jp.pegasus.pivot.model.SSTextData;
import jp.pegasus.pivot.skin.SSTabPaneSkin;
import jp.pegasus.pivot.ss.SSTabPane;
import jp.pegasus.pivot.ss.SSTextInput;
import jp.pegasus.util.JsonUtil;
import org.apache.pivot.collections.HashMap;
import org.apache.pivot.wtk.*;
import org.apache.pivot.wtk.content.ButtonData;

public class SSTabPaneEditor extends Window implements SSTabPane.TabEditor {
    private SSTabPane tabPane = null;
    private int index = 0;

    private SSTextInput textInput = SSTextInputFactory.createTextInput();

    private ContainerMouseListener displayMouseHandler = new ContainerMouseListener.Adapter() {
        @Override
        public boolean mouseDown(Container container, Mouse.Button button, int x, int y) {
            Display display = (Display) container;
            Window window = (Window) display.getComponentAt(x, y);

            if (window != SSTabPaneEditor.this) {
                endEdit(true);
            }

            return false;
        }

        @Override
        public boolean mouseWheel(Container container, Mouse.ScrollType scrollType,
                                  int scrollAmount, int wheelRotation, int x, int y) {
            Display display = (Display) container;
            Window window = (Window) display.getComponentAt(x, y);

            return (window != SSTabPaneEditor.this);
        }
    };

    public SSTabPaneEditor() {
        setContent(textInput);
    }

    public SSTabPane getSSTabPane() {
        return tabPane;
    }

    public int getIndex() {
        return index;
    }

    public SSTextInput getTextInput() {
        return textInput;
    }

    @Override
    public void beginEdit(SSTabPane tabPane) {
        this.tabPane = tabPane;
        this.index = tabPane.getSelectedIndex();

        // Get the data being edited
        Component tab = tabPane.getSelectedTab();
        ButtonData tabData = (ButtonData) SSTabPane.getTabData(tab);

        textInput.setText(tabData.getText());
        textInput.selectAll();

        // Render the tab data
        SSTabPaneSkin.TabButton tabButton = new SSTabPaneSkin().new TabButton(tab);
        SSButtonDataRenderer tabRenderer = (SSButtonDataRenderer) tabPane.getTabDataRenderer();
        tabRenderer.render(tabButton.getButtonData(), tabButton, false);

        final Point selectedTabButtonLocation = tabPane.getSelectedTabButtonLocation();
        Dimensions dimensions = tabRenderer.getPreferredSize();

        // Get the text bounds
        Bounds textBounds = tabRenderer.getTextBounds();
        Logger.log.info("textBounds:{}", textBounds.toString());

        // Calculate the bounds of what is being edited
        Insets padding = (Insets) textInput.getStyles().get("padding");
        Bounds editBounds = new Bounds(
        selectedTabButtonLocation.x + textBounds.x + padding.left,
        selectedTabButtonLocation.y + padding.top,
        dimensions.width - textBounds.x + padding.right,
        dimensions.height + padding.bottom);
        Logger.log.info("editBounds:{}", editBounds.toString());

        // Scroll to make the tab as visible as possible
        tabPane.scrollAreaToVisible(editBounds.x, editBounds.y, padding.left + 1, editBounds.height);

        // Constrain the bounds by what is visible through viewport ancestors
        editBounds = tabPane.getVisibleArea(editBounds);
        Point location = tabPane.mapPointToAncestor(tabPane.getDisplay(), editBounds.x, editBounds.y);

        textInput.setPreferredWidth(editBounds.width);
        setLocation(location.x, location.y + (editBounds.height - getPreferredHeight(-1)) / 2);

        // Open the editor
        open(tabPane.getWindow());
    }

    @Override
    @SuppressWarnings("unchecked")
    public void endEdit(boolean result) {
        if (result) {
            // Update the tab data
            final String text = textInput.getText();
            Logger.log.info("text: {}", text);

            ButtonData buttonData = new ButtonData(text);
            Component selectedTab = tabPane.getSelectedTab();
            SSTabPane.setTabData(selectedTab, buttonData);

            Window window = tabPane.getWindow();
            if (window != null) {
                SSData data = (SSData) window.getAttribute(SSData.DEFAULT_TYPE);
                final int selectFileIndex = data.getSelectFileIndex();
                SSFileData selectFileData = data.getFileDataList().get(selectFileIndex);
                Object sheetData = selectFileData.getSheetDataList().get(index);

                if (sheetData instanceof SSGridData) {
                    ((SSGridData) sheetData).setSheetName(text);
                } else if (sheetData instanceof SSTextData) {
                    ((SSTextData) sheetData).setSheetName(text);
                } else if (sheetData instanceof HashMap) {
                    ((HashMap) sheetData).put("sheetName", text);
                } else {
                    throw new RuntimeException();
                }

                JsonUtil.writeGrandData(data, Config.APP_DATA_FILE_PATH);
            }

            // Get a reference to node's parent
            // Re-select the tab, and make sure it's visible
//            tabPane.setSelectedIndex(index);
            tabPane.scrollAreaToVisible(tabPane.getBounds());
        }

        getOwner().moveToFront();
        tabPane.requestFocus();

        close();
    }

    @Override
    public boolean isEditing() {
        return (tabPane != null);
    }

    @Override
    public void open(Display display, Window owner) {
        if (tabPane == null) {
            throw new IllegalStateException();
        }

        super.open(display, owner);
        display.getContainerMouseListeners().add(displayMouseHandler);

        requestFocus();
    }

    @Override
    public void close() {
        Display display = getDisplay();
        display.getContainerMouseListeners().remove(displayMouseHandler);

        super.close();
    }

    @Override
    public boolean keyPressed(int keyCode, Keyboard.KeyLocation keyLocation) {
        boolean consumed;

        if (keyCode == Keyboard.KeyCode.ENTER) {
            endEdit(true);
            consumed = true;
        } else if (keyCode == Keyboard.KeyCode.ESCAPE) {
            endEdit(false);
            consumed = true;
        } else {
            consumed = false;
        }

        return consumed;
    }
}
