/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.editor;

import jp.pegasus.log.Logger;
import jp.pegasus.pivot.content.SSTableViewTextAreaCellRenderer;
import jp.pegasus.pivot.factory.SSCardPaneFactory;
import jp.pegasus.pivot.factory.SSTablePaneFactory;
import jp.pegasus.pivot.factory.SSTextAreaFactory;
import jp.pegasus.pivot.factory.ScrollPaneFactory;
import jp.pegasus.pivot.listener.SSCardPaneListener;
import jp.pegasus.pivot.ss.SSCardPane;
import jp.pegasus.pivot.ss.SSScrollPane;
import jp.pegasus.pivot.ss.SSTextArea;
import jp.pegasus.pivot.ss.tablepane.SSTablePane;
import jp.pegasus.pivot.ss.tableview.SSTableView;
import org.apache.pivot.collections.Dictionary;
import org.apache.pivot.collections.HashMap;
import org.apache.pivot.wtk.*;
import org.apache.pivot.wtk.Component;
import org.apache.pivot.wtk.Container;
import org.apache.pivot.wtk.Window;
import org.apache.pivot.wtk.media.Image;
import org.apache.pivot.wtk.skin.CardPaneSkin;

import java.awt.*;

public class SSTableViewRowEditor extends Window implements SSTableView.RowEditor {
    /**
     * Default table view row editor.
     */
    private class RowImage extends Image {
        private Bounds bounds = new Bounds(0, 0, 0, 0);

        @Override
        public int getWidth() {
            return bounds.width;
        }

        @Override
        public int getHeight() {
            return bounds.height;
        }

        @Override
        public void paint(Graphics2D graphics) {
            graphics.setPaint(Color.WHITE);
            graphics.setColor(Color.WHITE);
            graphics.translate(-bounds.x, -bounds.y);
            graphics.clipRect(bounds.x, bounds.y, bounds.width, bounds.height);

            tableView.paint(graphics);
        }
    }

    private SSTableView tableView = null;

    /**
     * int editColIndex = 0x0000FFFF & (editIndexPair >> 16);
     * int editRowIndex = 0x0000FFFF & (editIndexPair);
     */
    private int editIndexPair = -1;

    private SSScrollPane tableViewScrollPane = null;

    private SSScrollPane scrollPane =
    ScrollPaneFactory.createScrollPane(
    SSScrollPane.ScrollBarPolicy.NEVER,
    SSScrollPane.ScrollBarPolicy.FILL
    );

    private SSCardPane cardPane = SSCardPaneFactory.createCardPane();
    private SSTablePane tablePane = SSTablePaneFactory.createTablePane();
    private SSTablePane.Row editorRow = new SSTablePane.Row();

    private RowImage rowImage = new RowImage();
    private HashMap<String, org.apache.pivot.wtk.Component> cellEditors = new HashMap<String, org.apache.pivot.wtk.Component>();

    private ContainerMouseListener displayMouseHandler = new ContainerMouseListener.Adapter() {
        @Override
        public boolean mouseDown(org.apache.pivot.wtk.Container container, Mouse.Button button, int x, int y) {
            Display display = (Display) container;
            Window window = (Window) display.getComponentAt(x, y);

            boolean consumed;
            if (window != SSTableViewRowEditor.this && (window == null || !isOwner(window))) {
                endEdit(true);
                consumed = true;
            } else {
                consumed = false;
            }

            return consumed;
        }

        @Override
        public boolean mouseWheel(org.apache.pivot.wtk.Container container, Mouse.ScrollType scrollType,
                                  int scrollAmount, int wheelRotation, int x, int y) {
            Display display = (Display) container;
            Window window = (Window) display.getComponentAt(x, y);

            return (window != SSTableViewRowEditor.this);
        }
    };

    private static final int IMAGE_CARD_INDEX = 0;
    private static final int EDITOR_CARD_INDEX = 1;

    public SSTableViewRowEditor() {
        setContent(scrollPane);
        scrollPane.setView(cardPane);
        scrollPane.getViewportListeners().add(new ViewportListener.Adapter() {
            @Override
            public void scrollLeftChanged(Viewport viewport, int previousScrollLeft) {
                if (tableViewScrollPane != null) {
                    //log.debug("viewport.getScrollLeft: {}", viewport.getScrollLeft());
//                    tableViewScrollPane.setScrollLeft(viewport.getScrollLeft());
                    viewport.setScrollLeft(tableViewScrollPane.getScrollLeft());
                }
            }
        });

        cardPane.add(new ImageView(rowImage));
        cardPane.add(tablePane);
        cardPane.getCardPaneListeners().add(new SSCardPaneListener.Adapter() {
            @Override
            public void selectIndexChanged(SSCardPane cardPane, int previousSelectIndex) {
                if (previousSelectIndex == IMAGE_CARD_INDEX) {
                    final int editColIndex = 0x0000FFFF & (editIndexPair >> 16);
                    editorRow.get(editColIndex).requestFocus();
                } else {
                    endEdit(false);
                }
            }
        });

        tablePane.getStyles().put("horizontalSpacing", 1);
        tablePane.getStyles().put("backgroundColor", "#dddcd5");
        tablePane.getRows().add(editorRow);
    }

    public SSTableView getTableView() {
        return tableView;
    }


    public int getEditIndexPair() {
        return editIndexPair;
    }

    public void setEditIndexPair(int editIndexPair) {
        this.editIndexPair = editIndexPair;
    }

    /**
     * Gets this row editor's cell editor dictionary. The caller may specify
     * explicit editor components and place them in this dictionary by their
     * table view column names. Any column that does not have an entry in this
     * dictionary will have a {@link TextInput} implicitly associated with it
     * during editing.
     * <p/>
     * This row editor uses data binding to populate the cell editor components
     * and to get the data back out of those components, so it is the caller's
     * responsibility to set up the data binding keys in each component they
     * specify in this dictionary. The data binding key should equal the column
     * name that the cell editor serves.
     *
     * @return The cell editor dictionary.
     */
    public Dictionary<String, org.apache.pivot.wtk.Component> getCellEditors() {
        return cellEditors;
    }

    public void setCellEditors(HashMap<String, Component> cellEditors) {
        this.cellEditors = cellEditors;
    }

    public void beginEdit(SSTableView tableView, final int columnIndex, final int rowIndex) {
        beginEdit(tableView, columnIndex, rowIndex, null);
    }

    public void beginEdit(SSTableView tableView, final int colIdx, final int rowIdx, Character charactor) {
        this.tableView = tableView;
        editIndexPair = (colIdx << 16) | rowIdx;

        Container tableViewParent = tableView.getParent();
        tableViewScrollPane = (SSScrollPane) tableViewParent;

        /**
         * Add/create the editor components
         */
        SSTableView.ColumnSequence tableViewColumns = tableView.getColumns();
        SSTablePane.ColumnSequence tablePaneColumns = tablePane.getColumns();

        for (int i = 0, colNum = tableViewColumns.getLength(); i < colNum; i++) {
            tableViewColumns.get(i).setCellRenderer(new SSTableViewTextAreaCellRenderer());
//            tableViewColumns.get(i).setCellRenderer(new SSTableViewTextInputCellRenderer());

            /**
             * Add a new column to the table pane to match the table view column
             */
            SSTablePane.Column tablePaneColumn = new SSTablePane.Column();
            tablePaneColumn.setWidth(tableView.getColumnBounds(i).width);
            tablePaneColumns.add(tablePaneColumn);

            /**
             * Determine which component to use as the editor for this column
             */
            String colName = tableViewColumns.get(i).getName();
            org.apache.pivot.wtk.Component editorComponent = null;
            if (colName != null) {
                editorComponent = cellEditors.get(colName);
            }

            /**
             * Default to a text area
             */
            if (editorComponent == null) {
//                if(colIdx == i) {
//                    SSTextArea textArea = SSTextAreaFactory.createTextArea(SSTextArea.TextAreaType.Cell);
//                    if (colIdx == i) {
//                        SSTextAreaFactory.setSelect(textArea);
//                        textArea.setTextBindType(BindType.BOTH);
//                    } else {
//                        if (rowIdx % 2 != 0) {
//                            SSTextAreaFactory.setAlternate(textArea);
//                        }
//                    }
//                    textArea.setTextKey(colName);
//                    editorComponent = textArea;
//                    tableView.getDecorators().removeAll();
//                } else {
//                    SSLabel label = SSLabelFactory.createLabel();
//                    label.setTextKey(colName);
//                    editorComponent = label;
//                    tableView.getDecorators().removeAll();
//                }

                SSTextArea textArea = SSTextAreaFactory.createTextArea(SSTextArea.TextAreaType.Cell);

                if (colIdx == i) {
                    SSTextAreaFactory.setSelect(textArea);
                    textArea.setTextBindType(BindType.BOTH);
                    final String text = (charactor == null) ? textArea.getText() : textArea.getText() + charactor;
                    textArea.setText(text);
                } else {
                    if (rowIdx % 2 != 0) {
                        SSTextAreaFactory.setAlternate(textArea);
                    }
                }
                textArea.setTextKey(colName);
                editorComponent = textArea;
                tableView.getDecorators().removeAll();

//                SSTextInput textInput = SSTextInputFactory.createTextInput();
//
//                if (colIdx == i) {
//                    SSTextInputFactory.setSelect(textInput);
//                    textInput.setTextBindType(BindType.BOTH);
//                    final String text = (charactor == null) ? textInput.getText() : textInput.getText() + charactor;
//                    textInput.setText(text);
//                } else {
//                    if (rowIdx % 2 == 0) {
//                        SSTextInputFactory.setDefault(textInput);
//                    } else {
//                        SSTextInputFactory.setAlternate(textInput);
//                    }
//                }
//                textInput.setTextKey(colName);
//                editorComponent = textInput;
//                tableView.getDecorators().removeAll();
            }
            /**
             * Add the editor component to the table pane
             */
            editorRow.add(editorComponent);
        }

        /**
         * Get the data being edited
         */
        org.apache.pivot.collections.List<?> tableData = tableView.getTableData();
        Object tableRow = tableData.get(rowIdx);

        /**
         * Load the row data into the editor components
         */
        tablePane.load(tableRow);

        /**
         * Get the row bounds
         */
        Bounds rowBounds = tableView.getRowBounds(rowIdx);
        rowImage.bounds = rowBounds;

        /**
         * Scroll to make the row as visible as possible
         */
        //Logger.log.debug("rowBounds.x:{}, rowBounds.y:{}", rowBounds.x, rowBounds.y);
        //Logger.log.debug("rowBounds.width:{}, rowBounds.height:{}", rowBounds.width, rowBounds.height);
        tableView.scrollAreaToVisible(rowBounds.x, rowBounds.y, rowBounds.width, rowBounds.height);

        /**
         * Constrain the bounds by what is visible through viewport ancestors
         */
        rowBounds = tableView.getVisibleArea(rowBounds);
        org.apache.pivot.wtk.Point location = tableView.mapPointToAncestor(tableView.getDisplay(), rowBounds.x, rowBounds.y);

        /**
         * Set size and location and match scroll left
         */
        setPreferredWidth(rowBounds.width);
        setLocation(location.x, location.y + (rowBounds.height - getPreferredHeight(-1)) / 2);

        if (tableViewScrollPane != null) {
            //Logger.log.debug("tableViewScrollPane.getScrollLeft: {}", tableViewScrollPane.getScrollLeft());
            scrollPane.setScrollLeft(tableViewScrollPane.getScrollLeft());
        }

        /**
         * Open the editor
         */
        open(tableView.getWindow());

        /**
         * Start the transition
         */
        cardPane.setSelectIndex(EDITOR_CARD_INDEX);
    }

    @Override
    @SuppressWarnings("unchecked")
    public void endEdit(boolean result) {
        //log.debug("result: {}", result);

        final int editColIndex = 0x0000FFFF & (editIndexPair >> 16);
        int editRowIndex = 0x0000FFFF & (editIndexPair);
        if (result) {
            // Update the row data
            org.apache.pivot.collections.List<Object> tableData = (org.apache.pivot.collections.List<Object>) tableView.getTableData();

            Object tableRow = tableData.get(editRowIndex);
            tablePane.store(tableRow);

            if (tableData.getComparator() == null) {
                tableData.update(editRowIndex, tableRow);
            } else {
                tableData.remove(editRowIndex, 1);
                tableData.add(tableRow);

                // Re-select the item, and make sure it's visible
                editRowIndex = tableData.indexOf(tableRow);
                tableView.setSelectIndex(editRowIndex);
                tableView.scrollAreaToVisible(tableView.getRowBounds(editRowIndex));
            }
            editIndexPair = (editColIndex << 16) | editRowIndex;
        }

        if (cardPane.getSelectIndex() == EDITOR_CARD_INDEX) {
            cardPane.setSelectIndex(IMAGE_CARD_INDEX);
        } else {
            // Remove the editor components
            SSTablePane.ColumnSequence tablePaneColumns = tablePane.getColumns();
            tablePaneColumns.remove(0, tablePaneColumns.getLength());
            editorRow.remove(0, editorRow.getLength());

            getOwner().moveToFront();
            tableView.requestFocus();

            tableView = null;
            editIndexPair = (-1 << 16) | -1;

            tableViewScrollPane = null;

            close();
        }
    }

    @Override
    public boolean isEditing() {
        return (tableView != null);
    }

    /**
     * Returns the effect that is applied when the editor opens or closes.
     *
     * @return The edit effect, or <tt>null</tt> for no effect.
     */
    public CardPaneSkin.SelectionChangeEffect getEditEffect() {
        return (CardPaneSkin.SelectionChangeEffect) cardPane.getStyles().get("selectionChangeEffect");
    }

    /**
     * Sets the effect that is applied when the editor opens or closes.
     *
     * @param editEffect The edit effect, or <tt>null</tt> for no effect.
     */
    public void setEditEffect(CardPaneSkin.SelectionChangeEffect editEffect) {
        cardPane.getStyles().put("selectionChangeEffect", editEffect);
    }

    /**
     * Returns the edit effect duration.
     *
     * @return The effect duration in milliseconds.
     */
    public int getEditEffectDuration() {
        return (Integer) cardPane.getStyles().get("selectionChangeDuration");
    }


    /**
     * Sets the edit effect duration.
     *
     * @param effectDuration The effect duration in milliseconds.
     */
    public void setEditEffectDuration(int effectDuration) {
        cardPane.getStyles().put("selectionChangeDuration", effectDuration);
    }

    /**
     * Returns the edit effect rate.
     *
     * @return The effect rate.
     */
    public int getEditEffectRate() {
        return (Integer) cardPane.getStyles().get("selectionChangeRate");
    }

    /**
     * Sets the edit effect rate.
     *
     * @param effectRate The effect rate.
     */
    public void setEditEffectRate(int effectRate) {
        cardPane.getStyles().put("selectionChangeRate", effectRate);
    }

    @Override
    public void open(Display display, Window owner) {
        if (tableView == null) {
            throw new IllegalStateException();
        }

        super.open(display, owner);
        display.getContainerMouseListeners().add(displayMouseHandler);
    }

    @Override
    public void close() {
        Display display = getDisplay();
        display.getContainerMouseListeners().remove(displayMouseHandler);

        super.close();
    }

    @Override
    public boolean keyPressed(int keyCode, Keyboard.KeyLocation keyLocation) {
        boolean consumed;
        Logger.log.info("keyCode: {}", keyCode);

        if (keyCode == Keyboard.KeyCode.ENTER) {
            endEdit(true);
            consumed = true;
        } else if (keyCode == Keyboard.KeyCode.ESCAPE) {
            endEdit(false);
            consumed = true;
        } else {
            consumed = false;
        }

        return consumed;
    }

}
