/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.editor;

import jp.pegasus.config.Config;
import jp.pegasus.log.Logger;
import jp.pegasus.pivot.content.SSTreeViewNodeRenderer;
import jp.pegasus.pivot.model.SSData;
import jp.pegasus.pivot.model.SSFileData;
import jp.pegasus.pivot.ss.SSTreeNode;
import jp.pegasus.pivot.ss.SSTreeView;
import jp.pegasus.util.JsonUtil;
import org.apache.pivot.collections.List;
import org.apache.pivot.collections.Sequence;
import org.apache.pivot.wtk.*;

public class SSTreeViewNodeEditor extends Window implements SSTreeView.NodeEditor {
    private SSTreeView treeView = null;
    private Tree.Path path = null;

    private TextInput textInput = new TextInput();

    private ContainerMouseListener displayMouseHandler = new ContainerMouseListener.Adapter() {
        @Override
        public boolean mouseDown(Container container, Mouse.Button button, int x, int y) {
            Display display = (Display) container;
            Window window = (Window) display.getComponentAt(x, y);

            if (window != SSTreeViewNodeEditor.this) {
                endEdit(true);
            }

            return false;
        }

        @Override
        public boolean mouseWheel(Container container, Mouse.ScrollType scrollType,
                                  int scrollAmount, int wheelRotation, int x, int y) {
            Display display = (Display) container;
            Window window = (Window) display.getComponentAt(x, y);

            return (window != SSTreeViewNodeEditor.this);
        }
    };

    public SSTreeViewNodeEditor() {
        setContent(textInput);
    }

    public SSTreeView getSSTreeView() {
        return treeView;
    }

    public Tree.Path getPath() {
        return path;
    }

    public TextInput getTextInput() {
        return textInput;
    }

    @Override
    public void beginEdit(SSTreeView treeView, Tree.Path path) {
        this.treeView = treeView;
        this.path = path;

        // Get the data being edited
        List<?> treeData = treeView.getTreeData();
        SSTreeNode treeNode = (SSTreeNode) Sequence.Tree.get(treeData, path);

        textInput.setText(treeNode.getText());
        textInput.selectAll();

        // Get the node bounds
        Bounds nodeBounds = treeView.getNodeBounds(path);
        final int nodeIndent = treeView.getNodeIndent(path.getLength());
        nodeBounds = new Bounds(nodeBounds.x + nodeIndent, nodeBounds.y,
        nodeBounds.width - nodeIndent, nodeBounds.height);

        // Render the node data
        SSTreeViewNodeRenderer nodeRenderer = (SSTreeViewNodeRenderer) treeView.getNodeRenderer();
        nodeRenderer.render(treeNode, path, treeView.getRowIndex(path), treeView, false, false,
        SSTreeView.NodeCheckState.UNCHECKED, false, false);
        nodeRenderer.setSize(nodeBounds.width, nodeBounds.height);

        // Get the text bounds
        Bounds textBounds = nodeRenderer.getTextBounds();

        // Calculate the bounds of what is being edited
        Insets padding = (Insets) textInput.getStyles().get("padding");
        Bounds editBounds = new Bounds(
        nodeBounds.x + textBounds.x - (padding.left + 1),
        nodeBounds.y,
        nodeBounds.width - textBounds.x + (padding.left + 1),
        nodeBounds.height);

        // Scroll to make the node as visible as possible
        treeView.scrollAreaToVisible(
        editBounds.x,
        editBounds.y,
        textBounds.width + padding.left + 1,
        editBounds.height);

        // Constrain the bounds by what is visible through viewport ancestors
        editBounds = treeView.getVisibleArea(editBounds);
        Point location = treeView.mapPointToAncestor(treeView.getDisplay(), editBounds.x, editBounds.y);

        textInput.setPreferredWidth(editBounds.width);
        setLocation(location.x, location.y + (editBounds.height - getPreferredHeight(-1)) / 2);

        // Open the editor
        open(treeView.getWindow());
    }

    @Override
    @SuppressWarnings("unchecked")
    public void endEdit(boolean result) {
        if (result) {
            // Update the node data
            String text = textInput.getText();

            List<?> treeData = treeView.getTreeData();
            SSTreeNode treeNode = (SSTreeNode) Sequence.Tree.get(treeData, path);
            treeNode.setText(text);

            Window window = treeView.getWindow();
            if (window != null) {
                SSData data = (SSData) window.getAttribute(SSData.DEFAULT_TYPE);
                final int fileIndex = data.getSelectFileIndex();
                SSFileData fileData = data.getFileDataList().get(fileIndex);
                fileData.setFileName(treeNode.getText());
                Logger.log.info("treeNode.getText: {}", treeNode.getText());
                JsonUtil.writeGrandData(data, Config.APP_DATA_FILE_PATH);
            }

            // Get a reference to node's parent
            final int n = path.getLength();
            List<SSTreeNode> parentData;

            if (n == 1) {
                parentData = (List<SSTreeNode>) treeData;
            } else {
                Tree.Path parentPath = new Tree.Path(path, n - 1);
                parentData = (List<SSTreeNode>) Sequence.Tree.get(treeData, parentPath);
            }

            if (parentData.getComparator() == null) {
                int index = path.get(n - 1);
                parentData.remove(index, 1);
                parentData.insert(treeNode, index);
            } else {
                parentData.remove(path.get(n - 1), 1);
                parentData.add(treeNode);

                // Re-select the node, and make sure it's visible
                path = new Tree.Path(path, n - 1);
                path.add(parentData.indexOf(treeNode));
                treeView.setSelectedPath(path);
                treeView.scrollAreaToVisible(treeView.getNodeBounds(path));
            }
        }

        getOwner().moveToFront();
        treeView.requestFocus();

        treeView = null;
        path = null;

        close();
    }

    @Override
    public boolean isEditing() {
        return (treeView != null);
    }

    @Override
    public void open(Display display, Window owner) {
        if (treeView == null) {
            throw new IllegalStateException();
        }

        super.open(display, owner);
        display.getContainerMouseListeners().add(displayMouseHandler);

        requestFocus();
    }

    @Override
    public void close() {
        Display display = getDisplay();
        display.getContainerMouseListeners().remove(displayMouseHandler);

        super.close();
    }

    @Override
    public boolean keyPressed(int keyCode, Keyboard.KeyLocation keyLocation) {
        boolean consumed;

        if (keyCode == Keyboard.KeyCode.ENTER) {
            endEdit(true);
            consumed = true;
        } else if (keyCode == Keyboard.KeyCode.ESCAPE) {
            endEdit(false);
            consumed = true;
        } else {
            consumed = false;
        }

        return consumed;
    }
}
