/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.media;

import jp.pegasus.pivot.listener.SSDrawingListener;
import jp.pegasus.pivot.listener.drawing.CanvasListener;
import jp.pegasus.pivot.media.drawing.Canvas;
import org.apache.pivot.util.ListenerList;
import org.apache.pivot.wtk.Bounds;
import org.apache.pivot.wtk.Dimensions;
import org.apache.pivot.wtk.GraphicsUtilities;
import org.apache.pivot.wtk.media.Image;

import java.awt.*;


/**
 * Image representing a vector drawing.
 */
public class SSDrawing extends Image {
    private static class DrawingListenerList extends ListenerList<SSDrawingListener>
    implements SSDrawingListener {
        @Override
        public void canvasChanged(SSDrawing drawing, Canvas previousCanvas) {
            for (SSDrawingListener listener : this) {
                listener.canvasChanged(drawing, previousCanvas);
            }
        }

        @Override
        public void backgroundChanged(SSDrawing drawing, Paint previousBackground) {
            for (SSDrawingListener listener : this) {
                listener.backgroundChanged(drawing, previousBackground);
            }
        }
    }

    private Canvas canvas = null;
    private Paint background = null;

    private Dimensions size = null;

    private int baseline = -1;

    private CanvasListener canvasListener = new CanvasListener() {
        @Override
        public void regionUpdated(Canvas canvas, int x, int y, int width, int height) {
            imageListeners.regionUpdated(SSDrawing.this, x, y, width, height);
        }

        @Override
        public void canvasInvalidated(Canvas canvas) {
            int previousWidth = size.width;
            int previousHeight = size.height;

            invalidate();

            imageListeners.sizeChanged(SSDrawing.this, previousWidth, previousHeight);
        }
    };

    private DrawingListenerList drawingListeners = new DrawingListenerList();

    public SSDrawing() {
        setCanvas(new Canvas());
    }

    @Override
    public int getWidth() {
        validate();
        return size.width;
    }

    @Override
    public int getHeight() {
        validate();
        return size.height;
    }

    @Override
    public int getBaseline() {
        return baseline;
    }

    public void setBaseline(int baseline) {
        int previousBaseline = this.baseline;

        if (baseline != previousBaseline) {
            this.baseline = baseline;
            imageListeners.baselineChanged(this, previousBaseline);
        }
    }

    private void invalidate() {
        size = null;
    }

    private void validate() {
        if (size == null) {
            int width, height;
            if (canvas == null) {
                width = 0;
                height = 0;
            } else {
                Bounds canvasBounds = canvas.getBounds();
                width = Math.max(canvasBounds.x + canvasBounds.width, 0);
                height = Math.max(canvasBounds.y + canvasBounds.height, 0);
            }

            size = new Dimensions(width, height);
        }
    }

    public Canvas getCanvas() {
        return canvas;
    }

    public void setCanvas(Canvas canvas) {
        Canvas previousCanvas = this.canvas;

        if (previousCanvas != canvas) {
            this.canvas = canvas;

            if (previousCanvas != null) {
                previousCanvas.getCanvasListeners().remove(canvasListener);
            }

            if (canvas != null) {
                canvas.getCanvasListeners().add(canvasListener);
            }

            size = null;

            drawingListeners.canvasChanged(this, previousCanvas);
        }
    }

    public Paint getBackground() {
        return background;
    }

    public void setBackground(Paint background) {
        Paint previousBackground = this.background;
        if (previousBackground != background) {
            this.background = background;
            drawingListeners.backgroundChanged(this, previousBackground);
        }
    }

    public final void setBackground(String background) {
        if (background == null) {
            throw new IllegalArgumentException("background is null.");
        }

        setBackground(GraphicsUtilities.decodePaint(background));
    }

    @Override
    public void paint(Graphics2D graphics) {
        int width = getWidth();
        int height = getHeight();

        graphics.clipRect(0, 0, width, height);

        if (background != null) {
            graphics.setPaint(background);
            graphics.fillRect(0, 0, width, height);
        }

        if (canvas != null) {
            // TODO Make this configurable?
            graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
            RenderingHints.VALUE_ANTIALIAS_ON);

            canvas.draw(graphics);
        }
    }

    public ListenerList<SSDrawingListener> getDrawingListeners() {
        return drawingListeners;
    }
}
