/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.skin;

import jp.pegasus.log.Logger;
import jp.pegasus.pivot.editor.SSTableViewRowEditor;
import jp.pegasus.pivot.listener.tableview.SSTableViewColumnListener;
import jp.pegasus.pivot.listener.tableview.SSTableViewListener;
import jp.pegasus.pivot.listener.tableview.SSTableViewRowListener;
import jp.pegasus.pivot.listener.tableview.SSTableViewSelectListener;
import jp.pegasus.pivot.ss.SSScrollPane;
import jp.pegasus.pivot.ss.tableview.SSTableView;
import jp.pegasus.util.ColorUtil;
import org.apache.pivot.collections.ArrayList;
import org.apache.pivot.collections.Dictionary;
import org.apache.pivot.collections.List;
import org.apache.pivot.collections.Sequence;
import org.apache.pivot.util.Filter;
import org.apache.pivot.wtk.*;
import org.apache.pivot.wtk.Component;
import org.apache.pivot.wtk.skin.ComponentSkin;
import org.apache.pivot.wtk.skin.terra.TerraTheme;

import java.awt.*;

public class SSTableViewSkin extends ComponentSkin implements SSTableView.Skin,
SSTableViewListener, SSTableViewColumnListener, SSTableViewRowListener, SSTableViewSelectListener {

    private boolean validateSelect = false;
    private Font font;

    // Base Color
    //-------------------------------------------------
    private Color color;
    private Color disabledColor;
    private Color backgroundColor;
    private Color alternateRowBackgroundColor;
    private Color highlightBackgroundColor;
    private Color horizontalGridColor;
    private Color verticalGridColor;

    private Color inactiveSelectColor;
    private Color inactiveSelectBackgroundColor;
    private Color columnSelectColor;
    private Color columnSelectHorizontalGridColor;
    //-------------------------------------------------

    // Select
    //-------------------------------------------------
    private Color selectColor;
    private Color selectBackgroundColor;
    private Color selectHorizontalGridColor;
    private Color selectVerticalGridColor;
    private Color selectRangeGridColor;
    private Color selectCellBackgroundColor;

    // Copy
    //-------------------------------------------------
    private Color copyColor;
    private Color copyBackgroundColor;
    private Color copyHorizontalGridColor;
    private Color copyVerticalGridColor;
    private Color copyRangeGridColor;

    // Cut
    //-------------------------------------------------
    private Color cutColor;
    private Color cutBackgroundColor;
    private Color cutHorizontalGridColor;
    private Color cutVerticalGridColor;
    private Color cutRangeGridColor;

    // Paste
    //-------------------------------------------------
    private Color pasteColor;
    private Color pasteBackgroundColor;
    private Color pasteHorizontalGridColor;
    private Color pasteVerticalGridColor;
    private Color pasteRangeGridColor;

    // Status
    //-------------------------------------------------
    private boolean showHighlight;
    private boolean showHorizontalGridLines;
    private boolean showVerticalGridLines;
    private boolean includeTrailingVerticalGridLine;
    private boolean includeTrailingHorizontalGridLine;
    private boolean variableRowHeight;

    private ArrayList<Integer> columnWidths = null;
    private ArrayList<Integer> rowBoundaries = null;
    private int fixedRowHeight = -1;
    private int defaultWidthColumnCount = 0;

    private int highlightIndex = -1;
    private int selectIndex = -1;

    public SSTableViewSkin() {
        TerraTheme theme = (TerraTheme) Theme.getTheme();
        font = theme.getFont();
        color = theme.getColor(1);
        disabledColor = theme.getColor(7);
        backgroundColor = theme.getColor(4);
        selectColor = theme.getColor(4);
        selectBackgroundColor = theme.getColor(14);
        inactiveSelectColor = theme.getColor(1);
        inactiveSelectBackgroundColor = theme.getColor(9);
        highlightBackgroundColor = theme.getColor(10);
        alternateRowBackgroundColor = theme.getColor(11);
        columnSelectColor = null;
        columnSelectHorizontalGridColor = null;
        horizontalGridColor = theme.getColor(11);
        verticalGridColor = theme.getColor(11);

        showHighlight = true;
        showHorizontalGridLines = true;
        showVerticalGridLines = true;
        includeTrailingVerticalGridLine = false;
        includeTrailingHorizontalGridLine = false;
    }

    @Override
    @SuppressWarnings("unchecked")
    public void paint(Graphics2D graphics) {
        SSTableView tableView = (SSTableView) getComponent();

        List<Object> tableData = (List<Object>) tableView.getTableData();
        SSTableView.ColumnSequence columns = tableView.getColumns();

        // Status
        // ------------------------------------------------------------------------------------------
        final boolean isSelect = tableView.isSelect();
        final boolean isCopied = tableView.isCopied();
        final boolean isCutOut = tableView.isCutOut();
        final boolean isPasted = tableView.isPasted();

        final int width = getWidth();
        final int height = getHeight();
        //log.debug("width:{}, height:{}", width, height);
        //log.debug("tableView.getBounds:{}", tableView.getBounds());

        // Paint the background
        if (backgroundColor != null) {
            graphics.setPaint(backgroundColor);
            graphics.fillRect(0, 0, width, height);
        }

        // ------------------------------------------------------------------------------------------
        // Ensure that we only paint items that are visible
        // ------------------------------------------------------------------------------------------
        final int rowCount = tableData.getLength();
        int rowStart = 0;
        int rowEnd = rowCount - 1;

        final int colCount = columns.getLength();
        int colStart = 0;
        int colEnd = colCount - 1;

        Rectangle clipBounds = graphics.getClipBounds();
        //log.debug("clipBounds:{}", clipBounds.toString());
        if (clipBounds != null) {
            if (variableRowHeight) {
                rowStart = getRowAt(clipBounds.y);
                if (rowStart == -1) {
                    rowStart = rowCount - 1;
                }

                if (rowEnd != -1) {
                    int clipBottom = clipBounds.y + clipBounds.height - 1;
                    clipBottom = Math.min(clipBottom, rowBoundaries.get(rowEnd) - 1);
                    rowEnd = getRowAt(clipBottom);
                }
            } else {
                rowStart = Math.max(rowStart, (int) Math.floor(clipBounds.y / (double) (fixedRowHeight + 1)));
                rowEnd = Math.min(rowEnd, (int) Math.ceil((clipBounds.y + clipBounds.height) / (double) (fixedRowHeight + 1)) - 1);
            }

            colStart = getColumnAt(clipBounds.x + 1);
            if (colStart == -1) {
                colStart = colCount - 1;
            }

            if (colEnd != -1) {
                int clipRight = clipBounds.x + clipBounds.width - 1;
                clipRight = Math.min(clipRight, getColumnBoundary(colEnd) - 1);
                colEnd = getColumnAt(clipRight);
            }
        }

        final int colStartX = getColumnX(colStart);
        //log.debug("clipBounds:{}", clipBounds.toString());
        //log.debug("rowStart:{}, rowEnd:{}", rowStart, rowEnd);

        // Paint the row background
        if (alternateRowBackgroundColor != null) {
            for (int rowIndex = rowStart; rowIndex <= rowEnd; rowIndex++) {
                final int rowY = getRowY(rowIndex);
                final int rowHeight = getRowHeight(rowIndex);
                if (rowIndex % 2 > 0) {
                    graphics.setPaint(alternateRowBackgroundColor);
                    graphics.fillRect(0, rowY, width, rowHeight + 1);
                }
            }
        }

        // Paint the column backgrounds
        int colX = 0;
        if (columnSelectColor != null) {
            graphics.setColor(columnSelectColor);
            colX = colStartX;

            for (int colIndex = colStart; colIndex <= colEnd; colIndex++) {
                SSTableView.Column col = columns.get(colIndex);
                final int colWidth = columnWidths.get(colIndex);

                final String colName = col.getName();
                SortDirection sortDirection = tableView.getSort().get(colName);
                if (sortDirection != null) {
                    graphics.fillRect(colX, 0, colWidth, height);
                }
                colX += (colWidth + 1);
            }
        }

        // ------------------------------------------------------------------------------------------
        // Pant grid lines
        // ------------------------------------------------------------------------------------------
        // Paint the vertical grid lines
        if (showVerticalGridLines) {
            graphics.setPaint(verticalGridColor);
            colX = colStartX;
            for (int colIndex = colStart; colIndex <= colEnd; colIndex++) {
                colX += columnWidths.get(colIndex);
                if (colIndex < colCount - 1 || includeTrailingVerticalGridLine) {
                    GraphicsUtilities.drawLine(graphics, colX, clipBounds.y, clipBounds.height, Orientation.VERTICAL);
                }
                colX++;
            }
        }

        // Paint the horizontal grid lines
        if (showHorizontalGridLines) {
            graphics.setPaint(horizontalGridColor);
            for (int rowIndex = rowStart; rowIndex <= rowEnd; rowIndex++) {
                final int gridY = getRowY(rowIndex + 1) - 1;
                if (rowIndex < rowCount - 1
                || includeTrailingHorizontalGridLine) {
                    GraphicsUtilities.drawLine(graphics, clipBounds.x, gridY, clipBounds.width, Orientation.HORIZONTAL);
                }
            }
        }

        // ------------------------------------------------------------------------------------------
        // paint select area
        // ------------------------------------------------------------------------------------------
        final long selIndexQuartet = tableView.getSelectIndexQuartet();
        final long cpyIndexQuartet = tableView.getCopyIndexQuartet();
        final long cutIndexQuartet = tableView.getCutIndexQuartet();
        final long pstIndexQuartet = tableView.getPasteIndexQuartet();
        boolean cellSelect;
        boolean cellCopied;
        boolean cellCutOut;
        boolean cellPasted;
        Color blendBackgrondColor;
        Color blendHrztlGridColor;
        Color blendVrtclGridColor;

        final int cellIndexPair = tableView.getSelectCellIndexPair();
        final int cellColIndex = 0x0000FFFF & (cellIndexPair >> 16);
        final int cellRowIndex = 0x0000FFFF & (cellIndexPair);

        for (int rowIndex = rowStart; rowIndex <= rowEnd; rowIndex++) {
            Object rowData = tableData.get(rowIndex);
            final boolean rowHighlighted =
            (rowIndex == highlightIndex && tableView.getSelectMode() != SSTableView.SelectMode.NONE);
            final boolean rowDisabled = tableView.isRowDisabled(rowIndex);
            final int rowY = getRowY(rowIndex);
            final int rowHeight = getRowHeight(rowIndex);

            // Paint the cells
            colX = colStartX;

            for (int columnIndex = colStart; columnIndex <= colEnd; columnIndex++) {
                final int currentIndexPair = (columnIndex << 16) | rowIndex;
                final boolean incledeSelect = checkCell(selIndexQuartet, currentIndexPair);
                final boolean incledeCopy = checkCell(cpyIndexQuartet, currentIndexPair);
                final boolean incledeCut = checkCell(cutIndexQuartet, currentIndexPair);
                final boolean incledePaste = checkCell(pstIndexQuartet, currentIndexPair);

                if (isSelect) {
                    if (false) {
                    }
                    /**
                     * Select
                     */
                    else if (!isCopied && !isCutOut && !isPasted && incledeSelect) {
                        blendBackgrondColor = selectBackgroundColor;
                        blendHrztlGridColor = selectHorizontalGridColor;
                        blendVrtclGridColor = selectVerticalGridColor;
                        cellSelect = true;
                        cellCopied = false;
                        cellCutOut = false;
                        cellPasted = false;
                    }
                    /**
                     * Copy
                     */
                    else if (isCopied && !isCutOut && !isPasted) {
                        if (incledeSelect && incledeCopy) {
                            blendBackgrondColor = ColorUtil.blend(selectBackgroundColor, copyBackgroundColor, 0.5f);
                            blendHrztlGridColor = ColorUtil.blend(selectHorizontalGridColor, copyHorizontalGridColor, 0.5f);
                            blendVrtclGridColor = ColorUtil.blend(selectVerticalGridColor, copyVerticalGridColor, 0.5f);
                            cellSelect = true;
                            cellCopied = true;
                        } else if (!incledeSelect && incledeCopy) {
                            blendBackgrondColor = copyBackgroundColor;
                            blendHrztlGridColor = copyHorizontalGridColor;
                            blendVrtclGridColor = copyVerticalGridColor;
                            cellSelect = false;
                            cellCopied = true;
                        } else if (incledeSelect && !incledeCopy) {
                            blendBackgrondColor = selectBackgroundColor;
                            blendHrztlGridColor = selectHorizontalGridColor;
                            blendVrtclGridColor = selectVerticalGridColor;
                            cellSelect = true;
                            cellCopied = false;
                        } else {
                            blendBackgrondColor = backgroundColor;
                            blendHrztlGridColor = horizontalGridColor;
                            blendVrtclGridColor = verticalGridColor;
                            cellSelect = false;
                            cellCopied = false;
                        }
                        cellCutOut = false;
                        cellPasted = false;
                    }
                    /**
                     * Cut
                     */
                    else if (!isCopied && isCutOut && !isPasted) {
                        if (incledeSelect && incledeCut) {
                            blendBackgrondColor = ColorUtil.blend(selectBackgroundColor, cutBackgroundColor, 0.5f);
                            blendHrztlGridColor = ColorUtil.blend(selectHorizontalGridColor, cutHorizontalGridColor, 0.5f);
                            blendVrtclGridColor = ColorUtil.blend(selectVerticalGridColor, cutVerticalGridColor, 0.5f);
                            cellSelect = true;
                            cellCutOut = true;
                        } else if (!incledeSelect && incledeCut) {
                            blendBackgrondColor = cutBackgroundColor;
                            blendHrztlGridColor = cutHorizontalGridColor;
                            blendVrtclGridColor = cutVerticalGridColor;
                            cellSelect = false;
                            cellCutOut = true;
                        } else if (incledeSelect && !incledeCut) {
                            blendBackgrondColor = selectBackgroundColor;
                            blendHrztlGridColor = selectHorizontalGridColor;
                            blendVrtclGridColor = selectVerticalGridColor;
                            cellSelect = true;
                            cellCutOut = false;
                        } else {
                            blendBackgrondColor = backgroundColor;
                            blendHrztlGridColor = horizontalGridColor;
                            blendVrtclGridColor = verticalGridColor;
                            cellSelect = false;
                            cellCutOut = false;
                        }
                        cellCopied = false;
                        cellPasted = false;
                    }
                    /**
                     * Paste
                     */
                    else if (!isCopied && !isCutOut && isPasted) {
                        if (incledeSelect && incledePaste) {
                            blendBackgrondColor = ColorUtil.blend(selectBackgroundColor, pasteBackgroundColor, 0.5f);
                            blendHrztlGridColor = ColorUtil.blend(selectHorizontalGridColor, pasteHorizontalGridColor, 0.5f);
                            blendVrtclGridColor = ColorUtil.blend(selectVerticalGridColor, pasteVerticalGridColor, 0.5f);
                            cellSelect = true;
                            cellPasted = true;
                        } else if (!incledeSelect && incledePaste) {
                            blendBackgrondColor = pasteBackgroundColor;
                            blendHrztlGridColor = pasteHorizontalGridColor;
                            blendVrtclGridColor = pasteVerticalGridColor;
                            cellSelect = false;
                            cellPasted = true;
                        } else if (incledeSelect && !incledePaste) {
                            blendBackgrondColor = selectBackgroundColor;
                            blendHrztlGridColor = selectHorizontalGridColor;
                            blendVrtclGridColor = selectVerticalGridColor;
                            cellSelect = true;
                            cellPasted = false;
                        } else {
                            blendBackgrondColor = backgroundColor;
                            blendHrztlGridColor = horizontalGridColor;
                            blendVrtclGridColor = verticalGridColor;
                            cellSelect = false;
                            cellPasted = false;
                        }
                        cellCopied = false;
                        cellCutOut = false;
                    } else {
                        blendBackgrondColor = backgroundColor;
                        blendHrztlGridColor = horizontalGridColor;
                        blendVrtclGridColor = verticalGridColor;
                        cellSelect = false;
                        cellCopied = false;
                        cellCutOut = false;
                        cellPasted = false;
                    }
                } else {
                    blendBackgrondColor = backgroundColor;
                    blendHrztlGridColor = horizontalGridColor;
                    blendVrtclGridColor = verticalGridColor;
                    cellSelect = false;
                    cellCopied = false;
                    cellCutOut = false;
                    cellPasted = false;
                }

                SSTableView.Column col = columns.get(columnIndex);
                SSTableView.CellRenderer cellRenderer = col.getCellRenderer();

                final int colWidth = columnWidths.get(columnIndex);

                if (cellSelect || cellCopied || cellCutOut || cellPasted) {
                    if (rowIndex == cellRowIndex && columnIndex == cellColIndex) {
                        graphics.setPaint(selectCellBackgroundColor);
                    } else {
                        if (rowIndex % 2 == 0) {
                            blendBackgrondColor = ColorUtil.lighter(blendBackgrondColor, 0.03f);
                        }
                        graphics.setPaint(blendBackgrondColor);
                    }
                    graphics.fillRect(colX - 1, rowY - 1, colWidth + 1, rowHeight + 1);

                    graphics.setPaint(blendVrtclGridColor);
                    GraphicsUtilities.drawLine(graphics, colX - 1, rowY - 1, rowHeight + 1, Orientation.VERTICAL);
                    graphics.setPaint(blendHrztlGridColor);
                    GraphicsUtilities.drawLine(graphics, colX - 1, rowY - 1, colWidth + 1, Orientation.HORIZONTAL);
                }

                Graphics2D rendererGraphics = (Graphics2D) graphics.create(colX, rowY, colWidth, rowHeight);

                cellRenderer.render(rowData, rowIndex, columnIndex, tableView, col.getName(),
                cellSelect, cellCopied, cellCutOut, cellPasted, rowHighlighted, rowDisabled);

                cellRenderer.setSize(colWidth, rowHeight);
                cellRenderer.paint(rendererGraphics);
                rendererGraphics.dispose();

                colX += (colWidth + 1);
            }
        }

        if (isSelect) {
            final int[] selectRange = tableView.getSelectRange();
            final Rectangle selectRect = new Rectangle(selectRange[0] - 1, selectRange[1] - 1, selectRange[2] + 1, selectRange[3] + 1);

            // Copy
            if (isCopied) {
                Graphics2D copyGraphics = (Graphics2D) graphics.create();
                final int[] copyRange = tableView.getCopyRange();
                copyGraphics.setPaint(copyRangeGridColor);
                copyGraphics.drawRect(copyRange[0] - 1, copyRange[1] - 1, copyRange[2] + 1, copyRange[3] + 1);

                final Rectangle copyRect = new Rectangle(copyRange[0] - 1, copyRange[1] - 1, copyRange[2] + 1, copyRange[3] + 1);
                final Rectangle insecCopyRect = copyRect.intersection(selectRect);
                if (!insecCopyRect.isEmpty()) {
                    Color blendCopyRangeGridColor = ColorUtil.blend(selectRangeGridColor, copyRangeGridColor, 0.5f);
                    copyGraphics.setPaint(blendCopyRangeGridColor);
                    copyGraphics.drawRect(insecCopyRect.x, insecCopyRect.y, insecCopyRect.width, insecCopyRect.height);
                }
            }

            // Cut
            if (isCutOut) {
                Graphics2D cutGraphics = (Graphics2D) graphics.create();
                final int[] cutRange = tableView.getCutRange();
                cutGraphics.setPaint(cutRangeGridColor);
                cutGraphics.drawRect(cutRange[0] - 1, cutRange[1] - 1, cutRange[2] + 1, cutRange[3] + 1);

                final Rectangle cutRect = new Rectangle(cutRange[0] - 1, cutRange[1] - 1, cutRange[2] + 1, cutRange[3] + 1);
                final Rectangle insecCutRect = cutRect.intersection(selectRect);
                if (!insecCutRect.isEmpty()) {
                    Color blendCutRangeGridColor = ColorUtil.blend(selectRangeGridColor, cutRangeGridColor, 0.5f);
                    cutGraphics.setPaint(blendCutRangeGridColor);
                    cutGraphics.drawRect(insecCutRect.x, insecCutRect.y, insecCutRect.width, insecCutRect.height);
                }
            }

            // Paste
            if (isPasted) {
                Graphics2D pasteGraphics = (Graphics2D) graphics.create();
                final int[] pasteRange = tableView.getPasteRange();
                pasteGraphics.setPaint(pasteRangeGridColor);
                pasteGraphics.drawRect(pasteRange[0] - 1, pasteRange[1] - 1, pasteRange[2] + 1, pasteRange[3] + 1);

                final Rectangle pasteRect = new Rectangle(pasteRange[0] - 1, pasteRange[1] - 1, pasteRange[2] + 1, pasteRange[3] + 1);
                final Rectangle insecPasteRect = pasteRect.intersection(selectRect);
                if (!insecPasteRect.isEmpty()) {
                    Color blendPasteRangeGridColor = ColorUtil.blend(selectRangeGridColor, pasteRangeGridColor, 0.5f);
                    pasteGraphics.setPaint(blendPasteRangeGridColor);
                    pasteGraphics.drawRect(insecPasteRect.x, insecPasteRect.y, insecPasteRect.width, insecPasteRect.height);
                }
            }

            // Other
            if (!((isCopied || isCutOut) && isPasted)) {
                SSTableViewRowEditor tableViewRowEditor = tableView.getRowEditor();
                if (tableViewRowEditor == null) {
                    Graphics2D selectGraphics = (Graphics2D) graphics.create();
                    selectGraphics.setPaint(selectRangeGridColor);
                    selectGraphics.drawRect(selectRange[0] - 1, selectRange[1] - 1, selectRange[2] + 1, selectRange[3] + 1);
                }
            }
        }
    }

    private boolean checkCell(final long indexQuartet, final int currentIndexPair) {
        final int colIndexFr = (int) (0x000000000000FFFFl & (indexQuartet >> 48));
        final int rowIndexFr = (int) (0x000000000000FFFFl & (indexQuartet >> 32));
        final int colIndexTo = (int) (0x000000000000FFFFl & (indexQuartet >> 16));
        final int rowIndexTo = (int) (0x000000000000FFFFl & (indexQuartet));

        final int currentCol = 0x0000FFFF & (currentIndexPair >> 16);
        final int currentRow = 0x0000FFFF & (currentIndexPair);

        if (
        colIndexFr <= currentCol && currentCol <= colIndexTo &&
        rowIndexFr <= currentRow && currentRow <= rowIndexTo) {
            return true;
        } else {
            return false;
        }
    }

    private int getRowY(final int rowIndex) {
        int rowY;
        if (variableRowHeight) {
            if (rowIndex == 0) {
                rowY = 0;
            } else {
//                rowY = rowBoundaries.get(rowIndex - 1);
                rowY = rowBoundaries.get(rowIndex - 1) + 1;
            }
        } else {
            rowY = rowIndex * (fixedRowHeight + 1);
        }
        return rowY;
    }

    private int getRowHeight(final int rowIndex) {
        int rowHeight;
        if (variableRowHeight) {
            rowHeight = rowBoundaries.get(rowIndex);

            if (rowIndex > 0) {
//                rowHeight -= rowBoundaries.get(rowIndex - 1);
                rowHeight -= rowBoundaries.get(rowIndex - 1) + 1;
//                Logger.log.debug("rowHeight: {}", rowHeight);
            }
        } else {
            rowHeight = fixedRowHeight;
        }

        return rowHeight;
    }

    @Override
    @SuppressWarnings("unchecked")
    public void layout() {
        columnWidths = getColumnWidths((SSTableView) getComponent(), getWidth());

        SSTableView tableView = (SSTableView) getComponent();
        SSTableView.ColumnSequence columns = tableView.getColumns();

        if (variableRowHeight) {
            org.apache.pivot.collections.List<Object> tableData = (org.apache.pivot.collections.List<Object>) tableView.getTableData();

            final int n = tableData.getLength();
            rowBoundaries = new ArrayList<Integer>(n);

            int rowY = 0;
            for (int i = 0; i < n; i++) {
                Object rowData = tableData.get(i);

                int rowHeight = 0;
                for (int colIdx = 0, colCount = columns.getLength(); colIdx < colCount; colIdx++) {
                    SSTableView.Column column = columns.get(colIdx);
                    SSTableView.CellRenderer cellRenderer = column.getCellRenderer();

                    final int columnWidth = columnWidths.get(colIdx);

                    cellRenderer.render(rowData, i, colIdx, tableView, column.getName(), false, false, false, false, false, false);
                    rowHeight = Math.max(rowHeight, cellRenderer.getPreferredHeight(columnWidth));
                }

                rowY += rowHeight;
                rowBoundaries.add(rowY);
                rowY++;
            }
        } else {
            fixedRowHeight = calculateFixedRowHeight(tableView);
        }

        if (validateSelect) {
            // Ensure that the select is visible
            Sequence<Span> selectRanges = tableView.getSelectRanges();

            if (selectRanges.getLength() > 0) {
                final int rangeStart = selectRanges.get(0).start;
                final int rangeEnd = selectRanges.get(selectRanges.getLength() - 1).end;

                Bounds selectBounds = getRowBounds(rangeStart);
                selectBounds = selectBounds.union(getRowBounds(rangeEnd));

                Bounds visibleSelectBounds = tableView.getVisibleArea(selectBounds);
                if (visibleSelectBounds != null
                && visibleSelectBounds.height < selectBounds.height) {
                    tableView.scrollAreaToVisible(selectBounds);
                }
            }
        }

        validateSelect = false;
    }

    /**
     * Calculates the table row height, which is determined as the maximum
     * preferred height of all cell renderers.
     */
    private int calculateFixedRowHeight(SSTableView tableView) {
        int fixedRowHeight = 0;
        SSTableView.ColumnSequence columns = tableView.getColumns();

        for (int i = 0, columnNumber = columns.getLength(); i < columnNumber; i++) {
            SSTableView.Column column = columns.get(i);
            SSTableView.CellRenderer cellRenderer = column.getCellRenderer();
            cellRenderer.render(null, -1, i, tableView, column.getName(), false, false, false, false, false, false);

            fixedRowHeight = Math.max(fixedRowHeight, cellRenderer.getPreferredHeight(-1));
        }

        return fixedRowHeight;
    }

    // Table view skin methods
    @Override
    public int getRowAt(final int y) {
        if (y < 0) {
            throw new IllegalArgumentException("y is negative");
        }

        SSTableView tableView = (SSTableView) getComponent();

        int rowIndex;
        if (variableRowHeight) {
            if (y == 0) {
                rowIndex = 0;
            } else {
                rowIndex = ArrayList.binarySearch(rowBoundaries, y);
                if (rowIndex < 0) {
//                    rowIndex = 0;
                    rowIndex = -(rowIndex + 1);
                }
            }
        } else {
            rowIndex = (y / (fixedRowHeight + 1));

            @SuppressWarnings("unchecked")
            org.apache.pivot.collections.List<Object> tableData = (org.apache.pivot.collections.List<Object>) tableView.getTableData();
            if (rowIndex >= tableData.getLength()) {
                rowIndex = 0;
//                rowIndex = -1;
            }
        }

        return rowIndex;
    }

    @Override
    public int getPreferredHeight(int width) {
        int preferredHeight = 0;

        SSTableView tableView = (SSTableView) getComponent();

        final int n = tableView.getTableData().getLength();

        if (variableRowHeight) {
            ArrayList<Integer> columnWidths = getColumnWidths(tableView, width);

            for (int i = 0; i < n; i++) {
                preferredHeight += getVariableRowHeight(i, columnWidths);
            }
        } else {
            int fixedRowHeight = calculateFixedRowHeight(tableView);
            preferredHeight = fixedRowHeight * n;
        }

        // Include space for horizontal grid lines
        preferredHeight += (n - 1);

        if (includeTrailingHorizontalGridLine) {
            preferredHeight++;
        }

        return preferredHeight;
    }

    @Override
    public int getColumnAt(int x) {
        if (x < 0) {
            throw new IllegalArgumentException("x is negative");
        }

        SSTableView tableView = (SSTableView) getComponent();

//        int columnIndex = -1;
        int columnIndex = 0;

        int i = 0;
        int n = tableView.getColumns().getLength();
        int columnX = 0;
        while (i < n && x > columnX) {
            columnX += (columnWidths.get(i) + 1);
            i++;
        }

        if (x <= columnX) {
            columnIndex = i - 1;
        }

        return (columnIndex < 0) ? 0 : columnIndex;
    }

    @Override
    @SuppressWarnings("unchecked")
    public Bounds getCellBounds(final int rowIndex, final int colIndex) {
        SSTableView tableView = (SSTableView) getComponent();
        org.apache.pivot.collections.List<Object> tableData = (org.apache.pivot.collections.List<Object>) tableView.getTableData();

        if (rowIndex < 0 || rowIndex >= tableData.getLength()) {
            Logger.log.debug("rowIndex:{}", rowIndex);
            Logger.log.debug("tableData.getLength():{}", tableData.getLength());
            throw new IndexOutOfBoundsException();
        }

        int cellX = 0;
        for (int i = 0; i < colIndex; i++) {
            cellX += (columnWidths.get(i) + 1);
        }

        final int rowHeight = getRowHeight(rowIndex);

        return new Bounds(cellX, rowIndex * (rowHeight + 1), columnWidths.get(colIndex), rowHeight);
    }

    private int getColumnX(final int columnIndex) {
        int cellX = 0;

        for (int i = 0; i < columnIndex; i++) {
            cellX += (columnWidths.get(i) + 1);
        }

        return cellX;
    }

    public int getColumnBoundary(final int columnIndex) {
        int cellX = -1;
        for (int i = 0; i <= columnIndex; i++) {
            cellX += (columnWidths.get(i) + 1);
        }
        return cellX;
    }

    private ArrayList<Integer> createColumnBoundaries(final int columnCount) {
        ArrayList<Integer> columnBoundaries = new ArrayList<Integer>(columnCount);
        for (int i = 0; i < columnCount; i++) {
            columnBoundaries.add((columnWidths.get(i) + 1));
        }

        return columnBoundaries;
    }

    public static ArrayList<Integer> getColumnWidths(SSTableView tableView, final int width) {
        int fixedWidth = 0;
        int relativeWidth = 0;

        SSTableView.ColumnSequence columns = tableView.getColumns();
        final int columnLength = columns.getLength();

        ArrayList<Integer> columnWidths = new ArrayList<Integer>(columnLength);

        for (int i = 0; i < columnLength; i++) {
            SSTableView.Column column = columns.get(i);

            if (column.isRelative()) {
                columnWidths.add(0);
                relativeWidth += column.getWidth();
            } else {
                int columnWidth = column.getWidth();

                if (columnWidth == -1) {
                    // Calculate the maximum cell width
                    columnWidth = 0;

                    SSTableView.CellRenderer cellRenderer = column.getCellRenderer();
                    org.apache.pivot.collections.List<?> tableData = tableView.getTableData();

                    int rowIndex = 0;
                    for (Object rowData : tableData) {
                        cellRenderer.render(
                        rowData, rowIndex++, i, tableView, column.getName(), false, false, false, false, false, false);
                        columnWidth = Math.max(cellRenderer.getPreferredWidth(-1), columnWidth);
                    }
                }

                columnWidth = Math.min(Math.max(columnWidth, column.getMinimumWidth()), column.getMaximumWidth());
                columnWidths.add(columnWidth);
                fixedWidth += columnWidth;
            }
        }

        fixedWidth += columnLength - 1;
        final int variableWidth = Math.max(width - fixedWidth, 0);

        for (int i = 0; i < columnLength; i++) {
            SSTableView.Column column = columns.get(i);

            if (column.isRelative()) {
                final int columnWidth = (int) Math.round(
                (double) (column.getWidth() * variableWidth) / (double) relativeWidth);
                columnWidths.update(i, Math.min(Math.max(columnWidth, column.getMinimumWidth()),
                column.getMaximumWidth()));
            }
        }

        return columnWidths;
    }

    @Override
    public Bounds getColumnBounds(final int columnIndex) {
        int columnX = 0;
        for (int i = 0; i < columnIndex; i++) {
            columnX += (columnWidths.get(i) + 1);
        }

        return new Bounds(columnX, 0, columnWidths.get(columnIndex), getHeight());
    }

    public void selectRangesChanged(SSTableView tableView, Sequence<Span> previousSelectRanges) {
        if (previousSelectRanges != null
        && previousSelectRanges != tableView.getSelectRanges()) {
            if (tableView.isValid()) {
                // Repaint the area occupied by the previous select
                if (previousSelectRanges != null
                && previousSelectRanges.getLength() > 0) {
                    final int rangeStart = previousSelectRanges.get(0).start;
                    final int rangeEnd = previousSelectRanges.get(previousSelectRanges.getLength() - 1).end;

                    Bounds previousSelectBounds = getRowBounds(rangeStart);
                    previousSelectBounds = previousSelectBounds.union(getRowBounds(rangeEnd));

                    repaintComponent(previousSelectBounds);
                }

                // Repaint the area occupied by the current select
                Sequence<Span> selectRanges = tableView.getSelectRanges();
                if (selectRanges.getLength() > 0) {
                    int rangeStart = selectRanges.get(0).start;
                    int rangeEnd = selectRanges.get(selectRanges.getLength() - 1).end;

                    Bounds selectBounds = getRowBounds(rangeStart);
                    selectBounds = selectBounds.union(getRowBounds(rangeEnd));
                    repaintComponent(selectBounds);

                    // Ensure that the select is visible
                    Bounds visibleSelectBounds = tableView.getVisibleArea(selectBounds);
                    if (visibleSelectBounds != null
                    && visibleSelectBounds.height < selectBounds.height) {
                        // TODO Repainting the entire component is a workaround for PIVOT-490
                        repaintComponent();

                        //tableView.scrollAreaToVisible(selectBounds);
                    }
                }
            } else {
                validateSelect = true;
            }
        }
    }

    @Override
    public Bounds getRowBounds(final int rowIndex) {
        return new Bounds(0, getRowY(rowIndex), getWidth(), getRowHeight(rowIndex));
    }

    @Override
    public boolean mouseMove(Component component, int x, int y) {
        boolean consumed = super.mouseMove(component, x, y);

        SSTableView tableView = (SSTableView) component;

        final int previousHighlightIndex = this.highlightIndex;
        highlightIndex = getRowAt(y);

        if (previousHighlightIndex != highlightIndex && tableView.getSelectMode() != SSTableView.SelectMode.NONE && showHighlight) {
            if (previousHighlightIndex != -1) {
                repaintComponent(getRowBounds(previousHighlightIndex));
            }

            if (highlightIndex != -1) {
                repaintComponent(getRowBounds(highlightIndex));
            }
        }

        return consumed;
    }

    @Override
    public boolean mouseUp(Component component, Mouse.Button button, int x, int y) {
        Logger.log.debug("");
        boolean consumed = super.mouseUp(component, button, x, y);

        SSTableView tableView = (SSTableView) component;
        if (selectIndex != -1 && tableView.getFirstSelectIndex() != tableView.getLastSelectIndex()) {
            tableView.setSelectIndex(selectIndex);
            selectIndex = -1;
        }

        return consumed;
    }

    @Override
    public boolean mouseWheel(Component component, Mouse.ScrollType scrollType, int scrollAmount, int wheelRotation, int x, int y) {
        //log.debug("");
        SSTableView tableView = (SSTableView) component;

        @SuppressWarnings("unchecked")
        org.apache.pivot.collections.List<Object> tableData =
        (org.apache.pivot.collections.List<Object>) tableView.getTableData();

        int rowOriginalStart = 0;
        int rowOriginalEnd = tableData.getLength() - 1;

        Bounds visibleArea = tableView.getVisibleArea();
        if (variableRowHeight) {
            rowOriginalStart = getRowAt(visibleArea.y);
            if (rowOriginalStart == -1) {
                rowOriginalStart = tableData.getLength();
            }
            if (rowOriginalEnd != -1) {
                int clipBottom = visibleArea.y + visibleArea.height - 1;
                clipBottom = Math.min(clipBottom, rowBoundaries.get(rowOriginalEnd) - 1);
                rowOriginalEnd = getRowAt(clipBottom);
            }
        } else {
            rowOriginalStart = Math.max(rowOriginalStart, (int) Math.floor(visibleArea.y
            / (double) (fixedRowHeight + 1)));
            rowOriginalEnd = Math.min(rowOriginalEnd, (int) Math.ceil((visibleArea.y
            + visibleArea.height) / (double) (fixedRowHeight + 1)) - 1);
        }
//        log.debug("rowOriginalStart:{}, rowOriginalEnd:{}", rowOriginalStart, rowOriginalEnd);

        int rowStart = rowOriginalStart + wheelRotation;
        int rowEnd = rowOriginalEnd + wheelRotation;
        if (rowStart < 0) {
            rowStart = 0;
        } else if (rowEnd > tableData.getLength()) {
            rowStart -= (rowEnd - tableData.getLength() + 1);
        }
//        log.debug("rowStart:{}, rowEnd:{}", rowStart, rowEnd);

        Bounds rowBounds = tableView.getRowBounds(rowStart);
        SSScrollPane scrollPane = (SSScrollPane) tableView.getParent();
        scrollPane.setScrollTop(rowBounds.y);

        return false;
    }

    @Override
    public boolean mouseClick(Component component, Mouse.Button button, int x, int y, int count) {
        boolean consumed = super.mouseClick(component, button, x, y, count);

        SSTableView tableView = (SSTableView) component;
        if (selectIndex != -1 && count == 2) {
            SSTableView.RowEditor rowEditor = tableView.getRowEditor();

            if (rowEditor != null) {
                if (rowEditor.isEditing()) {
                    rowEditor.endEdit(true);
                }

                rowEditor.beginEdit(tableView, getColumnAt(x), selectIndex);
            }
        }

        selectIndex = -1;

        return consumed;
    }

    @Override
    public boolean mouseDown(Component component, Mouse.Button button, int x, int y) {
        boolean consumed = super.mouseDown(component, button, x, y);
        //Logger.log.debug("consumed: {}", consumed);

        SSTableView tableView = (SSTableView) component;
        int rowIndex = getRowAt(y);

        if (rowIndex >= 0 && !tableView.isRowDisabled(rowIndex)) {
            SSTableView.SelectMode selectMode = tableView.getSelectMode();

            if (button == Mouse.Button.RIGHT) {
                if (selectMode != SSTableView.SelectMode.NONE && !tableView.isRowSelect(rowIndex)) {
                    tableView.setSelectIndex(rowIndex);
                }
            } else {
                Keyboard.Modifier commandModifier = Platform.getCommandModifier();

                if (Keyboard.isPressed(Keyboard.Modifier.SHIFT) && selectMode == SSTableView.SelectMode.MULTI) {
                    Filter<?> disabledRowFilter = tableView.getDisabledRowFilter();

                    if (disabledRowFilter == null) {
                        // Select the range
                        int startIndex = tableView.getFirstSelectIndex();
                        int endIndex = tableView.getLastSelectIndex();
                        Span selectRange = (rowIndex > startIndex) ?
                        new Span(startIndex, rowIndex) : new Span(rowIndex, endIndex);

                        ArrayList<Span> selectRanges = new ArrayList<Span>();
                        selectRanges.add(selectRange);

                        tableView.setSelectRanges(selectRanges);
                    }
                } else if (Keyboard.isPressed(commandModifier) && selectMode == SSTableView.SelectMode.MULTI) {
                    // Toggle the item's select state
                    if (tableView.isRowSelect(rowIndex)) {
                        tableView.removeSelectIndex(rowIndex);
                    } else {
                        tableView.addSelectIndex(rowIndex);
                    }
                } else if (Keyboard.isPressed(commandModifier) && selectMode == SSTableView.SelectMode.SINGLE) {
                    // Toggle the item's select state
                    if (tableView.isRowSelect(rowIndex)) {
                        tableView.setSelectIndex(-1);
                    } else {
                        tableView.setSelectIndex(rowIndex);
                    }
                } else {
                    if (selectMode != SSTableView.SelectMode.NONE) {
                        if (tableView.isRowSelect(rowIndex)) {
                            selectIndex = rowIndex;
                        } else {
                            tableView.setSelectIndex(rowIndex);
                        }
                    }
                }
            }
        }

        tableView.requestFocus();

        return consumed;
    }

    //-------------------------------------------------------------------------------------------

    @Override
    public int getBaseline(final int width, final int height) {
        SSTableView tableView = (SSTableView) getComponent();
        @SuppressWarnings("unchecked")
        org.apache.pivot.collections.List<Object> tableData =
        (org.apache.pivot.collections.List<Object>) tableView.getTableData();

        int baseline = -1;

        SSTableView.ColumnSequence columns = tableView.getColumns();
        ArrayList<Integer> columnWidths = getColumnWidths(tableView, width);

        if (variableRowHeight) {
            final int rowHeight = getVariableRowHeight(0, columnWidths);
            Object rowData = tableData.get(0);

            for (int i = 0, columnNumber = columns.getLength(); i < columnNumber; i++) {
                SSTableView.Column column = columns.get(i);
                SSTableView.CellRenderer cellRenderer = column.getCellRenderer();
                cellRenderer.render(rowData, 0, i, tableView, column.getName(), false, false, false, false, false, false);
                baseline = Math.max(baseline, cellRenderer.getBaseline(columnWidths.get(i), rowHeight));
            }

        } else {
            final int rowHeight = calculateFixedRowHeight(tableView);

            for (int i = 0, columnNumber = columns.getLength(); i < columnNumber; i++) {
                SSTableView.Column column = columns.get(i);
                SSTableView.CellRenderer cellRenderer = column.getCellRenderer();
                cellRenderer.render(null, -1, i, tableView, column.getName(), false, false, false, false, false, false);
                baseline = Math.max(baseline, cellRenderer.getBaseline(columnWidths.get(i), rowHeight));
            }
        }

        return baseline;
    }

    @Override
    public void mouseOver(Component component) {
//        log.debug("");
//        super.mouseOver(component);
    }

    @Override
    public void mouseOut(Component component) {
//        log.debug("");
//        super.mouseOut(component);
    }

    public static int getPreferredWidth(SSTableView tableView, boolean includeTrailingVerticalGridLine) {
        int preferredWidth = 0;

        SSTableView.ColumnSequence columns = tableView.getColumns();
        org.apache.pivot.collections.List<?> tableData = tableView.getTableData();

        int n = columns.getLength();
        for (int i = 0; i < n; i++) {
            SSTableView.Column column = columns.get(i);

            if (!column.isRelative()) {
                int columnWidth = column.getWidth();

                if (columnWidth == -1) {
                    // Calculate the maximum cell width
                    columnWidth = 0;

                    SSTableView.CellRenderer cellRenderer = column.getCellRenderer();

                    int rowIndex = 0;
                    for (Object rowData : tableData) {
                        cellRenderer.render(
                        rowData, rowIndex++, i, tableView, column.getName(), false, false, false, false, false, false);
                        columnWidth = Math.max(cellRenderer.getPreferredWidth(-1), columnWidth);
                    }
                }

                preferredWidth += Math.min(Math.max(columnWidth, column.getMinimumWidth()), column.getMaximumWidth());
            } else {
                preferredWidth += column.getMinimumWidth();
            }
        }

        // Include space for vertical gridlines; even if we are not painting them,
        // the header does
        preferredWidth += (n - 1);

        if (includeTrailingVerticalGridLine) {
            preferredWidth++;
        }

        return preferredWidth;
    }

    public Color getSelectHorizontalGridColor() {
        return selectHorizontalGridColor;
    }

    public void setSelectHorizontalGridColor(Color selectHorizontalGridColor) {
        this.selectHorizontalGridColor = selectHorizontalGridColor;
    }

    public void setSelectHorizontalGridColor(String selectHorizontalGridColor) {
        if (selectHorizontalGridColor == null) {
            throw new IllegalArgumentException("selectHorizontalGridColor is null.");
        }
        setSelectHorizontalGridColor(GraphicsUtilities.decodeColor(selectHorizontalGridColor));
    }

    public Color getCopyHorizontalGridColor() {
        return copyHorizontalGridColor;
    }

    public void setCopyHorizontalGridColor(Color copyHorizontalGridColor) {
        this.copyHorizontalGridColor = copyHorizontalGridColor;
    }

    public void setCopyHorizontalGridColor(String copyHorizontalGridColor) {
        if (copyHorizontalGridColor == null) {
            throw new IllegalArgumentException("copyHorizontalGridColor is null.");
        }
        setCopyHorizontalGridColor(GraphicsUtilities.decodeColor(copyHorizontalGridColor));
    }

    public Color getCopyVerticalGridColor() {
        return copyVerticalGridColor;
    }

    public void setCopyVerticalGridColor(Color copyVerticalGridColor) {
        this.copyVerticalGridColor = copyVerticalGridColor;
    }

    public void setCopyVerticalGridColor(String copyVerticalGridColor) {
        if (copyVerticalGridColor == null) {
            throw new IllegalArgumentException("copyVerticalGridColor is null.");
        }
        setCopyVerticalGridColor(GraphicsUtilities.decodeColor(copyVerticalGridColor));
    }

    public Color getSelectVerticalGridColor() {
        return selectVerticalGridColor;
    }

    public void setSelectVerticalGridColor(Color selectVerticalGridColor) {
        this.selectVerticalGridColor = selectVerticalGridColor;
    }

    public void setSelectVerticalGridColor(String selectVerticalGridColor) {
        if (selectVerticalGridColor == null) {
            throw new IllegalArgumentException("selectVerticalGridColor is null.");
        }
        setSelectVerticalGridColor(GraphicsUtilities.decodeColor(selectVerticalGridColor));
    }

    public Color getSelectCellBackgroundColor() {
        return selectCellBackgroundColor;
    }

    public void setSelectCellBackgroundColor(Color selectCellBackgroundColor) {
        this.selectCellBackgroundColor = selectCellBackgroundColor;
    }

    public void setSelectCellBackgroundColor(String selectCellBackgroundColor) {
        if (selectCellBackgroundColor == null) {
            throw new IllegalArgumentException("selectCellBackgroundColor is null.");
        }
        setSelectCellBackgroundColor(GraphicsUtilities.decodeColor(selectCellBackgroundColor));
    }

    public Color getSelectRangeGridColor() {
        return selectRangeGridColor;
    }

    public void setSelectRangeGridColor(Color selectRangeGridColor) {
        this.selectRangeGridColor = selectRangeGridColor;
    }

    public void setSelectRangeGridColor(String selectRangeGridColor) {
        if (selectRangeGridColor == null) {
            throw new IllegalArgumentException("selectRangeGridColor is null.");
        }
        setSelectRangeGridColor(GraphicsUtilities.decodeColor(selectRangeGridColor));
    }

    public Color getCopyRangeGridColor() {
        return copyRangeGridColor;
    }

    public void setCopyRangeGridColor(Color copyRangeGridColor) {
        this.copyRangeGridColor = copyRangeGridColor;
    }

    public void setCopyRangeGridColor(String copyRangeGridColor) {
        if (copyRangeGridColor == null) {
            throw new IllegalArgumentException("copyRangeGridColor is null.");
        }
        setCopyRangeGridColor(GraphicsUtilities.decodeColor(copyRangeGridColor));
    }

    @Override
    public boolean keyPressed(Component component, int keyCode, Keyboard.KeyLocation keyLocation) {
        boolean consumed = super.keyPressed(component, keyCode, keyLocation);
        Logger.log.info("keyCode: {}, consumed: {}", keyCode, consumed);

        SSTableView tableView = (SSTableView) component;
        SSTableView.SelectMode selectMode = tableView.getSelectMode();

        switch (keyCode) {
            case Keyboard.KeyCode.UP: {
                if (selectMode != SSTableView.SelectMode.NONE) {
                    int index = tableView.getFirstSelectIndex();

                    do {
                        index--;
                    } while (index >= 0 && tableView.isRowDisabled(index));

                    if (index >= 0) {
                        if (Keyboard.isPressed(Keyboard.Modifier.SHIFT)
                        && tableView.getSelectMode() == SSTableView.SelectMode.MULTI) {
                            tableView.addSelectIndex(index);
                        } else {
                            tableView.setSelectIndex(index);
                        }
                    }

                    consumed = true;
                }

                break;
            }

            case Keyboard.KeyCode.DOWN: {
                if (selectMode != SSTableView.SelectMode.NONE) {
                    int index = tableView.getLastSelectIndex();
                    int count = tableView.getTableData().getLength();

                    do {
                        index++;
                    } while (index < count && tableView.isRowDisabled(index));

                    if (index < count) {
                        if (Keyboard.isPressed(Keyboard.Modifier.SHIFT)
                        && tableView.getSelectMode() == SSTableView.SelectMode.MULTI) {
                            tableView.addSelectIndex(index);
                        } else {
                            tableView.setSelectIndex(index);
                        }
                    }

                    consumed = true;
                }

                break;
            }
        }

        // Clear the highlight
        if (highlightIndex != -1
        && tableView.getSelectMode() != SSTableView.SelectMode.NONE
        && showHighlight) {
            repaintComponent(getRowBounds(highlightIndex));
        }

        highlightIndex = -1;

        return consumed;
    }

    public ArrayList<Integer> getRowBoundaries() {
        return rowBoundaries;
    }

    public boolean isVariableRowHeight() {
        return variableRowHeight;
    }

    public void setVariableRowHeight(boolean variableRowHeight) {
        this.variableRowHeight = variableRowHeight;
        this.rowBoundaries = null;
        this.fixedRowHeight = -1;
        invalidateComponent();
    }

    public void setFixedRowHeight(int fixedRowHeight) {
        this.fixedRowHeight = fixedRowHeight;
    }

    public int getFixedRowHeight() {
        return fixedRowHeight;
    }

    @Override
    public void install(Component component) {
        super.install(component);

        SSTableView tableView = (SSTableView) component;
        tableView.getTableViewListeners().add(this);
        tableView.getTableViewColumnListeners().add(this);
        tableView.getTableViewRowListeners().add(this);
        tableView.getTableViewSelectListeners().add(this);
    }

    @SuppressWarnings("unchecked")
    protected int getVariableRowHeight(int rowIndex, ArrayList<Integer> columnWidths) {
        SSTableView tableView = (SSTableView) getComponent();
        List<Object> tableData = (List<Object>) tableView.getTableData();

        SSTableView.ColumnSequence columns = tableView.getColumns();
        Object rowData = tableData.get(rowIndex);

        int rowHeight = 0;
        for (int i = 0, n = columns.getLength(); i < n; i++) {
            SSTableView.Column column = columns.get(i);
            SSTableView.CellRenderer cellRenderer = column.getCellRenderer();
            cellRenderer.render(rowData, rowIndex, i, tableView, column.getName(), false, false, false, false, false, false);

            rowHeight = Math.max(rowHeight, cellRenderer.getPreferredHeight(columnWidths.get(i)));
        }

        return rowHeight;
    }

    //-----------------------------------------------------------------------------------------------------------

    public Font getFont() {
        return font;
    }

    public void setFont(Font font) {
        if (font == null) {
            throw new IllegalArgumentException("font is null.");
        }

        this.font = font;
        invalidateComponent();
    }

    public final void setFont(String font) {
        if (font == null) {
            throw new IllegalArgumentException("font is null.");
        }

        setFont(decodeFont(font));
    }

    public final void setFont(Dictionary<String, ?> font) {
        if (font == null) {
            throw new IllegalArgumentException("font is null.");
        }

        setFont(Theme.deriveFont(font));
    }

    public Color getColor() {
        return color;
    }

    public void setColor(Color color) {
        if (color == null) {
            throw new IllegalArgumentException("color is null.");
        }

        this.color = color;
        repaintComponent();
    }

    public final void setColor(String color) {
        if (color == null) {
            throw new IllegalArgumentException("color is null.");
        }

        setColor(GraphicsUtilities.decodeColor(color));
    }

    public final void setColor(int color) {
        TerraTheme theme = (TerraTheme) Theme.getTheme();
        setColor(theme.getColor(color));
    }

    public Color getDisabledColor() {
        return disabledColor;
    }

    public void setDisabledColor(Color disabledColor) {
        if (disabledColor == null) {
            throw new IllegalArgumentException("disabledColor is null.");
        }

        this.disabledColor = disabledColor;
        repaintComponent();
    }

    public final void setDisabledColor(String disabledColor) {
        if (disabledColor == null) {
            throw new IllegalArgumentException("disabledColor is null.");
        }

        setDisabledColor(GraphicsUtilities.decodeColor(disabledColor));
    }

    public final void setDisabledColor(int disabledColor) {
        TerraTheme theme = (TerraTheme) Theme.getTheme();
        setDisabledColor(theme.getColor(disabledColor));
    }

    public Color getBackgroundColor() {
        return backgroundColor;
    }

    public void setBackgroundColor(Color backgroundColor) {
        this.backgroundColor = backgroundColor;
        repaintComponent();
    }

    public final void setBackgroundColor(String backgroundColor) {
        if (backgroundColor == null) {
            throw new IllegalArgumentException("backgroundColor is null.");
        }

        setBackgroundColor(GraphicsUtilities.decodeColor(backgroundColor));
    }

    public final void setBackgroundColor(int backgroundColor) {
        TerraTheme theme = (TerraTheme) Theme.getTheme();
        setBackgroundColor(theme.getColor(backgroundColor));
    }

    public Color getSelectColor() {
        return selectColor;
    }

    public void setSelectColor(Color selectColor) {
        if (selectColor == null) {
            throw new IllegalArgumentException("selectColor is null.");
        }

        this.selectColor = selectColor;
        repaintComponent();
    }

    public final void setSelectColor(String selectColor) {
        if (selectColor == null) {
            throw new IllegalArgumentException("selectColor is null.");
        }

        setSelectColor(GraphicsUtilities.decodeColor(selectColor));
    }

    public Color getCopyColor() {
        return copyColor;
    }

    public void setCopyColor(Color copyColor) {
        this.copyColor = copyColor;
    }

    public final void setCopyColor(String copyColor) {
        if (copyColor == null) {
            throw new IllegalArgumentException("copyColor is null.");
        }

        setCopyColor(GraphicsUtilities.decodeColor(copyColor));
    }

    public Color getSelectBackgroundColor() {
        return selectBackgroundColor;
    }

    public void setSelectBackgroundColor(Color selectBackgroundColor) {
        if (selectBackgroundColor == null) {
            throw new IllegalArgumentException("selectBackgroundColor is null.");
        }

        this.selectBackgroundColor = selectBackgroundColor;
        repaintComponent();
    }

    public final void setSelectBackgroundColor(String selectBackgroundColor) {
        if (selectBackgroundColor == null) {
            throw new IllegalArgumentException("selectBackgroundColor is null.");
        }

        setSelectBackgroundColor(GraphicsUtilities.decodeColor(selectBackgroundColor));
    }

    public final void setSelectBackgroundColor(int selectBackgroundColor) {
        TerraTheme theme = (TerraTheme) Theme.getTheme();
        setSelectBackgroundColor(theme.getColor(selectBackgroundColor));
    }

    public Color getCopyBackgroundColor() {
        return copyBackgroundColor;
    }

    public void setCopyBackgroundColor(Color copyBackgroundColor) {
        this.copyBackgroundColor = copyBackgroundColor;
    }

    public final void setCopyBackgroundColor(String copyBackgroundColor) {
        if (copyBackgroundColor == null) {
            throw new IllegalArgumentException("copyBackgroundColor is null.");
        }

        setCopyBackgroundColor(GraphicsUtilities.decodeColor(copyBackgroundColor));
    }

    public Color getInactiveSelectColor() {
        return inactiveSelectColor;
    }

    public void setInactiveSelectColor(Color inactiveSelectColor) {
        if (inactiveSelectColor == null) {
            throw new IllegalArgumentException("inactiveSelectColor is null.");
        }

        this.inactiveSelectColor = inactiveSelectColor;
        repaintComponent();
    }

    public final void setInactiveSelectColor(String inactiveSelectColor) {
        if (inactiveSelectColor == null) {
            throw new IllegalArgumentException("inactiveSelectColor is null.");
        }

        setInactiveSelectColor(GraphicsUtilities.decodeColor(inactiveSelectColor));
    }

    public final void setInactiveSelectColor(int inactiveSelectColor) {
        TerraTheme theme = (TerraTheme) Theme.getTheme();
        setInactiveSelectColor(theme.getColor(inactiveSelectColor));
    }

    public Color getInactiveSelectBackgroundColor() {
        return inactiveSelectBackgroundColor;
    }

    public void setInactiveSelectBackgroundColor(Color inactiveSelectBackgroundColor) {
        if (inactiveSelectBackgroundColor == null) {
            throw new IllegalArgumentException("inactiveSelectBackgroundColor is null.");
        }

        this.inactiveSelectBackgroundColor = inactiveSelectBackgroundColor;
        repaintComponent();
    }

    public final void setInactiveSelectBackgroundColor(String inactiveSelectBackgroundColor) {
        if (inactiveSelectBackgroundColor == null) {
            throw new IllegalArgumentException("inactiveSelectBackgroundColor is null.");
        }

        setInactiveSelectBackgroundColor(GraphicsUtilities.decodeColor(inactiveSelectBackgroundColor));
    }

    public final void setInactiveSelectBackgroundColor(int inactiveSelectBackgroundColor) {
        TerraTheme theme = (TerraTheme) Theme.getTheme();
        setInactiveSelectBackgroundColor(theme.getColor(inactiveSelectBackgroundColor));
    }

    public Color getHighlightBackgroundColor() {
        return highlightBackgroundColor;
    }

    public void setHighlightBackgroundColor(Color highlightBackgroundColor) {
        if (highlightBackgroundColor == null) {
            throw new IllegalArgumentException("highlightBackgroundColor is null.");
        }

        this.highlightBackgroundColor = highlightBackgroundColor;
        repaintComponent();
    }

    public final void setHighlightBackgroundColor(String highlightBackgroundColor) {
        if (highlightBackgroundColor == null) {
            throw new IllegalArgumentException("highlightBackgroundColor is null.");
        }

        setHighlightBackgroundColor(GraphicsUtilities.decodeColor(highlightBackgroundColor));
    }

    public final void setHighlightBackgroundColor(int highlightBackgroundColor) {
        TerraTheme theme = (TerraTheme) Theme.getTheme();
        setHighlightBackgroundColor(theme.getColor(highlightBackgroundColor));
    }

    public Color getAlternateRowBackgroundColor() {
        return alternateRowBackgroundColor;
    }

    public void setAlternateRowBackgroundColor(Color alternateRowBackgroundColor) {
        this.alternateRowBackgroundColor = alternateRowBackgroundColor;
        repaintComponent();
    }

    public final void setAlternateRowBackgroundColor(String alternateRowBackgroundColor) {
        if (alternateRowBackgroundColor == null) {
            throw new IllegalArgumentException("alternateRowBackgroundColor is null.");
        }

        setAlternateRowBackgroundColor(GraphicsUtilities.decodeColor(alternateRowBackgroundColor));
    }

    public final void setAlternateRowBackgroundColor(int alternateRowBackgroundColor) {
        TerraTheme theme = (TerraTheme) Theme.getTheme();
        setAlternateRowBackgroundColor(theme.getColor(alternateRowBackgroundColor));
    }

    public Color getColumnSelectColor() {
        return columnSelectColor;
    }

    public void setColumnSelectColor(Color columnSelectColor) {
        this.columnSelectColor = columnSelectColor;
        repaintComponent();
    }

    public final void setColumnSelectColor(String columnSelectColor) {
        if (columnSelectColor == null) {
            throw new IllegalArgumentException("columnSelectColor is null.");
        }

        setColumnSelectColor(GraphicsUtilities.decodeColor(columnSelectColor));
    }

    public final void setColumnSelectColor(int columnSelectColor) {
        TerraTheme theme = (TerraTheme) Theme.getTheme();
        setColumnSelectColor(theme.getColor(columnSelectColor));
    }

    public Color getColumnSelectHorizontalGridColor() {
        return columnSelectHorizontalGridColor;
    }

    public void setColumnSelectHorizontalGridColor(Color columnSelectHorizontalGridColor) {
        if (columnSelectHorizontalGridColor == null) {
            throw new IllegalArgumentException("columnSelectHorizontalGridColor is null.");
        }

        this.columnSelectHorizontalGridColor = columnSelectHorizontalGridColor;
        repaintComponent();
    }

    public final void setColumnSelectHorizontalGridColor(String columnSelectHorizontalGridColor) {
        if (columnSelectHorizontalGridColor == null) {
            throw new IllegalArgumentException("columnSelectHorizontalGridColor is null.");
        }

        setColumnSelectHorizontalGridColor(GraphicsUtilities.decodeColor(columnSelectHorizontalGridColor));
    }

    public final void setColumnSelectHorizontalGridColor(int columnSelectHorizontalGridColor) {
        TerraTheme theme = (TerraTheme) Theme.getTheme();
        setColumnSelectHorizontalGridColor(theme.getColor(columnSelectHorizontalGridColor));
    }

    public Color getHorizontalGridColor() {
        return horizontalGridColor;
    }

    public void setHorizontalGridColor(Color horizontalGridColor) {
        if (horizontalGridColor == null) {
            throw new IllegalArgumentException("horizontalGridColor is null.");
        }

        this.horizontalGridColor = horizontalGridColor;
        repaintComponent();
    }

    public final void setHorizontalGridColor(String horizontalGridColor) {
        if (horizontalGridColor == null) {
            throw new IllegalArgumentException("horizontalGridColor is null.");
        }

        setHorizontalGridColor(GraphicsUtilities.decodeColor(horizontalGridColor));
    }

    public final void setHorizontalGridColor(int horizontalGridColor) {
        TerraTheme theme = (TerraTheme) Theme.getTheme();
        setHorizontalGridColor(theme.getColor(horizontalGridColor));
    }

    public Color getVerticalGridColor() {
        return verticalGridColor;
    }

    public void setVerticalGridColor(Color verticalGridColor) {
        if (verticalGridColor == null) {
            throw new IllegalArgumentException("verticalGridColor is null.");
        }

        this.verticalGridColor = verticalGridColor;
        repaintComponent();
    }

    public final void setVerticalGridColor(String verticalGridColor) {
        if (verticalGridColor == null) {
            throw new IllegalArgumentException("verticalGridColor is null.");
        }

        setVerticalGridColor(GraphicsUtilities.decodeColor(verticalGridColor));
    }

    public final void setVerticalGridColor(int verticalGridColor) {
        TerraTheme theme = (TerraTheme) Theme.getTheme();
        setVerticalGridColor(theme.getColor(verticalGridColor));
    }

    public boolean getShowHighlight() {
        return showHighlight;
    }

    public void setShowHighlight(boolean showHighlight) {
        this.showHighlight = showHighlight;
        repaintComponent();
    }

    public boolean getShowHorizontalGridLines() {
        return showHorizontalGridLines;
    }

    public void setShowHorizontalGridLines(boolean showHorizontalGridLines) {
        this.showHorizontalGridLines = showHorizontalGridLines;
        repaintComponent();
    }

    public boolean getShowVerticalGridLines() {
        return showVerticalGridLines;
    }

    public void setShowVerticalGridLines(boolean showVerticalGridLines) {
        this.showVerticalGridLines = showVerticalGridLines;
        repaintComponent();
    }

    public boolean getIncludeTrailingVerticalGridLine() {
        return includeTrailingVerticalGridLine;
    }

    public void setIncludeTrailingVerticalGridLine(boolean includeTrailingVerticalGridLine) {
        this.includeTrailingVerticalGridLine = includeTrailingVerticalGridLine;
        invalidateComponent();
    }

    public boolean getIncludeTrailingHorizontalGridLine() {
        return includeTrailingHorizontalGridLine;
    }

    public void setIncludeTrailingHorizontalGridLine(boolean includeTrailingHorizontalGridLine) {
        this.includeTrailingHorizontalGridLine = includeTrailingHorizontalGridLine;
        invalidateComponent();
    }

    // Component state events
    @Override
    public void enabledChanged(Component component) {
        super.enabledChanged(component);

        repaintComponent();
    }

    @Override
    public void focusedChanged(Component component, Component obverseComponent) {
        super.focusedChanged(component, obverseComponent);

        repaintComponent();
    }

    @Override
    public int getPreferredWidth(int height) {
        return getPreferredWidth((SSTableView) getComponent(), includeTrailingVerticalGridLine);
    }

    @Override
    public Dimensions getPreferredSize() {
        return new Dimensions(getPreferredWidth(-1), getPreferredHeight(-1));
    }

    @Override
    public boolean isFocusable() {
        SSTableView tableView = (SSTableView) getComponent();
        return (tableView.getSelectMode() != SSTableView.SelectMode.NONE);
    }

    @Override
    public boolean isOpaque() {
        return (backgroundColor != null
        && backgroundColor.getTransparency() == Transparency.OPAQUE);
    }

    // Table view events
    @Override
    public void tableDataChanged(SSTableView tableView, List<?> previousTableData) {
        invalidateComponent();
    }

    @Override
    public void columnSourceChanged(SSTableView tableView, SSTableView previousColumnSource) {
        if (previousColumnSource != null) {
            previousColumnSource.getTableViewColumnListeners().remove(this);
        }

        SSTableView columnSource = tableView.getColumnSource();

        if (columnSource != null) {
            columnSource.getTableViewColumnListeners().add(this);
        }

        invalidateComponent();
    }

    @Override
    public void rowEditorChanged(SSTableView tableView, SSTableView.RowEditor previousRowEditor) {
        // No-op
    }

    @Override
    public void selectModeChanged(SSTableView tableView, SSTableView.SelectMode previousSelectMode) {
        repaintComponent();
    }

    @Override
    public void disabledRowFilterChanged(SSTableView tableView, Filter<?> previousDisabledRowFilter) {
        repaintComponent();
    }

    // Table view column events
    @Override
    public void columnInserted(SSTableView tableView, int index) {
        SSTableView.Column column = tableView.getColumns().get(index);

        if (column.getWidth() == -1) {
            defaultWidthColumnCount++;
        }

        invalidateComponent();
    }

    @Override
    public void columnsRemoved(SSTableView tableView, int index, Sequence<SSTableView.Column> columns) {
        for (int i = 0, n = columns.getLength(); i < n; i++) {
            SSTableView.Column column = columns.get(i);

            if (column.getWidth() == -1) {
                defaultWidthColumnCount--;
            }
        }

        invalidateComponent();
    }

    @Override
    public void columnNameChanged(SSTableView.Column column, String previousName) {
        invalidateComponent();
    }

    @Override
    public void columnHeaderDataChanged(SSTableView.Column column, Object previousHeaderData) {
        // No-op
    }

    @Override
    public void columnHeaderDataRendererChanged(SSTableView.Column column,
                                                SSTableView.HeaderDataRenderer previousHeaderDataRenderer) {
        // No-op
    }

    @Override
    public void columnWidthChanged(SSTableView.Column column, int previousWidth, boolean previousRelative) {
        if (column.getWidth() == -1) {
            defaultWidthColumnCount++;
        } else {
            defaultWidthColumnCount--;
        }

        invalidateComponent();
    }

    @Override
    public void columnWidthLimitsChanged(SSTableView.Column column, int previousMinimumWidth, int previousMaximumWidth) {
        invalidateComponent();
    }

    @Override
    public void columnFilterChanged(SSTableView.Column column, Object previousFilter) {
        // No-op
    }

    @Override
    public void columnCellRendererChanged(SSTableView.Column column, SSTableView.CellRenderer previousCellRenderer) {
        invalidateComponent();
    }

    // Table view row events
    @Override
    public void rowInserted(SSTableView tableView, int index) {
        invalidateComponent();
    }

    @Override
    public void rowsRemoved(SSTableView tableView, int index, int count) {
        invalidateComponent();
    }

    @Override
    public void rowUpdated(SSTableView tableView, int index) {
        if (variableRowHeight
        || defaultWidthColumnCount > 0) {
            invalidateComponent();
        } else {
            repaintComponent(getRowBounds(index));
        }
    }

    @Override
    public void rowsCleared(SSTableView listView) {
        invalidateComponent();
    }

    @Override
    public void rowsSorted(SSTableView tableView) {
        if (variableRowHeight) {
            invalidateComponent();
        } else {
            repaintComponent();
        }
    }

    // Table view select detail events
    @Override
    public void selectRangeAdded(SSTableView tableView, int rangeStart, int rangeEnd) {
        if (tableView.isValid()) {
            Bounds selectBounds = getRowBounds(rangeStart);
            selectBounds = selectBounds.union(getRowBounds(rangeEnd));
            repaintComponent(selectBounds);

            // Ensure that the select is visible
            Bounds visibleSelectBounds = tableView.getVisibleArea(selectBounds);
            if (visibleSelectBounds.height < selectBounds.height) {
                tableView.scrollAreaToVisible(selectBounds);
            }
        } else {
            validateSelect = true;
        }
    }

    @Override
    public void selectRangeRemoved(SSTableView tableView, int rangeStart, int rangeEnd) {
        // Repaint the area containing the removed select
        if (tableView.isValid()) {
            Bounds selectBounds = getRowBounds(rangeStart);
            selectBounds = selectBounds.union(getRowBounds(rangeEnd));
            repaintComponent(selectBounds);
        }
    }

    @Override
    public void selectRowChanged(SSTableView tableView, Object previousSelectRow) {
        // No-op
    }

    // PASTE
    public Color getPasteColor() {
        return pasteColor;
    }

    public void setPasteColor(Color pasteColor) {
        this.pasteColor = pasteColor;
    }

    public final void setPasteColor(String pasteColor) {
        if (pasteColor == null) {
            throw new IllegalArgumentException("pasteColor is null.");
        }
        setPasteColor(GraphicsUtilities.decodeColor(pasteColor));
    }

    public Color getPasteBackgroundColor() {
        return pasteBackgroundColor;
    }

    public void setPasteBackgroundColor(Color pasteBackgroundColor) {
        this.pasteBackgroundColor = pasteBackgroundColor;
    }

    public final void setPasteBackgroundColor(String pasteBackgroundColor) {
        if (pasteBackgroundColor == null) {
            throw new IllegalArgumentException("pasteBackgroundColor is null.");
        }
        setPasteBackgroundColor(GraphicsUtilities.decodeColor(pasteBackgroundColor));
    }

    public Color getPasteHorizontalGridColor() {
        return pasteHorizontalGridColor;
    }

    public void setPasteHorizontalGridColor(Color pasteHorizontalGridColor) {
        this.pasteHorizontalGridColor = pasteHorizontalGridColor;
    }

    public final void setPasteHorizontalGridColor(String pasteHorizontalGridColor) {
        if (pasteHorizontalGridColor == null) {
            throw new IllegalArgumentException("pasteHorizontalGridColor is null.");
        }

        setPasteHorizontalGridColor(GraphicsUtilities.decodeColor(pasteHorizontalGridColor));
    }

    public Color getPasteVerticalGridColor() {
        return pasteVerticalGridColor;
    }

    public void setPasteVerticalGridColor(Color pasteVerticalGridColor) {
        this.pasteVerticalGridColor = pasteVerticalGridColor;
    }

    public final void setPasteVerticalGridColor(String pasteVerticalGridColor) {
        if (pasteVerticalGridColor == null) {
            throw new IllegalArgumentException("pasteVerticalGridColor is null.");
        }
        setPasteVerticalGridColor(GraphicsUtilities.decodeColor(pasteVerticalGridColor));
    }

    public Color getPasteRangeGridColor() {
        return pasteRangeGridColor;
    }

    public void setPasteRangeGridColor(Color pasteRangeGridColor) {
        this.pasteRangeGridColor = pasteRangeGridColor;
    }

    public final void setPasteRangeGridColor(String pasteRangeGridColor) {
        if (pasteRangeGridColor == null) {
            throw new IllegalArgumentException("pasteRangeGridColor is null.");
        }
        setPasteRangeGridColor(GraphicsUtilities.decodeColor(pasteRangeGridColor));
    }

    public Color getCutColor() {
        return cutColor;
    }

    public void setCutColor(Color cutColor) {
        this.cutColor = cutColor;
    }

    public final void setCutColor(String cutColor) {
        if (cutColor == null) {
            throw new IllegalArgumentException("cutColor is null.");
        }
        setCutColor(GraphicsUtilities.decodeColor(cutColor));
    }

    public Color getCutBackgroundColor() {
        return cutBackgroundColor;
    }

    public void setCutBackgroundColor(Color cutBackgroundColor) {
        this.cutBackgroundColor = cutBackgroundColor;
    }

    public final void setCutBackgroundColor(String cutBackgroundColor) {
        if (cutBackgroundColor == null) {
            throw new IllegalArgumentException("cutBackgroundColor is null.");
        }
        setCutBackgroundColor(GraphicsUtilities.decodeColor(cutBackgroundColor));
    }

    public Color getCutHorizontalGridColor() {
        return cutHorizontalGridColor;
    }

    public void setCutHorizontalGridColor(Color cutHorizontalGridColor) {
        this.cutHorizontalGridColor = cutHorizontalGridColor;
    }

    public final void setCutHorizontalGridColor(String cutHorizontalGridColor) {
        if (cutHorizontalGridColor == null) {
            throw new IllegalArgumentException("cutHorizontalGridColor is null.");
        }
        setCutHorizontalGridColor(GraphicsUtilities.decodeColor(cutHorizontalGridColor));
    }

    public Color getCutVerticalGridColor() {
        return cutVerticalGridColor;
    }

    public void setCutVerticalGridColor(Color cutVerticalGridColor) {
        this.cutVerticalGridColor = cutVerticalGridColor;
    }

    public final void setCutVerticalGridColor(String cutVerticalGridColor) {
        if (cutVerticalGridColor == null) {
            throw new IllegalArgumentException("cutVerticalGridColor is null.");
        }
        setCutVerticalGridColor(GraphicsUtilities.decodeColor(cutVerticalGridColor));
    }

    public Color getCutRangeGridColor() {
        return cutRangeGridColor;
    }

    public void setCutRangeGridColor(Color cutRangeGridColor) {
        this.cutRangeGridColor = cutRangeGridColor;
    }

    public final void setCutRangeGridColor(String cutRangeGridColor) {
        if (cutRangeGridColor == null) {
            throw new IllegalArgumentException("cutRangeGridColor is null.");
        }
        setCutRangeGridColor(GraphicsUtilities.decodeColor(cutRangeGridColor));
    }

}
