/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.skin;

import jp.pegasus.log.Logger;
import jp.pegasus.pivot.listener.textarea.SSTextAreaContentListener;
import jp.pegasus.pivot.listener.textarea.SSTextAreaListener;
import jp.pegasus.pivot.listener.textarea.SSTextAreaSelectListener;
import jp.pegasus.pivot.ss.SSTextArea;
import org.apache.pivot.collections.ArrayList;
import org.apache.pivot.collections.Dictionary;
import org.apache.pivot.collections.Sequence;
import org.apache.pivot.wtk.*;
import org.apache.pivot.wtk.Component;
import org.apache.pivot.wtk.Insets;

import java.awt.*;
import java.awt.font.FontRenderContext;
import java.awt.font.LineMetrics;
import java.awt.geom.Area;

/**
 * Text area skin.
 */
public class SSTextAreaSkin extends SSComponentSkin implements SSTextArea.Skin, SSTextAreaListener,
SSTextAreaContentListener, SSTextAreaSelectListener {
    private class BlinkCaretCallback implements Runnable {
        @Override
        public void run() {
            caretOn = !caretOn;

            if (select == null) {
                SSTextArea SSTextArea = (SSTextArea) getComponent();
                SSTextArea.repaint(caret.x, caret.y, caret.width, caret.height);
            }
        }
    }

    private class ScrollSelectCallback implements Runnable {
        @Override
        public void run() {
            SSTextArea ssTextArea = (SSTextArea) getComponent();
            int selectStart = ssTextArea.getSelectStart();
            int selectLength = ssTextArea.getSelectLength();
            int selectEnd = selectStart + selectLength - 1;

            switch (scrollDirection) {
                case UP: {
                    // Get previous offset
                    final int index = getNextInsertionPoint(mouseX, selectStart, scrollDirection);
                    if (index != -1) {
                        ssTextArea.setSelect(index, selectEnd - index + 1);
                        scrollCharacterToVisible(index + 1);
                    }
                    break;
                }

                case DOWN: {
                    // Get next offset
                    int index = getNextInsertionPoint(mouseX, selectEnd, scrollDirection);
                    if (index != -1) {
                        // If the next character is a paragraph terminator, increment
                        // the select
                        if (index < ssTextArea.getCharacterCount()
                        && ssTextArea.getCharacterAt(index) == '\n') {
                            index++;
                        }
                        ssTextArea.setSelect(selectStart, index - selectStart);
                        scrollCharacterToVisible(index - 1);
                    }
                    break;
                }
            }
        }
    }

    private int caretX = 0;
    private Rectangle caret = new Rectangle();
    private Area select = null;

    private boolean caretOn = false;

    private int anchor = -1;
    private SSTextArea.ScrollDirection scrollDirection = null;
    private int mouseX = -1;

    private BlinkCaretCallback blinkCaretCallback = new BlinkCaretCallback();
    private ApplicationContext.ScheduledCallback scheduledBlinkCaretCallback = null;

    private ScrollSelectCallback scrollSelectCallback = new ScrollSelectCallback();
    private ApplicationContext.ScheduledCallback scheduledScrollSelectCallback = null;

    private Font font;
    private Color color;
    private Color backgroundColor;
    private Color inactiveColor;
    private Color selectColor;
    private Color selectBackgroundColor;
    private Color inactiveSelectColor;
    private Color inactiveSelectBackgroundColor;
    private Insets margin;
    private boolean wrapText;
    private int tabWidth;

    private ArrayList<SSTextAreaSkinParagraphView> paragraphViews = new ArrayList<SSTextAreaSkinParagraphView>();

    private static final int SCROLL_RATE = 30;

    public SSTextAreaSkin() {
        Theme theme = Theme.getTheme();
        font = theme.getFont();
        color = Color.BLACK;
        backgroundColor = null;
        inactiveColor = Color.GRAY;
        selectColor = Color.LIGHT_GRAY;
        selectBackgroundColor = Color.BLACK;
        inactiveSelectColor = Color.LIGHT_GRAY;
        inactiveSelectBackgroundColor = Color.BLACK;
        margin = new Insets(4);
        wrapText = true;
        tabWidth = 4;
    }

    @Override
    public void install(Component component) {
        super.install(component);

        SSTextArea ssTextArea = (SSTextArea) component;
        ssTextArea.getTextAreaListeners().add(this);
        ssTextArea.getTextAreaContentListeners().add(this);
        ssTextArea.getTextAreaSelectListeners().add(this);
    }

    @Override
    public int getPreferredWidth(int height) {
        int preferredWidth = 0;

        for (SSTextAreaSkinParagraphView paragraphView : paragraphViews) {
            paragraphView.setBreakWidth(Integer.MAX_VALUE);
            preferredWidth = Math.max(preferredWidth, paragraphView.getWidth());
        }

        preferredWidth += margin.left + margin.right;

        return preferredWidth;
    }

    @Override
    public int getPreferredHeight(int width) {
        int preferredHeight = 0;

        // Include margin in constraint
        int breakWidth = (wrapText && width != -1) ? Math.max(width - (margin.left + margin.right), 0) : Integer.MAX_VALUE;

        for (SSTextAreaSkinParagraphView paragraphView : paragraphViews) {
            paragraphView.setBreakWidth(breakWidth);
            preferredHeight += paragraphView.getHeight();
        }

        preferredHeight += margin.top + margin.bottom;

        return preferredHeight;
    }

    @Override
    public Dimensions getPreferredSize() {
        int preferredWidth = 0;
        int preferredHeight = 0;

        for (SSTextAreaSkinParagraphView paragraphView : paragraphViews) {
            paragraphView.setBreakWidth(Integer.MAX_VALUE);
            preferredWidth = Math.max(preferredWidth, paragraphView.getWidth());
            preferredHeight += paragraphView.getHeight();
        }

        preferredWidth += margin.left + margin.right;
        preferredHeight += margin.top + margin.bottom;

        return new Dimensions(preferredWidth, preferredHeight);
    }

    @Override
    public void layout() {
        SSTextArea ssTextArea = (SSTextArea) getComponent();

        final int width = getWidth();
        final int breakWidth = (wrapText) ? Math.max(width - (margin.left + margin.right), 0) : Integer.MAX_VALUE;

        int y = margin.top;

        int rowOffset = 0;
        for (SSTextAreaSkinParagraphView paragraphView : paragraphViews) {
            paragraphView.setBreakWidth(breakWidth);
            paragraphView.setX(margin.left);
            paragraphView.setY(y);
            y += paragraphView.getHeight();

            paragraphView.setRowOffset(rowOffset);
            rowOffset += paragraphView.getRowCount();
        }

        updateSelect();
        caretX = caret.x;

        if (ssTextArea.isFocused()) {
            scrollCharacterToVisible(ssTextArea.getSelectStart());
            showCaret(ssTextArea.getSelectLength() == 0);
        } else {
            showCaret(false);
        }
    }

    @Override
    public int getBaseline(int width, int height) {
        FontRenderContext fontRenderContext = Platform.getFontRenderContext();
        LineMetrics lm = font.getLineMetrics("", fontRenderContext);

        return Math.round(margin.top + lm.getAscent());
    }

    @Override
    public void paint(Graphics2D graphics) {
        SSTextArea ssTextArea = (SSTextArea) getComponent();
        final int width = getWidth();
        final int height = getHeight();

        // Draw the background
        if (backgroundColor != null) {
            graphics.setPaint(backgroundColor);
            graphics.fillRect(0, 0, width, height);
        }

        // Draw the caret/select
        if (select == null) {
            if (caretOn && ssTextArea.isFocused()) {
                graphics.setColor(ssTextArea.isEditable() ? color : inactiveColor);
                graphics.fill(caret);
            }
        } else {
            graphics.setColor(ssTextArea.isFocused()
            && ssTextArea.isEditable() ? selectBackgroundColor : inactiveSelectBackgroundColor);
            graphics.fill(select);
        }

        // Draw the text
        graphics.setFont(font);
        graphics.translate(0, margin.top);

        for (int i = 0, n = paragraphViews.getLength(); i < n; i++) {
            SSTextAreaSkinParagraphView paragraphView = paragraphViews.get(i);

            int x = paragraphView.getX();
            graphics.translate(x, 0);
            paragraphView.paint(graphics);
            graphics.translate(-x, 0);

            graphics.translate(0, paragraphView.getHeight());
        }
    }

    @Override
    public boolean isOpaque() {
        return (backgroundColor != null && backgroundColor.getTransparency() == Transparency.OPAQUE);
    }

    @Override
    public int getInsertionPoint(int x, int y) {
        int index = -1;

        if (paragraphViews.getLength() > 0) {
            if (y > getHeight() - margin.bottom) {
                // Select the character at x in the first row
                SSTextAreaSkinParagraphView paragraphView = paragraphViews.get(paragraphViews.getLength() - 1);
                index = paragraphView.getNextInsertionPoint(x, -1, SSTextArea.ScrollDirection.UP)
                + paragraphView.getParagraph().getOffset();
            } else if (y < margin.top) {
                // Select the character at x in the last row
                SSTextAreaSkinParagraphView paragraphView = paragraphViews.get(0);
                index = paragraphView.getNextInsertionPoint(x, -1, SSTextArea.ScrollDirection.DOWN);
            } else {
                // Select the character at x in the row at y
                for (int i = 0, n = paragraphViews.getLength(); i < n; i++) {
                    SSTextAreaSkinParagraphView paragraphView = paragraphViews.get(i);

                    int paragraphViewY = paragraphView.getY();
                    if (y >= paragraphViewY && y < paragraphViewY + paragraphView.getHeight()) {
                        index = paragraphView.getInsertionPoint(x - paragraphView.getX(), y - paragraphViewY) + paragraphView.getParagraph().getOffset();
                        break;
                    }
                }
            }
        }

        return index;
    }

    @Override
    public int getNextInsertionPoint(int x, int from, SSTextArea.ScrollDirection direction) {
        int index = -1;

        if (paragraphViews.getLength() > 0) {
            if (from == -1) {
                int i = (direction == SSTextArea.ScrollDirection.DOWN) ? 0 : paragraphViews.getLength() - 1;

                SSTextAreaSkinParagraphView paragraphView = paragraphViews.get(i);
                index = paragraphView.getNextInsertionPoint(x - paragraphView.getX(), -1, direction);

                if (index != -1) {
                    index += paragraphView.getParagraph().getOffset();
                }
            } else {
                SSTextArea ssTextArea = (SSTextArea) getComponent();
                int i = ssTextArea.getParagraphAt(from);

                SSTextAreaSkinParagraphView paragraphView = paragraphViews.get(i);
                index = paragraphView.getNextInsertionPoint(x - paragraphView.getX(), from - paragraphView.getParagraph().getOffset(), direction);

                if (index == -1) {
                    // Move to the next or previous paragraph view
                    if (direction == SSTextArea.ScrollDirection.DOWN) {
                        paragraphView = (i < paragraphViews.getLength() - 1) ? paragraphViews.get(i + 1) : null;
                    } else {
                        paragraphView = (i > 0) ? paragraphViews.get(i - 1) : null;
                    }

                    if (paragraphView != null) {
                        index = paragraphView.getNextInsertionPoint(x - paragraphView.getX(), -1, direction);
                    }
                }

                if (index != -1) {
                    index += paragraphView.getParagraph().getOffset();
                }
            }
        }

        return index;
    }

    @Override
    public int getRowAt(int index) {
        int rowIndex = -1;

        if (paragraphViews.getLength() > 0) {
            SSTextArea SSTextArea = (SSTextArea) getComponent();
            SSTextAreaSkinParagraphView paragraphView = paragraphViews.get(SSTextArea.getParagraphAt(index));

            rowIndex = paragraphView.getRowAt(index - paragraphView.getParagraph().getOffset()) + paragraphView.getRowOffset();
        }

        return rowIndex;
    }

    public int getRowOffset(int index) {
        int rowOffset = -1;

        if (paragraphViews.getLength() > 0) {
            SSTextArea SSTextArea = (SSTextArea) getComponent();
            SSTextAreaSkinParagraphView paragraphView = paragraphViews.get(SSTextArea.getParagraphAt(index));

            rowOffset = paragraphView.getRowOffset(index - paragraphView.getParagraph().getOffset()) + paragraphView.getParagraph().getOffset();
        }

        return rowOffset;
    }

    public int getRowLength(int index) {
        int rowLength = -1;

        if (paragraphViews.getLength() > 0) {
            SSTextArea SSTextArea = (SSTextArea) getComponent();
            SSTextAreaSkinParagraphView paragraphView = paragraphViews.get(SSTextArea.getParagraphAt(index));

            rowLength = paragraphView.getRowLength(index - paragraphView.getParagraph().getOffset());
        }

        return rowLength;
    }

    @Override
    public int getRowCount() {
        int rowCount = 0;

        for (SSTextAreaSkinParagraphView paragraphView : paragraphViews) {
            rowCount += paragraphView.getRowCount();
        }

        return rowCount;
    }

    public Bounds getCharacterBounds(int index) {
        Bounds characterBounds = null;

        if (paragraphViews.getLength() > 0) {
            SSTextArea SSTextArea = (SSTextArea) getComponent();
            SSTextAreaSkinParagraphView paragraphView = paragraphViews.get(SSTextArea.getParagraphAt(index));
            characterBounds = paragraphView.getCharacterBounds(index - paragraphView.getParagraph().getOffset());

            characterBounds = new Bounds(
            characterBounds.x + paragraphView.getX(),
            characterBounds.y + paragraphView.getY(),
            characterBounds.width,
            characterBounds.height);
        }

        return characterBounds;
    }

    public Area getSelect() {
        return select;
    }

    private void scrollCharacterToVisible(final int index) {
        Bounds characterBounds = getCharacterBounds(index);

        if (characterBounds != null) {
            SSTextArea SSTextArea = (SSTextArea) getComponent();
            SSTextArea.scrollAreaToVisible(characterBounds.x, characterBounds.y, characterBounds.width, characterBounds.height);
        }
    }

    public Font getFont() {
        return font;
    }

    public void setFont(Font font) {
        if (font == null) {
            throw new IllegalArgumentException("font is null.");
        }

        this.font = font;
        invalidateComponent();
    }

    public final void setFont(String font) {
        if (font == null) {
            throw new IllegalArgumentException("font is null.");
        }
        setFont(decodeFont(font));
    }

    public final void setFont(Dictionary<String, ?> font) {
        if (font == null) {
            throw new IllegalArgumentException("font is null.");
        }
        setFont(Theme.deriveFont(font));
    }

    public Color getColor() {
        return color;
    }

    public void setColor(Color color) {
        if (color == null) {
            throw new IllegalArgumentException("color is null.");
        }
        this.color = color;
        repaintComponent();
    }

    public final void setColor(String color) {
        if (color == null) {
            throw new IllegalArgumentException("color is null.");
        }
        setColor(GraphicsUtilities.decodeColor(color));
    }

    public Color getBackgroundColor() {
        return backgroundColor;
    }

    public void setBackgroundColor(Color backgroundColor) {
        this.backgroundColor = backgroundColor;
        repaintComponent();
    }

    public final void setBackgroundColor(String backgroundColor) {
        if (backgroundColor == null) {
            throw new IllegalArgumentException("backgroundColor is null");
        }
        setBackgroundColor(GraphicsUtilities.decodeColor(backgroundColor));
    }

    public Color getInactiveColor() {
        return inactiveColor;
    }

    public void setInactiveColor(Color inactiveColor) {
        if (inactiveColor == null) {
            throw new IllegalArgumentException("inactiveColor is null.");
        }
        this.inactiveColor = inactiveColor;
        repaintComponent();
    }

    public final void setInactiveColor(String inactiveColor) {
        if (inactiveColor == null) {
            throw new IllegalArgumentException("inactiveColor is null.");
        }
        setColor(GraphicsUtilities.decodeColor(inactiveColor));
    }

    public Color getSelectColor() {
        return selectColor;
    }

    public void setSelectColor(Color selectColor) {
        if (selectColor == null) {
            throw new IllegalArgumentException("selectColor is null.");
        }
        this.selectColor = selectColor;
        repaintComponent();
    }

    public final void setSelectColor(String selectColor) {
        if (selectColor == null) {
            throw new IllegalArgumentException("selectColor is null.");
        }
        setSelectColor(GraphicsUtilities.decodeColor(selectColor));
    }

    public Color getSelectBackgroundColor() {
        return selectBackgroundColor;
    }

    public void setSelectBackgroundColor(Color selectBackgroundColor) {
        if (selectBackgroundColor == null) {
            throw new IllegalArgumentException("selectBackgroundColor is null.");
        }
        this.selectBackgroundColor = selectBackgroundColor;
        repaintComponent();
    }

    public final void setSelectBackgroundColor(String selectBackgroundColor) {
        if (selectBackgroundColor == null) {
            throw new IllegalArgumentException("selectBackgroundColor is null.");
        }
        setSelectBackgroundColor(GraphicsUtilities.decodeColor(selectBackgroundColor));
    }

    public Color getInactiveSelectColor() {
        return inactiveSelectColor;
    }

    public void setInactiveSelectColor(Color inactiveSelectColor) {
        if (inactiveSelectColor == null) {
            throw new IllegalArgumentException("inactiveSelectColor is null.");
        }
        this.inactiveSelectColor = inactiveSelectColor;
        repaintComponent();
    }

    public final void setInactiveSelectColor(String inactiveSelectColor) {
        if (inactiveSelectColor == null) {
            throw new IllegalArgumentException("inactiveSelectColor is null.");
        }
        setInactiveSelectColor(GraphicsUtilities.decodeColor(inactiveSelectColor));
    }

    public Color getInactiveSelectBackgroundColor() {
        return inactiveSelectBackgroundColor;
    }

    public void setInactiveSelectBackgroundColor(Color inactiveSelectBackgroundColor) {
        if (inactiveSelectBackgroundColor == null) {
            throw new IllegalArgumentException("inactiveSelectBackgroundColor is null.");
        }
        this.inactiveSelectBackgroundColor = inactiveSelectBackgroundColor;
        repaintComponent();
    }

    public final void setInactiveSelectBackgroundColor(String inactiveSelectBackgroundColor) {
        if (inactiveSelectBackgroundColor == null) {
            throw new IllegalArgumentException("inactiveSelectBackgroundColor is null.");
        }
        setInactiveSelectBackgroundColor(GraphicsUtilities.decodeColor(inactiveSelectBackgroundColor));
    }

    public Insets getMargin() {
        return margin;
    }

    public void setMargin(Insets margin) {
        if (margin == null) {
            throw new IllegalArgumentException("margin is null.");
        }
        this.margin = margin;
        invalidateComponent();
    }

    public final void setMargin(Dictionary<String, ?> margin) {
        if (margin == null) {
            throw new IllegalArgumentException("margin is null.");
        }
        setMargin(new Insets(margin));
    }

    public final void setMargin(int margin) {
        setMargin(new Insets(margin));
    }

    public final void setMargin(Number margin) {
        if (margin == null) {
            throw new IllegalArgumentException("margin is null.");
        }
        setMargin(margin.intValue());
    }

    public final void setMargin(String margin) {
        if (margin == null) {
            throw new IllegalArgumentException("margin is null.");
        }
        setMargin(Insets.decode(margin));
    }

    public boolean getWrapText() {
        return wrapText;
    }

    public void setWrapText(boolean wrapText) {
        this.wrapText = wrapText;
        invalidateComponent();
    }

    public int getTabWidth() {
        return tabWidth;
    }

    public void setTabWidth(int tabWidth) {
        if (tabWidth < 0) {
            throw new IllegalArgumentException("tabWidth is negative.");
        }
        this.tabWidth = tabWidth;
    }

    @Override
    public boolean mouseMove(Component component, int x, int y) {
        boolean consumed = super.mouseMove(component, x, y);

        if (Mouse.getCapturer() == component) {
            SSTextArea ssTextArea = (SSTextArea) getComponent();

            Bounds visibleArea = ssTextArea.getVisibleArea();
            visibleArea = new Bounds(visibleArea.x, visibleArea.y, visibleArea.width,
            visibleArea.height);

            if (y >= visibleArea.y && y < visibleArea.y + visibleArea.height) {
                // Stop the scroll select timer
                if (scheduledScrollSelectCallback != null) {
                    scheduledScrollSelectCallback.cancel();
                    scheduledScrollSelectCallback = null;
                }

                scrollDirection = null;
                int index = getInsertionPoint(x, y);

                if (index != -1) {
                    // Select the range
                    if (index > anchor) {
                        ssTextArea.setSelect(anchor, index - anchor);
                    } else {
                        ssTextArea.setSelect(index, anchor - index);
                    }
                }
            } else {
                if (scheduledScrollSelectCallback == null) {
                    scrollDirection = (y < visibleArea.y) ? SSTextArea.ScrollDirection.UP : SSTextArea.ScrollDirection.DOWN;
                    scheduledScrollSelectCallback = ApplicationContext.scheduleRecurringCallback(
                    scrollSelectCallback, SCROLL_RATE);

                    // Run the callback once now to scroll the select immediately
                    scrollSelectCallback.run();
                }
            }

            mouseX = x;
        } else {
            if (Mouse.isPressed(Mouse.Button.LEFT) && Mouse.getCapturer() == null && anchor != -1) {
                // Capture the mouse so we can select text
                Mouse.capture(component);
            }
        }

        return consumed;
    }

    @Override
    public boolean mouseDown(Component component, Mouse.Button button, int x, int y) {
        boolean consumed = super.mouseDown(component, button, x, y);

        SSTextArea SSTextArea = (SSTextArea) component;

        if (button == Mouse.Button.LEFT) {
            anchor = getInsertionPoint(x, y);

            if (anchor != -1) {
                if (Keyboard.isPressed(Keyboard.Modifier.SHIFT)) {
                    // Select the range
                    int selectStart = SSTextArea.getSelectStart();

                    if (anchor > selectStart) {
                        SSTextArea.setSelect(selectStart, anchor - selectStart);
                    } else {
                        SSTextArea.setSelect(anchor, selectStart - anchor);
                    }
                } else {
                    // Move the caret to the insertion point
                    SSTextArea.setSelect(anchor, 0);
                    consumed = true;
                }
            }

            caretX = caret.x;

            // Set focus to the text input
            SSTextArea.requestFocus();
        }

        return consumed;
    }

    @Override
    public boolean mouseUp(Component component, Mouse.Button button, int x, int y) {
        boolean consumed = super.mouseUp(component, button, x, y);

        if (Mouse.getCapturer() == component) {
            // Stop the scroll select timer
            if (scheduledScrollSelectCallback != null) {
                scheduledScrollSelectCallback.cancel();
                scheduledScrollSelectCallback = null;
            }
            Mouse.release();
        }
        anchor = -1;
        scrollDirection = null;
        mouseX = -1;

        return consumed;
    }

    @Override
    public boolean keyTyped(Component component, char character) {
        boolean consumed = super.keyTyped(component, character);

        if (paragraphViews.getLength() > 0) {
            SSTextArea SSTextArea = (SSTextArea) getComponent();

            if (SSTextArea.isEditable()) {
                // Ignore characters in the control range and the ASCII delete
                // character as well as meta key presses
                if (character > 0x1F && character != 0x7F && !Keyboard.isPressed(Keyboard.Modifier.META)) {
                    final int selectLength = SSTextArea.getSelectLength();

                    if (SSTextArea.getCharacterCount() - selectLength + 1 > SSTextArea.getMaximumLength()) {
                        Toolkit.getDefaultToolkit().beep();
                    } else {
                        final int selectStart = SSTextArea.getSelectStart();
                        SSTextArea.removeText(selectStart, selectLength);
                        SSTextArea.insertText(Character.toString(character), selectStart);
                    }

                    showCaret(true);
                }
            }
        }

        return consumed;
    }

    @Override
    public boolean keyPressed(Component component, int keyCode, Keyboard.KeyLocation keyLocation) {
        boolean consumed = false;
        Logger.log.info("keyCode: {}", keyCode);

        if (paragraphViews.getLength() > 0) {
            SSTextArea textArea = (SSTextArea) getComponent();
            SSTextArea.TextAreaType textAreaType = textArea.getTextAreaType();
            Keyboard.Modifier commandModifier = Platform.getCommandModifier();
            Keyboard.Modifier wordNavigationModifier = Platform.getWordNavigationModifier();

            /**
             * Enter
             */
            if (keyCode == Keyboard.KeyCode.ENTER && textArea.isEditable() && Keyboard.getModifiers() == 0) {
                Logger.log.info("ENTER");
                if (SSTextArea.TextAreaType.Sheet.equals(textAreaType)) {
                    actionEnter(textArea);
                    consumed = true;
                }
            }
            /**
             * Alt + Enter
             */
            else if (keyCode == Keyboard.KeyCode.ENTER && textArea.isEditable() && Keyboard.isPressed(Keyboard.Modifier.ALT)) {
                Logger.log.info("ALT+ENTER");
                if (SSTextArea.TextAreaType.Cell.equals(textAreaType)) {
                    actionEnter(textArea);
                    consumed = true;
                }
            }
            /**
             * Delete
             */
            else if (keyCode == Keyboard.KeyCode.DELETE && textArea.isEditable()) {
                int index = textArea.getSelectStart();

                if (index < textArea.getCharacterCount()) {
                    int count = Math.max(textArea.getSelectLength(), 1);
                    textArea.removeText(index, count);

                    consumed = true;
                }
            }
            /**
             * Backspace
             */
            else if (keyCode == Keyboard.KeyCode.BACKSPACE && textArea.isEditable()) {
                int index = textArea.getSelectStart();
                int count = textArea.getSelectLength();

                if (count == 0
                && index > 0) {
                    textArea.removeText(index - 1, 1);
                    consumed = true;
                } else {
                    textArea.removeText(index, count);
                    consumed = true;
                }
            }
            /**
             * Tab + Ctrl
             */
            else if (keyCode == Keyboard.KeyCode.TAB && Keyboard.isPressed(Keyboard.Modifier.CTRL) && textArea.isEditable()) {
                int selectLength = textArea.getSelectLength();

                StringBuilder tabBuilder = new StringBuilder();
                for (int i = 0; i < tabWidth; i++) {
                    tabBuilder.append(" ");
                }

                if (textArea.getCharacterCount() - selectLength + tabWidth > textArea.getMaximumLength()) {
                    Toolkit.getDefaultToolkit().beep();
                } else {
                    int selectStart = textArea.getSelectStart();
                    textArea.removeText(selectStart, selectLength);
                    textArea.insertText(tabBuilder, selectStart);
                }

                showCaret(true);

                consumed = true;
            }
            /**
             * Home or (Left + Meta)
             */
            else if (keyCode == Keyboard.KeyCode.HOME || (keyCode == Keyboard.KeyCode.LEFT && Keyboard.isPressed(Keyboard.Modifier.META))) {
                // Move the caret to the beginning of the line
                int selectStart = textArea.getSelectStart();
                int selectLength = textArea.getSelectLength();
                int rowOffset = getRowOffset(selectStart);

                if (Keyboard.isPressed(Keyboard.Modifier.SHIFT)) {
                    selectLength += selectStart - rowOffset;
                }

                if (selectStart >= 0) {
                    textArea.setSelect(rowOffset, selectLength);
                    scrollCharacterToVisible(rowOffset);

                    caretX = caret.x;

                    consumed = true;
                }
            }
            /**
             * End or (Right + Meta)
             */
            else if (keyCode == Keyboard.KeyCode.END || (keyCode == Keyboard.KeyCode.RIGHT && Keyboard.isPressed(Keyboard.Modifier.META))) {
                // Move the caret to the end of the line
                int selectStart = textArea.getSelectStart();
                int selectLength = textArea.getSelectLength();

                int index = selectStart + selectLength;
                int rowOffset = getRowOffset(index);
                int rowLength = getRowLength(index);

                if (Keyboard.isPressed(Keyboard.Modifier.SHIFT)) {
                    selectLength += (rowOffset + rowLength) - index;
                } else {
                    selectStart = rowOffset + rowLength;
                    if (textArea.getCharacterAt(selectStart) != '\n') {
                        selectStart--;
                    }

                    selectLength = 0;
                }

                if (selectStart + selectLength <= textArea.getCharacterCount()) {
                    textArea.setSelect(selectStart, selectLength);
                    scrollCharacterToVisible(selectStart + selectLength);

                    caretX = caret.x;
                    if (select != null) {
                        caretX += select.getBounds2D().getWidth();
                    }

                    consumed = true;
                }
            }
            /**
             * Left
             */
            else if (keyCode == Keyboard.KeyCode.LEFT) {
                int selectStart = textArea.getSelectStart();
                int selectLength = textArea.getSelectLength();

                /**
                 * Left + wordNavigationModifier
                 */
                if (Keyboard.isPressed(wordNavigationModifier)) {
                    // Move the caret to the start of the next word to the left
                    if (selectStart > 0) {
                        // Skip over any space immediately to the left
                        int index = selectStart;
                        while (index > 0 && Character.isWhitespace(textArea.getCharacterAt(index - 1))) {
                            index--;
                        }

                        // Skip over any word-letters to the left
                        while (index > 0 && !Character.isWhitespace(textArea.getCharacterAt(index - 1))) {
                            index--;
                        }

                        /**
                         * Left + wordNavigationModifier + Shift
                         */
                        if (Keyboard.isPressed(Keyboard.Modifier.SHIFT)) {
                            selectLength += selectStart - index;
                        } else {
                            selectLength = 0;
                        }

                        selectStart = index;
                    }
                }
                /**
                 * Left + Shift
                 */
                else if (Keyboard.isPressed(Keyboard.Modifier.SHIFT)) {
                    // Add the previous character to the select
                    if (selectStart > 0) {
                        selectStart--;
                        selectLength++;
                    }
                } else {
                    // Move the caret back by one character
                    if (selectLength == 0 && selectStart > 0) {
                        selectStart--;
                    }

                    // Clear the select
                    selectLength = 0;
                }

                if (selectStart >= 0) {
                    textArea.setSelect(selectStart, selectLength);
                    scrollCharacterToVisible(selectStart);

                    caretX = caret.x;

                    consumed = true;
                }
            }
            /**
             * Right
             */
            else if (keyCode == Keyboard.KeyCode.RIGHT) {
                int selectStart = textArea.getSelectStart();
                int selectLength = textArea.getSelectLength();

                /**
                 * Right + wordNavigationModifier
                 */
                if (Keyboard.isPressed(wordNavigationModifier)) {
                    // Move the caret to the start of the next word to the right
                    if (selectStart < textArea.getCharacterCount()) {
                        int index = selectStart + selectLength;

                        // Skip over any space immediately to the right
                        while (index < textArea.getCharacterCount()
                        && Character.isWhitespace(textArea.getCharacterAt(index))) {
                            index++;
                        }

                        // Skip over any word-letters to the right
                        while (index < textArea.getCharacterCount()
                        && !Character.isWhitespace(textArea.getCharacterAt(index))) {
                            index++;
                        }

                        /**
                         * Right + wordNavigationModifier + Shift
                         */
                        if (Keyboard.isPressed(Keyboard.Modifier.SHIFT)) {
                            selectLength = index - selectStart;
                        } else {
                            selectStart = index;
                            selectLength = 0;
                        }
                    }
                }
                /**
                 * Right + Shift
                 */
                else if (Keyboard.isPressed(Keyboard.Modifier.SHIFT)) {
                    // Add the next character to the select
                    selectLength++;
                } else {
                    // Move the caret forward by one character
                    if (selectLength == 0) {
                        selectStart++;
                    } else {
                        selectStart += selectLength;
                    }

                    // Clear the select
                    selectLength = 0;
                }

                if (selectStart + selectLength <= textArea.getCharacterCount()) {
                    textArea.setSelect(selectStart, selectLength);
                    scrollCharacterToVisible(selectStart + selectLength);

                    caretX = caret.x;
                    if (select != null) {
                        caretX += select.getBounds2D().getWidth();
                    }

                    consumed = true;
                }
            }
            /**
             * Up
             */
            else if (keyCode == Keyboard.KeyCode.UP) {
                final int selectStart = textArea.getSelectStart();
                final int index = getNextInsertionPoint(caretX, selectStart, SSTextArea.ScrollDirection.UP);
                if (index != -1) {
                    int selectLength;
                    /**
                     * Up + Shift
                     */
                    if (Keyboard.isPressed(Keyboard.Modifier.SHIFT)) {
                        final int selectEnd = selectStart + textArea.getSelectLength() - 1;
                        selectLength = selectEnd - index + 1;
                    } else {
                        selectLength = 0;
                    }

                    textArea.setSelect(index, selectLength);
                    scrollCharacterToVisible(index);
                }

                consumed = true;
            }
            /**
             * Down
             */
            else if (keyCode == Keyboard.KeyCode.DOWN) {
                final int selectStart = textArea.getSelectStart();
                final int selectLength = textArea.getSelectLength();

                /**
                 * Down + Shift
                 */
                if (Keyboard.isPressed(Keyboard.Modifier.SHIFT)) {
                    int from;
                    int x;
                    if (selectLength == 0) {
                        // Get next insertion point from leading select character
                        from = selectStart;
                        x = caretX;
                    } else {
                        // Get next insertion point from right edge of trailing select
                        // character
                        from = selectStart + selectLength - 1;

                        Bounds trailingSelectBounds = getCharacterBounds(from);
                        x = trailingSelectBounds.x + trailingSelectBounds.width;
                    }

                    int index = getNextInsertionPoint(x, from, SSTextArea.ScrollDirection.DOWN);
                    if (index != -1) {
                        // If the next character is a paragraph terminator and is
                        // not the final terminator character, increment the select
                        if (index < textArea.getCharacterCount() - 1 && textArea.getCharacterAt(index) == '\n') {
                            index++;
                        }

                        textArea.setSelect(selectStart, index - selectStart);
                        scrollCharacterToVisible(index);
                    }
                } else {
                    int from;
                    if (selectLength == 0) {
                        // Get next insertion point from leading select character
                        from = selectStart;
                    } else {
                        // Get next insertion point from trailing select character
                        from = selectStart + selectLength - 1;
                    }

                    final int index = getNextInsertionPoint(caretX, from, SSTextArea.ScrollDirection.DOWN);
                    if (index != -1) {
                        textArea.setSelect(index, 0);
                        scrollCharacterToVisible(index);
                    }
                }

                consumed = true;
            }
            /**
             * commandModifier
             */
            else if (Keyboard.isPressed(commandModifier)) {
                /**
                 * commandModifier + A
                 */
                if (keyCode == Keyboard.KeyCode.A) {
                    textArea.setSelect(0, textArea.getCharacterCount());
                    consumed = true;
                }
                /**
                 * commandModifier + X
                 */
                else if (keyCode == Keyboard.KeyCode.X && textArea.isEditable()) {
                    textArea.cut();
                    consumed = true;
                }
                /**
                 * commandModifier + C
                 */
                else if (keyCode == Keyboard.KeyCode.C) {
                    textArea.copy();
                    consumed = true;
                }
                /**
                 * commandModifier + V
                 */
                else if (keyCode == Keyboard.KeyCode.V && textArea.isEditable()) {
                    textArea.paste();
                    consumed = true;
                }
                /**
                 * commandModifier + Z
                 */
                else if (keyCode == Keyboard.KeyCode.Z && textArea.isEditable()) {
                    /**
                     * commandModifier + Z + Shift
                     */
                    if (!Keyboard.isPressed(Keyboard.Modifier.SHIFT)) {
                        textArea.undo();
                    }

                    consumed = true;
                }
            } else {
                consumed = super.keyPressed(component, keyCode, keyLocation);
            }
        }

        return consumed;
    }

    @Override
    public void enabledChanged(Component component) {
        super.enabledChanged(component);
        repaintComponent();
    }

    @Override
    public void focusedChanged(Component component, Component obverseComponent) {
        super.focusedChanged(component, obverseComponent);

        SSTextArea ssTextArea = (SSTextArea) getComponent();
        if (ssTextArea.isFocused()
        && ssTextArea.getSelectLength() == 0) {
            if (ssTextArea.isValid()) {
                scrollCharacterToVisible(ssTextArea.getSelectStart());
            }

            showCaret(true);
        } else {
            showCaret(false);
        }

        repaintComponent();
    }

    @Override
    public void maximumLengthChanged(SSTextArea SSTextArea, int previousMaximumLength) {
        // No-op
    }

    @Override
    public void editableChanged(SSTextArea SSTextArea) {
        // No-op
    }

    @Override
    public void paragraphInserted(SSTextArea ssTextArea, int index) {
        // Create paragraph view and add as paragraph listener
        SSTextArea.Paragraph paragraph = ssTextArea.getParagraphs().get(index);
        SSTextAreaSkinParagraphView paragraphView = new SSTextAreaSkinParagraphView(this, paragraph);
        paragraph.getParagraphListeners().add(paragraphView);

        // Insert view
        paragraphViews.insert(paragraphView, index);

        invalidateComponent();
    }

    @Override
    public void paragraphsRemoved(SSTextArea ssTextArea, int index, Sequence<SSTextArea.Paragraph> removed) {
        // Remove paragraph views as paragraph listeners
        int count = removed.getLength();

        for (int i = 0; i < count; i++) {
            SSTextArea.Paragraph paragraph = removed.get(i);
            SSTextAreaSkinParagraphView paragraphView = paragraphViews.get(i + index);
            paragraph.getParagraphListeners().remove(paragraphView);
        }

        // Remove views
        paragraphViews.remove(index, count);

        invalidateComponent();
    }

    @Override
    public void textChanged(SSTextArea SSTextArea) {
        // No-op
    }

    @Override
    public void selectChanged(SSTextArea ssTextArea, int previousSelectStart,
                              int previousSelectLength) {
        // If the text area is valid, repaint the select state; otherwise,
        // the select will be updated in layout()
        if (ssTextArea.isValid()) {
            if (select == null) {
                // Repaint previous caret bounds
                ssTextArea.repaint(caret.x, caret.y, caret.width, caret.height);
            } else {
                // Repaint previous select bounds
                Rectangle bounds = select.getBounds();
                ssTextArea.repaint(bounds.x, bounds.y, bounds.width, bounds.height);
            }

            updateSelect();

            if (select == null) {
                showCaret(ssTextArea.isFocused());
            } else {
                showCaret(false);

                // Repaint current select bounds
                Rectangle bounds = select.getBounds();
                ssTextArea.repaint(bounds.x, bounds.y, bounds.width, bounds.height);
            }
        }
    }

    private void updateSelect() {
        SSTextArea ssTextArea = (SSTextArea) getComponent();

        if (paragraphViews.getLength() > 0) {
            // Update the caret
            int selectStart = ssTextArea.getSelectStart();

            Bounds leadingSelectBounds = getCharacterBounds(selectStart);
            caret = leadingSelectBounds.toRectangle();
            caret.width = 1;

            // Update the select
            int selectLength = ssTextArea.getSelectLength();

            if (selectLength > 0) {
                int selectEnd = selectStart + selectLength - 1;
                Bounds trailingSelectBounds = getCharacterBounds(selectEnd);
                select = new Area();

                int firstRowIndex = getRowAt(selectStart);
                int lastRowIndex = getRowAt(selectEnd);

                if (firstRowIndex == lastRowIndex) {
                    select.add(new Area(new Rectangle(leadingSelectBounds.x,
                    leadingSelectBounds.y, trailingSelectBounds.x
                    + trailingSelectBounds.width - leadingSelectBounds.x,
                    trailingSelectBounds.y + trailingSelectBounds.height
                    - leadingSelectBounds.y)));
                } else {
                    int width = getWidth();

                    select.add(new Area(new Rectangle(leadingSelectBounds.x,
                    leadingSelectBounds.y, width - margin.right - leadingSelectBounds.x,
                    leadingSelectBounds.height)));

                    if (lastRowIndex - firstRowIndex > 0) {
                        select.add(new Area(new Rectangle(margin.left, leadingSelectBounds.y
                        + leadingSelectBounds.height, width - (margin.left + margin.right),
                        trailingSelectBounds.y
                        - (leadingSelectBounds.y + leadingSelectBounds.height))));
                    }

                    select.add(new Area(new Rectangle(margin.left, trailingSelectBounds.y,
                    trailingSelectBounds.x + trailingSelectBounds.width - margin.left,
                    trailingSelectBounds.height)));
                }
            } else {
                select = null;
            }
        } else {
            // Clear the caret and the select
            caret = new Rectangle();
            select = null;
        }
    }

    private void showCaret(boolean show) {
        if (scheduledBlinkCaretCallback != null) {
            scheduledBlinkCaretCallback.cancel();
        }

        if (show) {
            caretOn = true;
            scheduledBlinkCaretCallback = ApplicationContext.scheduleRecurringCallback(blinkCaretCallback, Platform.getCursorBlinkRate());

            // Run the callback once now to show the cursor immediately
            blinkCaretCallback.run();
        } else {
            scheduledBlinkCaretCallback = null;
        }
    }

    private void actionEnter(SSTextArea textArea) {
        final int index = textArea.getSelectStart();
        textArea.removeText(index, textArea.getSelectLength());
        textArea.insertText("\n", index);
    }

}
