/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.skin;


import jp.pegasus.pivot.ss.SSTextArea;
import org.apache.pivot.collections.ArrayList;
import org.apache.pivot.text.CharSequenceCharacterIterator;
import org.apache.pivot.wtk.Bounds;
import org.apache.pivot.wtk.Platform;
import org.apache.pivot.wtk.Span;
import org.apache.pivot.wtk.Visual;

import java.awt.*;
import java.awt.font.FontRenderContext;
import java.awt.font.GlyphVector;
import java.awt.font.LineMetrics;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.awt.geom.Rectangle2D;

public class SSTextAreaSkinParagraphView implements Visual, SSTextArea.ParagraphListener {
    private static class Row {
        public final GlyphVector glyphVector;
        public final int offset;

        public Row(GlyphVector glyphVector, int offset) {
            this.glyphVector = glyphVector;
            this.offset = offset;
        }
    }

    private SSTextAreaSkin ssTextAreaSkin;
    private SSTextArea.Paragraph paragraph;

    private int x = 0;
    private int y = 0;
    private float width = 0;
    private float height = 0;

    private int breakWidth = Integer.MAX_VALUE;

    private int rowOffset = 0;

    private boolean valid = false;
    private ArrayList<Row> rows = new ArrayList<Row>();

    private static final int PARAGRAPH_TERMINATOR_WIDTH = 2;

    public SSTextAreaSkinParagraphView(SSTextAreaSkin ssTextAreaSkin, SSTextArea.Paragraph paragraph) {
        this.ssTextAreaSkin = ssTextAreaSkin;
        this.paragraph = paragraph;
    }

    public SSTextArea.Paragraph getParagraph() {
        return paragraph;
    }

    public int getX() {
        return x;
    }

    public void setX(int x) {
        this.x = x;
    }

    public int getY() {
        return y;
    }

    public void setY(int y) {
        this.y = y;
    }

    public int getRowOffset() {
        return rowOffset;
    }

    public void setRowOffset(int rowOffset) {
        this.rowOffset = rowOffset;
    }

    @Override
    public int getWidth() {
        validate();
        return (int) Math.ceil(width);
    }

    @Override
    public int getHeight() {
        validate();
        return (int) Math.ceil(height);
    }

    public int getBreakWidth() {
        return breakWidth;
    }

    public void setBreakWidth(int breakWidth) {
        int previousBreakWidth = this.breakWidth;
        if (previousBreakWidth != breakWidth) {
            this.breakWidth = breakWidth;
            invalidate();
        }
    }

    @Override
    public int getBaseline() {
        return -1;
    }

    @Override
    public void paint(Graphics2D graphics) {
        SSTextArea ssTextArea = (SSTextArea) ssTextAreaSkin.getComponent();

        final int selectionStart = ssTextArea.getSelectStart();
        final int selectionLength = ssTextArea.getSelectLength();
        Span selectionRange = new Span(selectionStart, selectionStart + selectionLength - 1);

        final int paragraphOffset = paragraph.getOffset();
        Span characterRange = new Span(paragraphOffset, paragraphOffset + paragraph.getCharacters().length() - 1);

        if (selectionLength > 0 && characterRange.intersects(selectionRange)) {
            final boolean focused = ssTextArea.isFocused();
            final boolean editable = ssTextArea.isEditable();

            // Determine the selected and unselected areas
            Area selection = ssTextAreaSkin.getSelect();
            Area selectedArea = selection.createTransformedArea(AffineTransform.getTranslateInstance(-x, -y));
            Area unselectedArea = new Area();
            unselectedArea.add(new Area(new Rectangle2D.Float(0, 0, width, height)));
            unselectedArea.subtract(new Area(selectedArea));

            // Paint the unselected text
            Graphics2D unselectedGraphics = (Graphics2D) graphics.create();
            unselectedGraphics.clip(unselectedArea);
            paint(unselectedGraphics, focused, editable, false);
            unselectedGraphics.dispose();

            // Paint the selected text
            Graphics2D selectedGraphics = (Graphics2D) graphics.create();
            selectedGraphics.clip(selectedArea);
            paint(selectedGraphics, focused, editable, true);
            selectedGraphics.dispose();
        } else {
            paint(graphics, ssTextArea.isFocused(), ssTextArea.isEditable(), false);
        }
    }

    private void paint(Graphics2D graphics, boolean focused, boolean editable, boolean selected) {
        Font font = ssTextAreaSkin.getFont();
        FontRenderContext fontRenderContext = Platform.getFontRenderContext();
        LineMetrics lm = font.getLineMetrics("", fontRenderContext);
        float ascent = lm.getAscent();
        float rowHeight = ascent + lm.getDescent();

        Rectangle clipBounds = graphics.getClipBounds();

        float rowY = 0;
        for (int i = 0, n = rows.getLength(); i < n; i++) {
            Row row = rows.get(i);

            Rectangle2D textBounds = row.glyphVector.getLogicalBounds();
            float rowWidth = (float) textBounds.getWidth();
            if (clipBounds.intersects(new Rectangle2D.Float(0, rowY, rowWidth, rowHeight))) {
                if (selected) {
                    graphics.setPaint(focused && editable ?
                    ssTextAreaSkin.getSelectColor() : ssTextAreaSkin.getInactiveSelectColor());
                } else {
                    graphics.setPaint(ssTextAreaSkin.getColor());
                }

                graphics.drawGlyphVector(row.glyphVector, 0, rowY + ascent);
            }

            rowY += textBounds.getHeight();
        }
    }

    public void invalidate() {
        valid = false;
    }

    public void validate() {
        // TODO Validate from known invalid offset rather than 0, so we don't need to
        // recalculate all glyph vectors
        if (!valid) {
            rows = new ArrayList<Row>();
            width = 0;
            height = 0;

            // Re-layout glyphs and recalculate size
            Font font = ssTextAreaSkin.getFont();
            FontRenderContext fontRenderContext = Platform.getFontRenderContext();

            CharSequence characters = paragraph.getCharacters();
            int n = characters.length();

            int i = 0;
            int start = 0;
            float rowWidth = 0;
            int lastWhitespaceIndex = -1;

            // NOTE We use a character iterator here only because it is the most
            // efficient way to measure the character bounds (as of Java 6, the version
            // of Font#getStringBounds() that takes a String performs a string copy,
            // whereas the version that takes a character iterator does not)
            CharSequenceCharacterIterator ci = new CharSequenceCharacterIterator(characters);
            while (i < n) {
                char c = characters.charAt(i);
                if (Character.isWhitespace(c)) {
                    lastWhitespaceIndex = i;
                }

                Rectangle2D characterBounds = font.getStringBounds(ci, i, i + 1,
                fontRenderContext);
                rowWidth += characterBounds.getWidth();

                if (rowWidth > breakWidth) {
                    if (lastWhitespaceIndex == -1) {
                        if (start == i) {
                            appendLine(characters, start, start + 1, font, fontRenderContext);
                        } else {
                            appendLine(characters, start, i, font, fontRenderContext);
                            i--;
                        }
                    } else {
                        appendLine(characters, start, lastWhitespaceIndex + 1,
                        font, fontRenderContext);
                        i = lastWhitespaceIndex;
                    }

                    start = i + 1;

                    rowWidth = 0;
                    lastWhitespaceIndex = -1;
                }

                i++;
            }

            appendLine(characters, start, i, font, fontRenderContext);

            width = Math.max(width, PARAGRAPH_TERMINATOR_WIDTH);
        }

        valid = true;
    }

    private void appendLine(CharSequence characters, int start, int end,
                            Font font, FontRenderContext fontRenderContext) {
        CharSequenceCharacterIterator line = new CharSequenceCharacterIterator(characters,
        start, end, start);
        GlyphVector glyphVector = font.createGlyphVector(fontRenderContext, line);
        rows.add(new Row(glyphVector, start));

        Rectangle2D textBounds = glyphVector.getLogicalBounds();
        width = Math.max(width, (float) textBounds.getWidth());
        height += textBounds.getHeight();
    }

    public int getInsertionPoint(int x, int y) {
        Font font = ssTextAreaSkin.getFont();
        FontRenderContext fontRenderContext = Platform.getFontRenderContext();
        LineMetrics lm = font.getLineMetrics("", fontRenderContext);
        float rowHeight = lm.getAscent() + lm.getDescent();

        int i = (int) Math.floor(y / rowHeight);

        return getRowInsertionPoint(i, x);
    }

    public int getNextInsertionPoint(int x, int from, SSTextArea.ScrollDirection direction) {
        // Identify the row that contains the from index
        int n = rows.getLength();
        int i;
        if (from == -1) {
            i = (direction == SSTextArea.ScrollDirection.DOWN) ? -1 : n;
        } else {
            i = getRowAt(from);
        }

        // Move to the next or previous row
        if (direction == SSTextArea.ScrollDirection.DOWN) {
            i++;
        } else {
            i--;
        }

        return (i < 0
        || i >= n) ? -1 : getRowInsertionPoint(i, x);
    }

    private int getRowInsertionPoint(int rowIndex, float x) {
        Row row = rows.get(rowIndex);

        Rectangle2D glyphVectorBounds = row.glyphVector.getLogicalBounds();
        float rowWidth = (float) glyphVectorBounds.getWidth();

        int index;
        if (x < 0) {
            index = 0;
        } else if (x > rowWidth) {
            index = row.glyphVector.getNumGlyphs();

            // If this is not the last row, decrement the index so the insertion
            // point remains on this line
            if (rowIndex < rows.getLength() - 1) {
                index--;
            }
        } else {
            index = 0;
            int n = row.glyphVector.getNumGlyphs();

            while (index < n) {
                Shape glyphBounds = row.glyphVector.getGlyphLogicalBounds(index);
                Rectangle2D glyphBounds2D = glyphBounds.getBounds2D();

                if (glyphBounds2D.contains(x, glyphBounds2D.getY())) {
                    // Determine the bias; if the user clicks on the right half of the
                    // character; select the next character
                    if (x - glyphBounds2D.getX() > glyphBounds2D.getWidth() / 2
                    && index < n - 1) {
                        index++;
                    }

                    break;
                }

                index++;
            }
        }

        return index + row.offset;
    }

    public int getRowAt(int index) {
        int rowIndex = rows.getLength() - 1;
        Row row = rows.get(rowIndex);

        while (row.offset > index) {
            row = rows.get(--rowIndex);
        }

        return rowIndex;
    }

    public int getRowOffset(int index) {
        Row row = rows.get(getRowAt(index));
        return row.offset;
    }

    public int getRowLength(int index) {
        Row row = rows.get(getRowAt(index));
        return row.glyphVector.getNumGlyphs();
    }

    public int getRowCount() {
        return rows.getLength();
    }

    public Bounds getCharacterBounds(int index) {
        Bounds characterBounds = null;

        CharSequence characters = paragraph.getCharacters();
        final int characterCount = characters.length();

        int rowIndex, x, width;
        if (index == characterCount) {
            // This is the terminator character
            rowIndex = rows.getLength() - 1;
            Row row = rows.get(rowIndex);

            Rectangle2D glyphVectorBounds = row.glyphVector.getLogicalBounds();
            x = (int) Math.floor(glyphVectorBounds.getWidth());
            width = PARAGRAPH_TERMINATOR_WIDTH;
        } else {
            // This is a visible character
            rowIndex = getRowAt(index);
            Row row = rows.get(rowIndex);

            Shape glyphBounds = row.glyphVector.getGlyphLogicalBounds(index - row.offset);
            Rectangle2D glyphBounds2D = glyphBounds.getBounds2D();
            x = (int) Math.floor(glyphBounds2D.getX());
            width = (int) Math.ceil(glyphBounds2D.getWidth());
        }

        Font font = ssTextAreaSkin.getFont();
        FontRenderContext fontRenderContext = Platform.getFontRenderContext();
        LineMetrics lm = font.getLineMetrics("", fontRenderContext);
        final float rowHeight = lm.getAscent() + lm.getDescent();

        characterBounds = new Bounds(x, (int) Math.floor(rowIndex * rowHeight), width,
        (int) Math.ceil(rowHeight));

        return characterBounds;
    }

    @Override
    public void textInserted(SSTextArea.Paragraph paragraph, int index, int count) {
        invalidate();
        ssTextAreaSkin.invalidateComponent();
    }

    @Override
    public void textRemoved(SSTextArea.Paragraph paragraph, int index, int count) {
        invalidate();
        ssTextAreaSkin.invalidateComponent();
    }
}
