/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.ss;

import jp.pegasus.pivot.listener.SSCardPaneListener;
import org.apache.pivot.collections.Sequence;
import org.apache.pivot.util.ListenerList;
import org.apache.pivot.util.Vote;
import org.apache.pivot.wtk.Component;
import org.apache.pivot.wtk.Container;

public class SSCardPane extends Container {

    private static class SSCardPaneListenerList extends ListenerList<SSCardPaneListener>
    implements SSCardPaneListener {
        @Override
        public Vote previewSelectIndexChange(SSCardPane cardPane, int selectIndex) {
            Vote vote = Vote.APPROVE;

            for (SSCardPaneListener listener : this) {
                vote = vote.tally(listener.previewSelectIndexChange(cardPane, selectIndex));
            }

            return vote;
        }

        @Override
        public void selectIndexChangeVetoed(SSCardPane cardPane, Vote reason) {
            for (SSCardPaneListener listener : this) {
                listener.selectIndexChangeVetoed(cardPane, reason);
            }
        }

        @Override
        public void selectIndexChanged(SSCardPane cardPane, int previousSelectIndex) {
            for (SSCardPaneListener listener : this) {
                listener.selectIndexChanged(cardPane, previousSelectIndex);
            }
        }
    }

    private int selectIndex = -1;
    private SSCardPaneListenerList cardPaneListeners = new SSCardPaneListenerList();

    public SSCardPane() {
        installSkin(SSCardPane.class);
    }

    /**
     * Returns the currently select card index.
     *
     * @return The select card index, or <tt>-1</tt> if no card is select.
     */
    public int getSelectIndex() {
        return selectIndex;
    }

    /**
     * Sets the select card index.
     *
     * @param selectIndex The select card index, or <tt>-1</tt> for no selection.
     */
    public void setSelectIndex(int selectIndex) {
        if (selectIndex < -1
        || selectIndex > getLength() - 1) {
            throw new IndexOutOfBoundsException();
        }

        int previousSelectIndex = this.selectIndex;

        if (previousSelectIndex != selectIndex) {
            Vote vote = cardPaneListeners.previewSelectIndexChange(this, selectIndex);

            if (vote == Vote.APPROVE) {
                this.selectIndex = selectIndex;
                cardPaneListeners.selectIndexChanged(this, previousSelectIndex);
            } else {
                cardPaneListeners.selectIndexChangeVetoed(this, vote);
            }
        }
    }

    public Component getSelectCard() {
        return (selectIndex == -1) ? null : get(selectIndex);
    }

    @Override
    public void insert(Component component, int index) {
        // Update the selection
        int previousSelectIndex = selectIndex;

        if (selectIndex >= index) {
            selectIndex++;
        }

        // Insert the component
        super.insert(component, index);

        // Fire selection change event, if necessary
        if (selectIndex != previousSelectIndex) {
            cardPaneListeners.selectIndexChanged(this, selectIndex);
        }
    }

    @Override
    public Sequence<Component> remove(int index, int count) {
        // Update the selection
        int previousSelectIndex = selectIndex;

        if (selectIndex >= index) {
            if (selectIndex < index + count) {
                selectIndex = -1;
            } else {
                selectIndex -= count;
            }
        }

        // Remove the components
        Sequence<Component> removed = super.remove(index, count);

        // Fire selection change event, if necessary
        if (selectIndex != previousSelectIndex) {
            cardPaneListeners.selectIndexChanged(this, selectIndex);
        }

        return removed;
    }

    public ListenerList<SSCardPaneListener> getCardPaneListeners() {
        return cardPaneListeners;
    }
}
