/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.ss;

import jp.pegasus.pivot.listener.SSLabelBindingListener;
import jp.pegasus.pivot.listener.SSLabelListener;
import org.apache.pivot.json.JSON;
import org.apache.pivot.util.ListenerList;
import org.apache.pivot.wtk.BindType;
import org.apache.pivot.wtk.Component;

public class SSLabel extends Component {
    /**
     * Translates between text and context data during data binding.
     */
    public interface TextBindMapping {
        /**
         * Converts a value from the bind context to a text representation
         * during a {@link Component#load(Object)} operation.
         *
         * @param value
         */
        public String toString(Object value);

        /**
         * Converts a text string to a value to be stored in the bind context
         * during a {@link Component#store(Object)} operation.
         *
         * @param text
         */
        public Object valueOf(String text);
    }

    private static class SSLabelListenerList extends ListenerList<SSLabelListener>
    implements SSLabelListener {
        @Override
        public void textChanged(SSLabel label, String previousText) {
            for (SSLabelListener listener : this) {
                listener.textChanged(label, previousText);
            }
        }
    }

    private static class SSLabelBindingListenerList extends ListenerList<SSLabelBindingListener>
    implements SSLabelBindingListener {
        @Override
        public void textKeyChanged(SSLabel label, String previousTextKey) {
            for (SSLabelBindingListener listener : this) {
                listener.textKeyChanged(label, previousTextKey);
            }
        }

        @Override
        public void textBindTypeChanged(SSLabel label, BindType previousTextBindType) {
            for (SSLabelBindingListener listener : this) {
                listener.textBindTypeChanged(label, previousTextBindType);
            }
        }

        @Override
        public void textBindMappingChanged(SSLabel label, SSLabel.TextBindMapping previousTextBindMapping) {
            for (SSLabelBindingListener listener : this) {
                listener.textBindMappingChanged(label, previousTextBindMapping);
            }
        }
    }

    private String text = null;
    private String textKey = null;
    private BindType textBindType = BindType.BOTH;
    private TextBindMapping textBindMapping = null;

    private SSLabelListenerList labelListeners = new SSLabelListenerList();
    private SSLabelBindingListenerList labelBindingListeners = new SSLabelBindingListenerList();

    public SSLabel() {
        this(null);
    }

    public SSLabel(String text) {
        this.text = text;

        installSkin(SSLabel.class);
    }

    public String getText() {
        return text;
    }

    public void setText(String text) {
        String previousText = this.text;
        if (previousText != text) {
            this.text = text;
            labelListeners.textChanged(this, previousText);
        }
    }

    /**
     * Returns the label's text key.
     *
     * @return The text key, or <tt>null</tt> if no text key is set.
     */
    public String getTextKey() {
        return textKey;
    }

    /**
     * Sets the label's text key.
     *
     * @param textKey The text key, or <tt>null</tt> to clear the binding.
     */
    public void setTextKey(String textKey) {
        String previousTextKey = this.textKey;

        if (previousTextKey != textKey) {
            this.textKey = textKey;
            labelBindingListeners.textKeyChanged(this, previousTextKey);
        }
    }

    public BindType getTextBindType() {
        return textBindType;
    }

    public void setTextBindType(BindType textBindType) {
        if (textBindType == null) {
            throw new IllegalArgumentException();
        }

        BindType previousTextBindType = this.textBindType;

        if (previousTextBindType != textBindType) {
            this.textBindType = textBindType;
            labelBindingListeners.textBindTypeChanged(this, previousTextBindType);
        }
    }

    public TextBindMapping getTextBindMapping() {
        return textBindMapping;
    }

    public void setTextBindMapping(TextBindMapping textBindMapping) {
        TextBindMapping previousTextBindMapping = this.textBindMapping;

        if (previousTextBindMapping != textBindMapping) {
            this.textBindMapping = textBindMapping;
            labelBindingListeners.textBindMappingChanged(this, previousTextBindMapping);
        }
    }

    @Override
    public void load(Object context) {
        if (textKey != null
        && JSON.containsKey(context, textKey)
        && textBindType != BindType.STORE) {
            Object value = JSON.get(context, textKey);

            if (textBindMapping == null) {
                value = (value == null) ? null : value.toString();
            } else {
                value = textBindMapping.toString(value);
            }

            setText((String) value);
        }
    }

    @Override
    public void store(Object context) {
        if (textKey != null && textBindType != BindType.LOAD) {
            String text = getText();
            JSON.put(context, textKey, (textBindMapping == null) ?
            text : textBindMapping.valueOf(text));
        }
    }

    @Override
    public void clear() {
        if (textKey != null) {
            setText(null);
        }
    }

    public ListenerList<SSLabelListener> getLabelListeners() {
        return labelListeners;
    }

    public ListenerList<SSLabelBindingListener> getLabelBindingListeners() {
        return labelBindingListeners;
    }
}
