/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.ss;

import org.apache.pivot.collections.ArrayList;
import org.apache.pivot.collections.Sequence;
import org.apache.pivot.collections.immutable.ImmutableList;
import org.apache.pivot.wtk.Span;

import java.util.Comparator;

public class SSRangeSelect {
    // The coalesced select ranges
    private ArrayList<Span> selectRanges = new ArrayList<Span>();

    // Comparator that determines the index of the first intersecting range.
    private static final Comparator<Span> START_COMPARATOR = new Comparator<Span>() {
        @Override
        public int compare(Span range1, Span range2) {
            return (range1.end - range2.start);
        }
    };

    // Comparator that determines the index of the last intersecting range.
    private static final Comparator<Span> END_COMPARATOR = new Comparator<Span>() {
        @Override
        public int compare(Span range1, Span range2) {
            return (range1.start - range2.end);
        }
    };

    // Comparator that determines if two ranges intersect.
    private static final Comparator<Span> INTERSECTION_COMPARATOR = new Comparator<Span>() {
        @Override
        public int compare(Span range1, Span range2) {
            return (range1.start > range2.end) ? 1 : (range2.start > range1.end) ? -1 : 0;
        }
    };

    /**
     * Adds a range to the selection, merging and removing intersecting ranges
     * as needed.
     *
     * @param start
     * @param end
     * @return A sequence containing the ranges that were added.
     */
    public Sequence<Span> addRange(int start, int end) {
        ArrayList<Span> addedRanges = new ArrayList<Span>();

        Span range = normalize(start, end);
        assert (range.start >= 0);

        int n = selectRanges.getLength();

        if (n == 0) {
            // The selection is currently empty; append the new range
            // and add it to the added range list
            selectRanges.add(range);
            addedRanges.add(range);
        } else {
            // Locate the lower bound of the intersection
            int i = ArrayList.binarySearch(selectRanges, range, START_COMPARATOR);
            if (i < 0) {
                i = -(i + 1);
            }

            // Merge the selection with the previous range, if necessary
            if (i > 0) {
                Span previousRange = selectRanges.get(i - 1);
                if (range.start == previousRange.end + 1) {
                    i--;
                }
            }

            if (i == n) {
                // The new range starts after the last existing selection
                // ends; append it and add it to the added range list
                selectRanges.add(range);
                addedRanges.add(range);
            } else {
                // Locate the upper bound of the intersection
                int j = ArrayList.binarySearch(selectRanges, range, END_COMPARATOR);
                if (j < 0) {
                    j = -(j + 1);
                } else {
                    j++;
                }

                // Merge the selection with the next range, if necessary
                if (j < n) {
                    Span nextRange = selectRanges.get(j);
                    if (range.end == nextRange.start - 1) {
                        j++;
                    }
                }

                if (i == j) {
                    selectRanges.insert(range, i);
                    addedRanges.add(range);
                } else {
                    // Create a new range representing the union of the intersecting ranges
                    Span lowerRange = selectRanges.get(i);
                    Span upperRange = selectRanges.get(j - 1);

                    range = new Span(Math.min(range.start, lowerRange.start),
                    Math.max(range.end, upperRange.end));

                    // Add the gaps to the added list
                    if (range.start < lowerRange.start) {
                        addedRanges.add(new Span(range.start, lowerRange.start - 1));
                    }

                    for (int k = i; k < j - 1; k++) {
                        Span selectRange = selectRanges.get(k);
                        Span nextSelectRange = selectRanges.get(k + 1);
                        addedRanges.add(new Span(selectRange.end + 1, nextSelectRange.start - 1));
                    }

                    if (range.end > upperRange.end) {
                        addedRanges.add(new Span(upperRange.end + 1, range.end));
                    }

                    // Remove all redundant ranges
                    selectRanges.update(i, range);

                    if (i < j) {
                        selectRanges.remove(i + 1, j - i - 1);
                    }
                }
            }
        }

        return addedRanges;
    }

    /**
     * Removes a range from the selection, truncating and removing intersecting
     * ranges as needed.
     *
     * @param start
     * @param end
     * @return A sequence containing the ranges that were removed.
     */
    public Sequence<Span> removeRange(int start, int end) {
        ArrayList<Span> removedRanges = new ArrayList<Span>();

        Span range = normalize(start, end);
        assert (range.start >= 0);

        int n = selectRanges.getLength();

        if (n > 0) {
            // Locate the lower bound of the intersection
            int i = ArrayList.binarySearch(selectRanges, range, START_COMPARATOR);
            if (i < 0) {
                i = -(i + 1);
            }

            if (i < n) {
                Span lowerRange = selectRanges.get(i);

                if (lowerRange.start < range.start
                && lowerRange.end > range.end) {
                    // Removing the range will split the intersecting selection
                    // into two ranges
                    selectRanges.update(i, new Span(lowerRange.start, range.start - 1));
                    selectRanges.insert(new Span(range.end + 1, lowerRange.end), i + 1);
                    removedRanges.add(range);
                } else {
                    Span leadingRemovedRange = null;
                    if (range.start > lowerRange.start) {
                        // Remove the tail of this range
                        selectRanges.update(i, new Span(lowerRange.start, range.start - 1));
                        leadingRemovedRange = new Span(range.start, lowerRange.end);
                        i++;
                    }

                    // Locate the upper bound of the intersection
                    int j = ArrayList.binarySearch(selectRanges, range, END_COMPARATOR);
                    if (j < 0) {
                        j = -(j + 1);
                    } else {
                        j++;
                    }

                    if (j > 0) {
                        Span upperRange = selectRanges.get(j - 1);

                        Span trailingRemovedRange = null;
                        if (range.end < upperRange.end) {
                            // Remove the head of this range
                            selectRanges.update(j - 1, new Span(range.end + 1, upperRange.end));
                            trailingRemovedRange = new Span(upperRange.start, range.end);
                            j--;
                        }

                        // Remove all cleared ranges
                        Sequence<Span> clearedRanges = selectRanges.remove(i, j - i);

                        // Construct the removed range list
                        if (leadingRemovedRange != null) {
                            removedRanges.add(leadingRemovedRange);
                        }

                        for (int k = 0, c = clearedRanges.getLength(); k < c; k++) {
                            removedRanges.add(clearedRanges.get(k));
                        }

                        if (trailingRemovedRange != null) {
                            removedRanges.add(trailingRemovedRange);
                        }
                    }
                }
            }
        }

        return removedRanges;
    }

    /**
     * Clears the selection.
     */
    public void clear() {
        selectRanges.clear();
    }

    /**
     * Returns the range at a given index.
     *
     * @param index
     */
    public Span get(int index) {
        return selectRanges.get(index);
    }

    /**
     * Returns the number of ranges in the selection.
     */
    public int getLength() {
        return selectRanges.getLength();
    }

    /**
     * Returns an immutable wrapper around the select ranges.
     */
    public ImmutableList<Span> getSelectRanges() {
        return new ImmutableList<Span>(selectRanges);
    }

    /**
     * Determines the index of a range in the selection.
     *
     * @param range
     * @return The index of the range, if it exists in the selection; <tt>-1</tt>,
     *         otherwise.
     */
    public int indexOf(Span range) {
        assert (range != null);

        int index = -1;
        int i = ArrayList.binarySearch(selectRanges, range, INTERSECTION_COMPARATOR);

        if (i >= 0) {
            index = (range.equals(selectRanges.get(i))) ? i : -1;
        }

        return index;
    }

    /**
     * Tests for the presence of an index in the selection.
     *
     * @param index
     * @return <tt>true</tt> if the index is select; <tt>false</tt>, otherwise.
     */
    public boolean containsIndex(int index) {
        Span range = new Span(index);
        int i = ArrayList.binarySearch(selectRanges, range, INTERSECTION_COMPARATOR);

        return (i >= 0);
    }

    /**
     * Inserts an index into the span sequence (e.g. when items are inserted
     * into the model data).
     *
     * @param index
     * @return The number of ranges that were updated.
     */
    public int insertIndex(int index) {
        int updated = 0;

        // Get the insertion point for the range corresponding to the given index
        Span range = new Span(index);
        int i = ArrayList.binarySearch(selectRanges, range, INTERSECTION_COMPARATOR);

        if (i < 0) {
            // The inserted index does not intersect with a select range
            i = -(i + 1);
        } else {
            // The inserted index intersects with a currently select range
            Span selectRange = selectRanges.get(i);

            // If the inserted index falls within the current range, increment
            // the endpoint only
            if (selectRange.start < index) {
                selectRanges.update(i, new Span(selectRange.start, selectRange.end + 1));

                // Start incrementing range bounds beginning at the next range
                i++;
            }
        }

        // Increment any subsequent selection indexes
        int n = selectRanges.getLength();
        while (i < n) {
            Span selectRange = selectRanges.get(i);
            selectRanges.update(i, new Span(selectRange.start + 1, selectRange.end + 1));
            updated++;
            i++;
        }

        return updated;
    }

    /**
     * Removes a range of indexes from the span sequence (e.g. when items
     * are removed from the model data).
     *
     * @param index
     * @param count
     * @return The number of ranges that were updated.
     */
    public int removeIndexes(int index, int count) {
        // Clear any selections in the given range
        Sequence<Span> removed = removeRange(index, (index + count) - 1);
        int updated = removed.getLength();

        // Decrement any subsequent selection indexes
        Span range = new Span(index);
        int i = ArrayList.binarySearch(selectRanges, range, INTERSECTION_COMPARATOR);
        assert (i < 0) : "i should be negative, since index should no longer be select";

        i = -(i + 1);

        // Determine the number of ranges to modify
        int n = selectRanges.getLength();
        while (i < n) {
            Span selectRange = selectRanges.get(i);
            selectRanges.update(i, new Span(selectRange.start - count, selectRange.end - count));
            updated++;
            i++;
        }

        return updated;
    }

    /**
     * Ensures that the start value is less than or equal to the end value.
     *
     * @param start
     * @param end
     * @return A span containing the normalized range.
     */
    public static Span normalize(int start, int end) {
        return new Span(Math.min(start, end), Math.max(start, end));
    }
}
