/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.ss;

import jp.pegasus.pivot.content.SSButtonDataRenderer;
import jp.pegasus.pivot.editor.SSTabPaneEditor;
import jp.pegasus.pivot.listener.tabpane.SSTabPaneAttributeListener;
import jp.pegasus.pivot.listener.tabpane.SSTabPaneListener;
import jp.pegasus.pivot.listener.tabpane.SSTabPaneSelectionListener;
import org.apache.pivot.beans.DefaultProperty;
import org.apache.pivot.collections.ArrayList;
import org.apache.pivot.collections.Sequence;
import org.apache.pivot.util.ImmutableIterator;
import org.apache.pivot.util.ListenerList;
import org.apache.pivot.util.Vote;
import org.apache.pivot.wtk.Button;
import org.apache.pivot.wtk.Component;
import org.apache.pivot.wtk.Container;
import org.apache.pivot.wtk.Point;

import java.util.Iterator;

@DefaultProperty("tabs")
public class SSTabPane extends Container {

    /**
     * Tree view node editor interface.
     */
    public interface TabEditor {
        /**
         * Called to begin editing a tab.
         *
         * @param tabPane
         */
        public void beginEdit(SSTabPane tabPane);

        /**
         * Terminates an edit operation.
         *
         * @param result <tt>true</tt> to perform the edit; <tt>false</tt> to cancel it.
         */
        public void endEdit(boolean result);

        /**
         * Tests whether an edit is currently in progress.
         */
        public boolean isEditing();
    }

    /**
     * Tab sequence implementation.
     */
    public final class TabSequence implements Sequence<Component>, Iterable<Component> {
        private TabSequence() {
        }

        @Override
        public int add(Component tab) {
            int index = getLength();
            insert(tab, index);

            return index;
        }

        @Override
        public void insert(Component tab, final int index) {
            if (tab == null) {
                throw new IllegalArgumentException("tab is null.");
            }

            // Add the tab to the component sequence
            SSTabPane.this.add(tab);
            tabs.insert(tab, index);

            // Update the selection
            final int previousSelectedIndex = selectedIndex;

            if (selectedIndex >= index) {
                selectedIndex++;
            }

            // Fire insert event
            tabPaneListeners.tabInserted(SSTabPane.this, index);

            // Fire selection change event, if necessary
            if (selectedIndex != previousSelectedIndex) {
                tabPaneSelectionListeners.selectedIndexChanged(SSTabPane.this, selectedIndex);
            }
        }

        @Override
        public Component update(final int index, Component tab) {
            throw new UnsupportedOperationException();
        }

        @Override
        public int remove(Component tab) {
            final int index = indexOf(tab);
            if (index != -1) {
                remove(index, 1);
            }

            return index;
        }

        @Override
        public Sequence<Component> remove(final int index, final int count) {
            Sequence<Component> removed;

            Vote vote = tabPaneListeners.previewRemoveTabs(SSTabPane.this, index, count);
            if (vote == Vote.APPROVE) {
                // Remove the tabs from the tab list
                removed = tabs.remove(index, count);

                // Update the selection
                final int previousSelectedIndex = selectedIndex;

                if (selectedIndex >= index) {
                    if (selectedIndex < index + count) {
                        selectedIndex = -1;
                    } else {
                        selectedIndex -= count;
                    }
                }

                // Fire remove event
                tabPaneListeners.tabsRemoved(SSTabPane.this, index, removed);

                // Remove the tabs from the component list
                for (int i = 0, n = removed.getLength(); i < n; i++) {
                    Component tab = removed.get(i);
                    SSTabPane.this.remove(tab);
                }

                // Fire selection change event, if necessary
                if (selectedIndex != previousSelectedIndex) {
                    tabPaneSelectionListeners.selectedIndexChanged(SSTabPane.this, selectedIndex);
                }
            } else {
                removed = null;
                tabPaneListeners.removeTabsVetoed(SSTabPane.this, vote);
            }

            return removed;
        }

        @Override
        public Component get(int index) {
            return tabs.get(index);
        }

        @Override
        public int indexOf(Component tab) {
            return tabs.indexOf(tab);
        }

        @Override
        public int getLength() {
            return tabs.getLength();
        }

        @Override
        public Iterator<Component> iterator() {
            return new ImmutableIterator<Component>(tabs.iterator());
        }
    }

    private enum Attribute {
        TAB_DATA,
        TOOLTIP_TEXT;
    }

    private static class SSTabPaneListenerList extends ListenerList<SSTabPaneListener>
    implements SSTabPaneListener {
        @Override
        public void tabInserted(SSTabPane tabPane, int index) {
            for (SSTabPaneListener listener : this) {
                listener.tabInserted(tabPane, index);
            }
        }

        @Override
        public Vote previewRemoveTabs(SSTabPane tabPane, int index, int count) {
            Vote vote = Vote.APPROVE;

            for (SSTabPaneListener listener : this) {
                vote = vote.tally(listener.previewRemoveTabs(tabPane, index, count));
            }

            return vote;
        }

        @Override
        public void tabsRemoved(SSTabPane tabPane, int index, Sequence<Component> tabs) {
            for (SSTabPaneListener listener : this) {
                listener.tabsRemoved(tabPane, index, tabs);
            }
        }

        @Override
        public void removeTabsVetoed(SSTabPane tabPane, Vote reason) {
            for (SSTabPaneListener listener : this) {
                listener.removeTabsVetoed(tabPane, reason);
            }
        }

        @Override
        public void cornerChanged(SSTabPane tabPane, Component previousCorner) {
            for (SSTabPaneListener listener : this) {
                listener.cornerChanged(tabPane, previousCorner);
            }
        }

        @Override
        public void tabDataRendererChanged(SSTabPane tabPane, Button.DataRenderer previousTabDataRenderer) {
            for (SSTabPaneListener listener : this) {
                listener.tabDataRendererChanged(tabPane, previousTabDataRenderer);
            }
        }

        @Override
        public void closeableChanged(SSTabPane tabPane) {
            for (SSTabPaneListener listener : this) {
                listener.closeableChanged(tabPane);
            }
        }

        @Override
        public void collapsibleChanged(SSTabPane tabPane) {
            for (SSTabPaneListener listener : this) {
                listener.collapsibleChanged(tabPane);
            }
        }
    }

    private static class SSTabPaneSelectionListenerList extends ListenerList<SSTabPaneSelectionListener>
    implements SSTabPaneSelectionListener {
        @Override
        public Vote previewSelectedIndexChange(SSTabPane tabPane, int selectedIndex) {
            Vote vote = Vote.APPROVE;

            for (SSTabPaneSelectionListener listener : this) {
                vote = vote.tally(listener.previewSelectedIndexChange(tabPane, selectedIndex));
            }

            return vote;
        }

        @Override
        public void selectedIndexChangeVetoed(SSTabPane tabPane, Vote reason) {
            for (SSTabPaneSelectionListener listener : this) {
                listener.selectedIndexChangeVetoed(tabPane, reason);
            }
        }

        @Override
        public void selectedIndexChanged(SSTabPane tabPane, int previousSelectedIndex) {
            for (SSTabPaneSelectionListener listener : this) {
                listener.selectedIndexChanged(tabPane, previousSelectedIndex);
            }
        }
    }

    private static class SSTabPaneAttributeListenerList extends ListenerList<SSTabPaneAttributeListener>
    implements SSTabPaneAttributeListener {
        @Override
        public void tabDataChanged(SSTabPane tabPane, Component component, Object previousTabData) {
            for (SSTabPaneAttributeListener listener : this) {
                listener.tabDataChanged(tabPane, component, previousTabData);
            }
        }

        @Override
        public void tooltipTextChanged(SSTabPane tabPane, Component component, String previousTooltipText) {
            for (SSTabPaneAttributeListener listener : this) {
                listener.tooltipTextChanged(tabPane, component, previousTooltipText);
            }
        }
    }

    private Point selectedTabButtonLocation = null;
    private TabEditor tabEditor = new SSTabPaneEditor();
    private ArrayList<Component> tabs = new ArrayList<Component>();
    private TabSequence tabSequence = new TabSequence();
    private Component corner = null;
    private int selectedIndex = -1;
    private Button.DataRenderer tabDataRenderer = DEFAULT_TAB_DATA_RENDERER;
    private boolean closeable = false;
    private boolean collapsible = false;

    private SSTabPaneListenerList tabPaneListeners = new SSTabPaneListenerList();
    private SSTabPaneSelectionListenerList tabPaneSelectionListeners = new SSTabPaneSelectionListenerList();
    private SSTabPaneAttributeListenerList tabPaneAttributeListeners = new SSTabPaneAttributeListenerList();

    private static final Button.DataRenderer DEFAULT_TAB_DATA_RENDERER = new SSButtonDataRenderer();

    public SSTabPane() {
        super();
        installSkin(SSTabPane.class);
    }

    public TabSequence getTabs() {
        return tabSequence;
    }

    public Component getCorner() {
        return corner;
    }

    public void setCorner(Component corner) {
        Component previousCorner = this.corner;

        if (previousCorner != corner) {
            // Remove any previous corner component
            this.corner = null;

            if (previousCorner != null) {
                remove(previousCorner);
            }

            // Set the new corner component
            if (corner != null) {
                insert(corner, 0);
            }

            this.corner = corner;

            tabPaneListeners.cornerChanged(this, previousCorner);
        }
    }

    public int getSelectedIndex() {
        return selectedIndex;
    }

    public void setSelectedIndex(int selectedIndex) {
        if (selectedIndex < -1 || selectedIndex > tabs.getLength() - 1) {
            throw new IndexOutOfBoundsException();
        }

        int previousSelectedIndex = this.selectedIndex;

        if (previousSelectedIndex != selectedIndex) {
            Vote vote = tabPaneSelectionListeners.previewSelectedIndexChange(this, selectedIndex);

            if (vote == Vote.APPROVE) {
                this.selectedIndex = selectedIndex;
                tabPaneSelectionListeners.selectedIndexChanged(this, previousSelectedIndex);
            } else {
                tabPaneSelectionListeners.selectedIndexChangeVetoed(this, vote);
            }
        }
    }

    public Component getSelectedTab() {
        return (selectedIndex == -1) ? null : tabs.get(selectedIndex);
    }

    public Button.DataRenderer getTabDataRenderer() {
        return tabDataRenderer;
    }

    public void setTabDataRenderer(Button.DataRenderer tabDataRenderer) {
        if (tabDataRenderer == null) {
            throw new IllegalArgumentException();
        }

        Button.DataRenderer previousTabDataRenderer = this.tabDataRenderer;
        if (previousTabDataRenderer != tabDataRenderer) {
            this.tabDataRenderer = tabDataRenderer;
            tabPaneListeners.tabDataRendererChanged(this, previousTabDataRenderer);
        }
    }

    public boolean isCloseable() {
        return closeable;
    }

    public void setCloseable(boolean closeable) {
        if (this.closeable != closeable) {
            this.closeable = closeable;
            tabPaneListeners.closeableChanged(this);
        }
    }

    /**
     * @return <tt>true</tt> if the SSTabPane is collapsible and no tab is
     *         selected; <tt>false</tt>, otherwise.
     * @see #isCollapsible()
     * @see #getSelectedIndex()
     */
    public boolean isCollapsed() {
        return collapsible && (selectedIndex == -1);
    }

    /**
     * Collapse or expand the SSTabPane (if it is collapsible).
     *
     * @param collapsed <tt>true</tt> to collapse, <tt>false</tt> to expand and
     *                  select the first tab. Use {@link #setSelectedIndex(int)} to expand and
     *                  select a specific Tab.
     * @see #isCollapsible()
     * @see #setSelectedIndex(int)
     */
    public void setCollapsed(final boolean collapsed) {
        if (collapsible && (isCollapsed() != collapsed)) {
            int index = (collapsed || tabs.getLength() == 0) ? -1 : 0;
            setSelectedIndex(index);
        }
    }

    public boolean isCollapsible() {
        return collapsible;
    }

    public void setCollapsible(final boolean collapsible) {
        if (this.collapsible != collapsible) {
            this.collapsible = collapsible;
            tabPaneListeners.collapsibleChanged(this);
        }
    }

    @Override
    public Sequence<Component> remove(int index, int count) {
        for (int i = index, n = index + count; i < n; i++) {
            Component component = get(i);

            if (component == corner
            || tabs.indexOf(component) >= 0) {
                throw new UnsupportedOperationException();
            }
        }

        // Call the base method to remove the components
        return super.remove(index, count);
    }

    public ListenerList<SSTabPaneListener> getTabPaneListeners() {
        return tabPaneListeners;
    }

    public ListenerList<SSTabPaneSelectionListener> getTabPaneSelectionListeners() {
        return tabPaneSelectionListeners;
    }

    public ListenerList<SSTabPaneAttributeListener> getTabPaneAttributeListeners() {
        return tabPaneAttributeListeners;
    }

    public static Object getTabData(Component component) {
        return component.getAttribute(Attribute.TAB_DATA);
    }

    public static void setTabData(Component component, Object tabData) {
        Object previousTabData = component.setAttribute(Attribute.TAB_DATA, tabData);

        if (previousTabData != tabData) {
            Container parent = component.getParent();

            if (parent instanceof SSTabPane) {
                SSTabPane tabPane = (SSTabPane) parent;
                tabPane.tabPaneAttributeListeners.tabDataChanged(tabPane, component,
                previousTabData);
            }
        }
    }

    public static String getTooltipText(Component component) {
        return (String) component.getAttribute(Attribute.TOOLTIP_TEXT);
    }

    public static void setTooltipText(Component component, String tooltipText) {
        String previousTooltipText = (String) component.setAttribute(Attribute.TOOLTIP_TEXT, tooltipText);

        if (previousTooltipText != tooltipText) {
            Container parent = component.getParent();

            if (parent instanceof SSTabPane) {
                SSTabPane tabPane = (SSTabPane) parent;
                tabPane.tabPaneAttributeListeners.tooltipTextChanged(tabPane, component, previousTooltipText);
            }
        }
    }

    public TabEditor getTabEditor() {
        return tabEditor;
    }

    public void setTabEditor(TabEditor tabEditor) {
        TabEditor previousTabEditor = this.tabEditor;

        if (previousTabEditor != tabEditor) {
            this.tabEditor = tabEditor;
//            tabPaneListeners.tabEditorChanged(this, previousTabEditor);
        }
        this.tabEditor = tabEditor;
    }

    public Point getSelectedTabButtonLocation() {
        return selectedTabButtonLocation;
    }

    public void setSelectedTabButtonLocation(Point selectedTabButtonLocation) {
        this.selectedTabButtonLocation = selectedTabButtonLocation;
    }
}
