/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.ss;

import jp.pegasus.pivot.listener.textarea.SSTextAreaBindingListener;
import jp.pegasus.pivot.listener.textarea.SSTextAreaContentListener;
import jp.pegasus.pivot.listener.textarea.SSTextAreaListener;
import jp.pegasus.pivot.listener.textarea.SSTextAreaSelectListener;
import jp.pegasus.pivot.model.SSElementalData;
import org.apache.pivot.collections.ArrayList;
import org.apache.pivot.collections.HashMap;
import org.apache.pivot.collections.LinkedList;
import org.apache.pivot.collections.Sequence;
import org.apache.pivot.json.JSON;
import org.apache.pivot.util.ImmutableIterator;
import org.apache.pivot.util.ListenerList;
import org.apache.pivot.wtk.*;
import org.apache.pivot.wtk.Component;

import java.awt.*;
import java.io.*;
import java.net.URL;
import java.util.Iterator;

public class SSTextArea extends Component {

    /**
     * Text Area Type
     */
    public enum TextAreaType {
        Cell, Sheet,
    }

    /**
     * Class representing a paragraph of text.
     */
    public static final class Paragraph {
        private static class ParagraphListenerList extends ListenerList<ParagraphListener>
        implements ParagraphListener {
            @Override
            public void textInserted(Paragraph paragraph, int index, int count) {
                for (ParagraphListener listener : this) {
                    listener.textInserted(paragraph, index, count);
                }
            }

            @Override
            public void textRemoved(Paragraph paragraph, int index, int count) {
                for (ParagraphListener listener : this) {
                    listener.textRemoved(paragraph, index, count);
                }
            }
        }

        private StringBuilder characters = new StringBuilder(INITIAL_PARAGRAPH_CAPACITY);
        private SSTextArea textArea = null;
        private int offset = -1;

        private ParagraphListenerList paragraphListeners = new ParagraphListenerList();

        public CharSequence getCharacters() {
            return characters;
        }

        public SSTextArea getTextArea() {
            return textArea;
        }

        public void append(char character) {
            if (textArea != null) {
                throw new IllegalStateException();
            }
            characters.append(character);
        }

        public void clear() {
            if (textArea != null) {
                throw new IllegalStateException();
            }
            characters.delete(0, characters.length());
        }

        public void insertText(CharSequence text, int index) {
            if (text == null) {
                throw new IllegalArgumentException();
            }

            if (index < 0 || index > characters.length()) {
                throw new IndexOutOfBoundsException();
            }

            int count = text.length();

            if (textArea != null && textArea.characterCount + count > textArea.maximumLength) {
                throw new IllegalArgumentException("Insertion of text would exceed maximum length.");
            }

            characters.insert(index, text);

            if (textArea != null) {
                // Update offsets and character count
                textArea.updateParagraphOffsets(textArea.paragraphs.indexOf(this) + 1, count);
                textArea.characterCount += count;

                // Update select state
                int previousSelectStart = textArea.selectStart;
                int previousSelectLength = textArea.selectLength;
                textArea.selectStart = offset + index + count;
                textArea.selectLength = 0;

                // Fire change events
                paragraphListeners.textInserted(this, index, count);
                textArea.textAreaContentListeners.textChanged(textArea);

                if (textArea.selectStart != previousSelectStart || textArea.selectLength != previousSelectLength) {
                    textArea.textAreaSelectListeners.selectChanged(textArea,
                    textArea.selectStart, textArea.selectLength);
                }
            }
        }

        public void removeText(int index) {
            removeText(index, characters.length() - index);
        }

        public void removeText(int index, int count) {
            if (index < 0 || index + count > characters.length()) {
                throw new IndexOutOfBoundsException();
            }

            characters.delete(index, index + count);

            if (textArea != null) {
                // Update offsets and character count
                textArea.updateParagraphOffsets(textArea.paragraphs.indexOf(this) + 1, -count);
                textArea.characterCount -= count;

                // Update select state
                final int previousSelectStart = textArea.selectStart;
                final int previousSelectLength = textArea.selectLength;
                textArea.selectStart = offset + index;
                textArea.selectLength = 0;

                // Fire change events
                paragraphListeners.textRemoved(this, index, count);
                textArea.textAreaContentListeners.textChanged(textArea);

                if (textArea.selectStart != previousSelectStart || textArea.selectLength != previousSelectLength) {
                    textArea.textAreaSelectListeners.selectChanged(textArea,
                    textArea.selectStart, textArea.selectLength);
                }
            }
        }

        public int getOffset() {
            return offset;
        }

        public ListenerList<ParagraphListener> getParagraphListeners() {
            return paragraphListeners;
        }
    }

    /**
     * Paragraph listener interface.
     */
    public interface ParagraphListener {
        /**
         * Paragraph listener interface adapter.
         */
        public static class Adapter implements ParagraphListener {
            @Override
            public void textInserted(Paragraph paragraph, int index, int count) {
            }

            @Override
            public void textRemoved(Paragraph paragraph, int index, int count) {
            }
        }

        /**
         * Called when text has been inserted into a paragraph.
         *
         * @param paragraph The source of the event.
         * @param index     The index at which the text was inserted.
         * @param count     The number of characters that were inserted.
         */
        public void textInserted(Paragraph paragraph, int index, int count);

        /**
         * Called when characters have been removed from a paragraph.
         *
         * @param paragraph The source of the event.
         * @param index     The index from which the text was removed.
         * @param count     The number of characters that were removed.
         */
        public void textRemoved(Paragraph paragraph, int index, int count);
    }

    /**
     * Enum representing a scroll direction.
     */
    public enum ScrollDirection {
        UP,
        DOWN
    }

    /**
     * Text area skin interface. Text area skins are required to implement
     * this.
     */
    public interface Skin {
        /**
         * Returns the insertion point for a given location.
         *
         * @param x
         * @param y
         */
        public int getInsertionPoint(int x, int y);

        /**
         * Returns the next insertion point given an x coordinate and a
         * character index.
         *
         * @param x
         * @param from
         * @param direction
         */
        public int getNextInsertionPoint(int x, int from, ScrollDirection direction);

        /**
         * Returns the row index of the character at a given index.
         *
         * @param index
         */
        public int getRowAt(int index);

        /**
         * Returns the index of the first character in the row containing
         * a given character index.
         *
         * @param index
         */
        public int getRowOffset(int index);

        /**
         * Returns the number of characters in the row containing a given
         * character index.
         *
         * @param index
         */
        public int getRowLength(int index);

        /**
         * Returns the total number of rows in the text area.
         */
        public int getRowCount();

        /**
         * Returns the bounds of the character at a given index.
         *
         * @param index
         */
        public Bounds getCharacterBounds(int index);
    }

    /**
     * Translates between text and context data during data binding.
     */
    public interface TextBindMapping {
        /**
         * Converts a value from the bind context to a text representation during a
         * {@link Component#load(Object)} operation.
         *
         * @param value
         */
        public String toString(Object value);

        /**
         * Converts a text string to a value to be stored in the bind context during a
         * {@link Component#store(Object)} operation.
         *
         * @param text
         */
        public Object valueOf(String text);
    }

    /**
     * Text area paragraph sequence.
     */
    public final class ParagraphSequence implements Sequence<Paragraph>, Iterable<Paragraph> {
        public int add(Paragraph paragraph) {
            int index = getLength();
            insert(paragraph, index);

            return index;
        }

        public void insert(Paragraph paragraph, int index) {
            if (paragraph == null) {
                throw new IllegalArgumentException("paragraph is null.");
            }

            if (paragraph.textArea != null) {
                throw new IllegalArgumentException("paragraph is already in use by another text area.");
            }

            // Determine insertion count, including terminator character
            int characterCount = paragraph.characters.length();

            if (getLength() > 0) {
                characterCount++;
            }

            if (SSTextArea.this.characterCount + characterCount > maximumLength) {
                throw new IllegalArgumentException("Insertion of text would exceed maximum length.");
            }

            // Set the paragraph offset
            if (index == paragraphs.getLength()) {
                paragraph.offset = SSTextArea.this.characterCount;

                // Include terminator character
                if (index > 0) {
                    paragraph.offset++;
                }
            } else {
                paragraph.offset = paragraphs.get(index).offset;
            }

            // Insert the paragraph
            paragraphs.insert(paragraph, index);
            paragraph.textArea = SSTextArea.this;

            // Update offsets and character count
            updateParagraphOffsets(index + 1, characterCount);
            SSTextArea.this.characterCount += characterCount;

            // Update select state
            final int previousSelectStart = selectStart;
            final int previousSelectLength = selectLength;
            selectStart = paragraph.offset + paragraph.characters.length();
            selectLength = 0;

            // Fire change events
            textAreaContentListeners.paragraphInserted(SSTextArea.this, index);
            textAreaContentListeners.textChanged(SSTextArea.this);

            if (selectStart != previousSelectStart || selectLength != previousSelectLength) {
                textAreaSelectListeners.selectChanged(SSTextArea.this, selectStart, selectLength);
            }
        }

        public Paragraph update(int index, Paragraph paragraph) {
            throw new UnsupportedOperationException();
        }

        public int remove(Paragraph paragraph) {
            int index = indexOf(paragraph);
            if (index != -1) {
                remove(index, 1);
            }

            return index;
        }

        public Sequence<Paragraph> remove(int index, int count) {
            Sequence<Paragraph> removed = paragraphs.remove(index, count);

            if (count > 0) {
                int characterCount = 0;
                for (int i = 0, n = removed.getLength(); i < n; i++) {
                    Paragraph paragraph = removed.get(i);
                    paragraph.textArea = null;
                    paragraph.offset = -1;
                    characterCount += paragraph.characters.length() + 1;
                }

                // Don't include the implicit final terminator in the character count
                if (getLength() == 0) {
                    characterCount--;
                }

                // Update offsets
                updateParagraphOffsets(index, -characterCount);
                SSTextArea.this.characterCount -= characterCount;

                // Update select state
                final int previousSelectStart = selectStart;
                final int previousSelectLength = selectLength;
                selectStart = (index == paragraphs.getLength()) ?
                SSTextArea.this.characterCount : paragraphs.get(index).offset;
                selectLength = 0;

                // Fire change events
                textAreaContentListeners.paragraphsRemoved(SSTextArea.this, index, removed);
                textAreaContentListeners.textChanged(SSTextArea.this);

                if (selectStart != previousSelectStart || selectLength != previousSelectLength) {
                    textAreaSelectListeners.selectChanged(SSTextArea.this, selectStart, selectLength);
                }
            }

            return removed;
        }

        public Paragraph get(int index) {
            return paragraphs.get(index);
        }

        public int indexOf(Paragraph paragraph) {
            return paragraphs.indexOf(paragraph);
        }

        public int getLength() {
            return paragraphs.getLength();
        }

        public Iterator<Paragraph> iterator() {
            return new ImmutableIterator<Paragraph>(paragraphs.iterator());
        }
    }

    private interface Edit {
        public void undo();
    }

    private class InsertTextEdit implements Edit {
        private final int index;
        private final int count;

        public InsertTextEdit(CharSequence text, int index) {
            this.index = index;
            count = text.length();
        }

        public void undo() {
            removeText(index, count, false);
        }
    }

    private class RemoveTextEdit implements Edit {
        private final int index;
        private final CharSequence text;

        public RemoveTextEdit(int index, int count) {
            this.index = index;
            text = getText(index, index + count);
        }

        public void undo() {
            insertText(text, index, false);
        }
    }

    private static class SSTextAreaListenerList extends ListenerList<SSTextAreaListener>
    implements SSTextAreaListener {
        @Override
        public void maximumLengthChanged(SSTextArea textArea, int previousMaximumLength) {
            for (SSTextAreaListener listener : this) {
                listener.maximumLengthChanged(textArea, previousMaximumLength);
            }
        }

        @Override
        public void editableChanged(SSTextArea textArea) {
            for (SSTextAreaListener listener : this) {
                listener.editableChanged(textArea);
            }
        }
    }

    private static class SSTextAreaContentListenerList extends ListenerList<SSTextAreaContentListener>
    implements SSTextAreaContentListener {
        @Override
        public void paragraphInserted(SSTextArea textArea, int index) {
            for (SSTextAreaContentListener listener : this) {
                listener.paragraphInserted(textArea, index);
            }
        }

        @Override
        public void paragraphsRemoved(SSTextArea ssTextArea, int index, Sequence<SSTextArea.Paragraph> removed) {
            for (SSTextAreaContentListener listener : this) {
                listener.paragraphsRemoved(ssTextArea, index, removed);
            }
        }

        @Override
        public void textChanged(SSTextArea ssTextArea) {
            for (SSTextAreaContentListener listener : this) {
                listener.textChanged(ssTextArea);
            }
        }
    }

    private static class SSTextAreaSelectListenerList extends ListenerList<SSTextAreaSelectListener>
    implements SSTextAreaSelectListener {
        @Override
        public void selectChanged(SSTextArea textArea, int previousSelectStart,
                                  int previousSelectLength) {
            for (SSTextAreaSelectListener listener : this) {
                listener.selectChanged(textArea, previousSelectStart,
                previousSelectLength);
            }
        }
    }

    private static class SSTextAreaBindingListenerList extends ListenerList<SSTextAreaBindingListener>
    implements SSTextAreaBindingListener {
        @Override
        public void textKeyChanged(SSTextArea ssTextArea, String previousTextKey) {
            for (SSTextAreaBindingListener listener : this) {
                listener.textKeyChanged(ssTextArea, previousTextKey);
            }
        }

        @Override
        public void textBindTypeChanged(SSTextArea ssTextArea, BindType previousTextBindType) {
            for (SSTextAreaBindingListener listener : this) {
                listener.textBindTypeChanged(ssTextArea, previousTextBindType);
            }
        }

        @Override
        public void textBindMappingChanged(SSTextArea ssTextArea, TextBindMapping previousTextBindMapping) {
            for (SSTextAreaBindingListener listener : this) {
                listener.textBindMappingChanged(ssTextArea, previousTextBindMapping);
            }
        }
    }

    private TextAreaType textAreaType = null;

    private ArrayList<Paragraph> paragraphs = new ArrayList<Paragraph>();
    private ParagraphSequence paragraphSequence = new ParagraphSequence();

    private int characterCount = 0;

    private int selectStart = 0;
    private int selectLength = 0;

    private int maximumLength = Integer.MAX_VALUE;
    private boolean editable = true;

    private String textKey = null;
    private BindType textBindType = BindType.BOTH;
    private TextBindMapping textBindMapping = null;

    private LinkedList<Edit> editHistory = new LinkedList<Edit>();

    private SSTextAreaListenerList textAreaListeners = new SSTextAreaListenerList();
    private SSTextAreaContentListenerList textAreaContentListeners = new SSTextAreaContentListenerList();
    private SSTextAreaSelectListenerList textAreaSelectListeners = new SSTextAreaSelectListenerList();
    private SSTextAreaBindingListenerList textAreaBindingListeners = new SSTextAreaBindingListenerList();

    private static final int INITIAL_PARAGRAPH_CAPACITY = 256;
    private static final int MAXIMUM_EDIT_HISTORY_LENGTH = 30;

    public SSTextArea() {
        installSkin(SSTextArea.class);
        setText("");
    }

    @Override
    protected void setSkin(org.apache.pivot.wtk.Skin skin) {
        if (!(skin instanceof SSTextArea.Skin)) {
            throw new IllegalArgumentException("Skin class must implement " + SSTextArea.Skin.class.getName());
        }
        super.setSkin(skin);
    }

    /**
     * Returns the text content of the text area.
     *
     * @return A string containing a copy of the text area's text content.
     */
    public String getText() {
        return getText(0, getCharacterCount());
    }

    /**
     * Returns a portion of the text content of the text area.
     *
     * @param beginIndex
     * @param endIndex
     * @return A string containing a copy of the text area's text content.
     */
    public String getText(int beginIndex, int endIndex) {
        if (beginIndex > endIndex) {
            throw new IllegalArgumentException();
        }

        if (beginIndex < 0 || endIndex > characterCount) {
            throw new IndexOutOfBoundsException();
        }

        final int count = endIndex - beginIndex;
        StringBuilder textBuilder = new StringBuilder(count);

        // Get paragraph and character offset at beginIndex
        int paragraphIndex = getParagraphAt(beginIndex);
        Paragraph paragraph = paragraphs.get(paragraphIndex);

        int characterOffset = beginIndex - paragraph.offset;

        // Read characters until endIndex is reached, appending to text builder
        // and moving to next paragraph as needed
        int i = 0;
        while (i < count) {
            if (characterOffset == paragraph.characters.length() && i < characterCount) {
                textBuilder.append('\n');
                paragraph = paragraphs.get(++paragraphIndex);
                characterOffset = 0;
            } else {
                textBuilder.append(paragraph.characters.charAt(characterOffset++));
            }

            i++;
        }

        return textBuilder.toString();
    }

    /**
     * Sets the text content of the text area.
     *
     * @param text
     */
    public void setText(String text) {
        if (text == null) {
            throw new IllegalArgumentException();
        }

        try {
            setText(new StringReader(text));
        } catch (IOException exception) {
            throw new RuntimeException(exception);
        }
    }

    public void setText(URL textURL) throws IOException {
        if (textURL == null) {
            throw new IllegalArgumentException();
        }

        InputStream inputStream = null;
        try {
            inputStream = textURL.openStream();
            setText(new InputStreamReader(inputStream));
        } finally {
            if (inputStream != null) {
                inputStream.close();
            }
        }
    }

    public void setText(Reader textReader) throws IOException {
        if (textReader == null) {
            throw new IllegalArgumentException();
        }

        // Construct the paragraph list
        ArrayList<Paragraph> paragraphs = new ArrayList<Paragraph>();
        int characterCount = 0;

        Paragraph paragraph = new Paragraph();

        int c = textReader.read();
        while (c != -1) {
            if (++characterCount > maximumLength) {
                throw new IllegalArgumentException("Text length is greater than maximum length.");
            }

            if (c == '\n') {
                paragraphs.add(paragraph);
                paragraph = new Paragraph();
            } else {
                paragraph.append((char) c);
            }

            c = textReader.read();
        }

        paragraphs.add(paragraph);

        // Clear the edit history
        editHistory.clear();

        // Update content
        paragraphSequence.remove(0, paragraphSequence.getLength());

        for (int i = 0, n = paragraphs.getLength(); i < n; i++) {
            paragraphSequence.add(paragraphs.get(i));
        }
    }

    public void insertText(CharSequence text, int index) {
        insertText(text, index, true);
    }

    private void insertText(CharSequence text, final int index, final boolean addToEditHistory) {
        if (text == null) {
            throw new IllegalArgumentException();
        }

        if (index < 0 || index > characterCount) {
            throw new IndexOutOfBoundsException();
        }

        if (text.length() > 0) {
            // Insert the text
            int paragraphIndex = getParagraphAt(index);
            Paragraph paragraph = paragraphs.get(paragraphIndex);

            int characterOffset = index - paragraph.offset;

            StringBuilder textBuilder = new StringBuilder();

            for (int i = 0, n = text.length(); i < n; i++) {
                char c = text.charAt(i);

                if (c == '\n') {
                    // Split paragraph at current offset
                    final int count = paragraph.characters.length();

                    CharSequence trailingCharacters = paragraph.characters.subSequence(characterOffset, count);
                    paragraph.removeText(characterOffset, count - characterOffset);
                    paragraph.insertText(textBuilder, characterOffset);

                    paragraph = new Paragraph();
                    paragraph.insertText(trailingCharacters, 0);
                    paragraphSequence.insert(paragraph, ++paragraphIndex);
                    characterOffset = 0;

                    textBuilder = new StringBuilder();
                } else {
                    // Append character
                    textBuilder.append(c);
                }
            }

            paragraph.insertText(textBuilder, characterOffset);

            // Add an insert history item
            if (addToEditHistory) {
                addHistoryItem(new InsertTextEdit(text, index));
            }
        }
    }

    public void removeText(int index, int count) {
        removeText(index, count, true);
    }

    private void removeText(final int index, final int count, final boolean addToEditHistory) {
        if (index < 0 || index + count > characterCount) {
            throw new IndexOutOfBoundsException();
        }

        if (count > 0) {
            // Add a remove history item
            if (addToEditHistory) {
                addHistoryItem(new RemoveTextEdit(index, count));
            }

            // Identify the leading and trailing paragraph indexes
            final int endParagraphIndex = getParagraphAt(index + count);
            Paragraph endParagraph = paragraphs.get(endParagraphIndex);

            int beginParagraphIndex = endParagraphIndex;
            Paragraph beginParagraph = endParagraph;

            while (beginParagraph.offset > index) {
                beginParagraph = paragraphs.get(--beginParagraphIndex);
            }

            if (beginParagraphIndex == endParagraphIndex) {
                // The removal affects only a single paragraph
                beginParagraph.removeText(index - beginParagraph.offset, count);
            } else {
                // The removal spans paragraphs; remove any intervening paragraphs and
                // merge the leading and trailing segments
                String leadingText = beginParagraph.characters.substring(0, index - beginParagraph.offset);
                endParagraph.removeText(0, (index + count) - endParagraph.offset);
                paragraphSequence.remove(beginParagraphIndex, endParagraphIndex - beginParagraphIndex);
                endParagraph.insertText(leadingText, 0);
            }
        }
    }

    private void updateParagraphOffsets(final int from, final int count) {
        if (count != 0) {
            for (int i = from, n = paragraphs.getLength(); i < n; i++) {
                Paragraph paragraph = paragraphs.get(i);
                paragraph.offset += count;
            }
        }
    }

    /**
     * Returns the text area's paragraph sequence.
     */
    public ParagraphSequence getParagraphs() {
        return paragraphSequence;
    }

    /**
     * Returns the index of the paragraph containing a given character index.
     *
     * @param index
     */
    public int getParagraphAt(final int index) {
        if (index < 0 || index > characterCount) {
            throw new IndexOutOfBoundsException();
        }

        int paragraphIndex = paragraphs.getLength() - 1;
        Paragraph paragraph = paragraphs.get(paragraphIndex);

        while (paragraph.offset > index) {
            paragraph = paragraphs.get(--paragraphIndex);
        }

        return paragraphIndex;
    }

    /**
     * Returns the character at a given index.
     *
     * @param index
     */
    public char getCharacterAt(final int index) {
        if (index < 0 || index >= characterCount) {
            throw new IndexOutOfBoundsException();
        }

        final int paragraphIndex = getParagraphAt(index);
        Paragraph paragraph = paragraphs.get(paragraphIndex);

        final int characterOffset = index - paragraph.offset;

        return (characterOffset == paragraph.characters.length()) ? '\n' : paragraph.characters.charAt(characterOffset);
    }

    /**
     * Returns the number of characters in the text area, including line break
     * characters.
     */
    public int getCharacterCount() {
        return characterCount;
    }

    /**
     * Places any selected text on the clipboard and deletes it from
     * the text input.
     */
    public void cut() {
        copy();
        removeText(selectStart, selectLength);
    }

    /**
     * Places any selected text on the clipboard.
     */
    public void copy() {
        // Copy select to clipboard
        String selectedText = getSelectedText();

        if (selectedText.length() > 0) {
            LocalManifest clipboardContent = new LocalManifest();
            clipboardContent.putText(selectedText);
            Clipboard.setContent(clipboardContent);
        }
    }

    /**
     * Inserts text from the clipboard into the text input.
     */
    public void paste() {
        Manifest clipboardContent = Clipboard.getContent();

        if (clipboardContent != null && clipboardContent.containsText()) {
            // Paste the string representation of the content
            String text = null;
            try {
                text = clipboardContent.getText();
            } catch (IOException exception) {
                // No-op
            }

            if (text != null) {
                if ((characterCount + text.length()) > maximumLength) {
                    Toolkit.getDefaultToolkit().beep();
                } else {
                    removeText(selectStart, selectLength);
                    insertText(text, selectStart);
                }
            }
        }
    }

    public void undo() {
        int n = editHistory.getLength();
        if (n > 0) {
            Edit edit = editHistory.remove(n - 1, 1).get(0);
            edit.undo();
        }
    }

    private void addHistoryItem(Edit edit) {
        editHistory.add(edit);

        if (editHistory.getLength() > MAXIMUM_EDIT_HISTORY_LENGTH) {
            editHistory.remove(0, 1);
        }
    }

    /**
     * Returns the starting index of the select.
     *
     * @return The starting index of the select.
     */
    public int getSelectStart() {
        return selectStart;
    }

    /**
     * Returns the length of the select.
     *
     * @return The length of the select; may be <tt>0</tt>.
     */
    public int getSelectLength() {
        return selectLength;
    }

    /**
     * Returns a span representing the current select.
     *
     * @return A span containing the current select. Both start and end points are
     *         inclusive. Returns <tt>null</tt> if the select is empty.
     */
    public Span getSelect() {
        return (selectLength == 0) ? null : new Span(selectStart, selectStart + selectLength - 1);
    }

    /**
     * Sets the select. The sum of the select start and length must be
     * less than the length of the text area's content.
     *
     * @param selectStart  The starting index of the select.
     * @param selectLength The length of the select.
     */
    public void setSelect(final int selectStart, final int selectLength) {
        if (selectLength < 0) {
            throw new IllegalArgumentException("selectLength is negative.");
        }

        if (selectStart < 0 || selectStart + selectLength > characterCount) {
            throw new IndexOutOfBoundsException();
        }

        final int previousSelectStart = this.selectStart;
        final int previousSelectLength = this.selectLength;

        if (previousSelectStart != selectStart || previousSelectLength != selectLength) {
            this.selectStart = selectStart;
            this.selectLength = selectLength;

            textAreaSelectListeners.selectChanged(this, previousSelectStart, previousSelectLength);
        }
    }

    /**
     * Sets the select.
     *
     * @param select
     * @see #setSelect(int, int)
     */
    public final void setSelect(Span select) {
        if (select == null) {
            throw new IllegalArgumentException("select is null.");
        }

        setSelect(Math.min(select.start, select.end), (int) select.getLength());
    }

    /**
     * Selects all text.
     */
    public void selectAll() {
        setSelect(0, characterCount);
    }

    /**
     * Clears the select.
     */
    public void clearSelect() {
        setSelect(0, 0);
    }

    /**
     * Returns the selected text.
     *
     * @return A string containing a copy of the selected text.
     */
    public String getSelectedText() {
        return getText(selectStart, selectStart + selectLength);
    }

    /**
     * Returns the maximum length of the text area's text content.
     *
     * @return The maximum length of the text area's text content.
     */
    public int getMaximumLength() {
        return maximumLength;
    }

    /**
     * Sets the maximum length of the text area's text content.
     *
     * @param maximumLength The maximum length of the text area's text content.
     */
    public void setMaximumLength(final int maximumLength) {
        if (maximumLength < 0) {
            throw new IllegalArgumentException("maximumLength is negative.");
        }

        final int previousMaximumLength = this.maximumLength;

        if (previousMaximumLength != maximumLength) {
            this.maximumLength = maximumLength;

            // Truncate the text, if necessary
            if (characterCount > maximumLength) {
                removeText(maximumLength, characterCount - maximumLength);
            }

            // Fire change events
            textAreaListeners.maximumLengthChanged(this, previousMaximumLength);
        }
    }

    /**
     * Returns the text area's editable flag.
     */
    public boolean isEditable() {
        return editable;
    }

    /**
     * Sets the text area's editable flag.
     *
     * @param editable
     */
    public void setEditable(boolean editable) {
        if (this.editable != editable) {
            if (!editable) {
                if (isFocused()) {
                    clearFocus();
                }
            }

            this.editable = editable;

            textAreaListeners.editableChanged(this);
        }
    }

    /**
     * Returns the text area's text key.
     *
     * @return The text key, or <tt>null</tt> if no text key is set.
     */
    public String getTextKey() {
        return textKey;
    }

    /**
     * Sets the text area's text key.
     *
     * @param textKey The text key, or <tt>null</tt> to clear the binding.
     */
    public void setTextKey(String textKey) {
        String previousTextKey = this.textKey;

        if (previousTextKey != textKey) {
            this.textKey = textKey;
            textAreaBindingListeners.textKeyChanged(this, previousTextKey);
        }
    }

    public BindType getTextBindType() {
        return textBindType;
    }

    public void setTextBindType(BindType textBindType) {
        if (textBindType == null) {
            throw new IllegalArgumentException();
        }

        BindType previousTextBindType = this.textBindType;

        if (previousTextBindType != textBindType) {
            this.textBindType = textBindType;
            textAreaBindingListeners.textBindTypeChanged(this, previousTextBindType);
        }
    }

    public TextBindMapping getTextBindMapping() {
        return textBindMapping;
    }

    public void setTextBindMapping(TextBindMapping textBindMapping) {
        TextBindMapping previousTextBindMapping = this.textBindMapping;

        if (previousTextBindMapping != textBindMapping) {
            this.textBindMapping = textBindMapping;
            textAreaBindingListeners.textBindMappingChanged(this, previousTextBindMapping);
        }
    }

    @Override
    public void load(Object context) {
        if (textKey != null && JSON.containsKey(context, textKey)
        && textBindType != BindType.STORE) {

            Object object = JSON.get(context, textKey);

            if (object instanceof SSElementalData) {
                SSElementalData elementalData = (SSElementalData) JSON.get(context, textKey);

                String value = null;
                if (textBindMapping == null) {
                    value = (elementalData == null) ? "" : ((elementalData.getText() == null) ? "" : elementalData.getText());
                } else {
                    value = textBindMapping.toString(elementalData);
                }
                setText(value);
            } else if (object instanceof HashMap) {
                HashMap cellData = (HashMap) JSON.get(context, textKey);
                String value = (String) cellData.get("text");
                value = (value == null) ? "" : value;
                setText(value);
            }
        }
    }

    @Override
    public void store(Object context) {
        if (textKey != null && textBindType != BindType.LOAD) {
            String text = getText();
            SSElementalData elementalData = new SSElementalData();
            elementalData.setText(text);

            JSON.put(context, textKey, (textBindMapping == null) ? elementalData : textBindMapping.valueOf(text));
//            JSON.put(context, textKey, (textBindMapping == null) ? text : textBindMapping.valueOf(text));
        }
    }

    @Override
    public void clear() {
        if (textKey != null) {
            setText("");
        }
    }

    public int getInsertionPoint(int x, int y) {
        SSTextArea.Skin textAreaSkin = (SSTextArea.Skin) getSkin();
        return textAreaSkin.getInsertionPoint(x, y);
    }

    public int getNextInsertionPoint(int x, int from, ScrollDirection direction) {
        SSTextArea.Skin textAreaSkin = (SSTextArea.Skin) getSkin();
        return textAreaSkin.getNextInsertionPoint(x, from, direction);
    }

    public int getRowAt(int index) {
        SSTextArea.Skin textAreaSkin = (SSTextArea.Skin) getSkin();
        return textAreaSkin.getRowAt(index);
    }

    public int getRowOffset(int index) {
        SSTextArea.Skin textAreaSkin = (SSTextArea.Skin) getSkin();
        return textAreaSkin.getRowOffset(index);
    }

    public int getRowLength(int index) {
        SSTextArea.Skin textAreaSkin = (SSTextArea.Skin) getSkin();
        return textAreaSkin.getRowLength(index);
    }

    public int getRowCount() {
        SSTextArea.Skin textAreaSkin = (SSTextArea.Skin) getSkin();
        return textAreaSkin.getRowCount();
    }

    public Bounds getCharacterBounds(int index) {
        SSTextArea.Skin textAreaSkin = (SSTextArea.Skin) getSkin();
        return textAreaSkin.getCharacterBounds(index);
    }

    public ListenerList<SSTextAreaListener> getTextAreaListeners() {
        return textAreaListeners;
    }

    public ListenerList<SSTextAreaContentListener> getTextAreaContentListeners() {
        return textAreaContentListeners;
    }

    public ListenerList<SSTextAreaSelectListener> getTextAreaSelectListeners() {
        return textAreaSelectListeners;
    }

    public ListenerList<SSTextAreaBindingListener> getTextAreaBindingListeners() {
        return textAreaBindingListeners;
    }

    public TextAreaType getTextAreaType() {
        return textAreaType;
    }

    public void setTextAreaType(TextAreaType textAreaType) {
        this.textAreaType = textAreaType;
    }
}
