/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.ss;

import org.apache.pivot.collections.ArrayList;
import org.apache.pivot.collections.List;
import org.apache.pivot.collections.ListListener;
import org.apache.pivot.collections.Sequence;
import org.apache.pivot.util.ImmutableIterator;
import org.apache.pivot.util.ListenerList;
import org.apache.pivot.util.concurrent.TaskExecutionException;
import org.apache.pivot.wtk.ApplicationContext;
import org.apache.pivot.wtk.media.Image;

import java.net.URL;
import java.util.Comparator;
import java.util.Iterator;

public class SSTreeBranch extends SSTreeNode implements List<SSTreeNode> {
    private Image expandedIcon = null;

    private ArrayList<SSTreeNode> treeNodes = new ArrayList<SSTreeNode>();
    private ListListenerList<SSTreeNode> listListeners = new ListListenerList<SSTreeNode>();

    public SSTreeBranch() {
        this(null, null, null);
    }

    public SSTreeBranch(Image icon) {
        this(icon, null, null);
    }

    public SSTreeBranch(String text) {
        this(null, text, null);
    }

    public SSTreeBranch(NodeType nodeType) {
        this(nodeType.getIconImage(), null, nodeType.toString(), nodeType);
    }

    public SSTreeBranch(String text, NodeType nodeType) {
        this(nodeType.getIconImage(), null, text, nodeType);
    }

    public SSTreeBranch(Image icon, String text, NodeType nodeType) {
        this(icon, null, text, nodeType);
    }

    public SSTreeBranch(Image icon, Image expandedIcon, String text, NodeType nodeType) {
        super(icon, text, nodeType);

        this.expandedIcon = expandedIcon;
    }

    public Image getExpandedIcon() {
        return expandedIcon;
    }

    public void setExpandedIcon(Image expandedIcon) {
        this.expandedIcon = expandedIcon;
    }

    /**
     * Sets the tree branch's expanded icon by URL.
     * <p/>
     * If the icon already exists in the application context resource cache,
     * the cached value will be used.
     * Otherwise, the icon will be loaded synchronously and added to the cache.
     *
     * @param expandedIconURL The location of the expanded icon to set.
     */
    public void setExpandedIcon(URL expandedIconURL) {
        if (expandedIconURL == null) {
            throw new IllegalArgumentException("expandedIconURL is null.");
        }

        Image icon = (Image) ApplicationContext.getResourceCache().get(expandedIconURL);

        if (icon == null) {
            try {
                icon = Image.load(expandedIconURL);
            } catch (TaskExecutionException exception) {
                throw new IllegalArgumentException(exception);
            }

            ApplicationContext.getResourceCache().put(expandedIconURL, icon);
        }

        setExpandedIcon(icon);
    }

    /**
     * Sets the tree branch's expanded icon by {@linkplain ClassLoader#getResource(String)
     * resource name}.
     *
     * @param expandedIconName The resource name of the expanded icon to set.
     * @see #setExpandedIcon(URL)
     */
    public void setExpandedIcon(String expandedIconName) {
        if (expandedIconName == null) {
            throw new IllegalArgumentException("expandedIconName is null.");
        }

        ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
        setExpandedIcon(classLoader.getResource(expandedIconName.substring(1)));
    }

    @Override
    public int add(SSTreeNode treeNode) {
        if (treeNode == null) {
            throw new IllegalArgumentException("treeNode is null.");
        }

        if (treeNode.getParent() != null) {
            throw new IllegalArgumentException("treeNode already has a parent.");
        }

        int index = treeNodes.add(treeNode);
        treeNode.setParent(this);
        listListeners.itemInserted(this, index);

        return index;
    }

    @Override
    public void insert(SSTreeNode treeNode, int index) {
        if (treeNode == null) {
            throw new IllegalArgumentException("treeNode is null.");
        }

        if (treeNode.getParent() != null) {
            throw new IllegalArgumentException("treeNode already has a parent.");
        }

        treeNodes.insert(treeNode, index);
        treeNode.setParent(this);
        listListeners.itemInserted(this, index);
    }

    @Override
    public SSTreeNode update(int index, SSTreeNode treeNode) {
        if (treeNode == null) {
            throw new IllegalArgumentException("treeNode is null.");
        }

        if (treeNode.getParent() != null) {
            throw new IllegalArgumentException("treeNode already has a parent.");
        }

        SSTreeNode previousSSTreeNode = treeNodes.update(index, treeNode);
        previousSSTreeNode.setParent(null);
        treeNode.setParent(this);
        listListeners.itemUpdated(this, index, previousSSTreeNode);

        return previousSSTreeNode;
    }

    @Override
    public int remove(SSTreeNode treeNode) {
        int index = indexOf(treeNode);
        if (index != -1) {
            remove(index, 1);
        }

        return index;
    }

    @Override
    public Sequence<SSTreeNode> remove(int index, int count) {
        Sequence<SSTreeNode> removed = treeNodes.remove(index, count);
        if (count > 0) {
            for (int i = 0, n = removed.getLength(); i < n; i++) {
                SSTreeNode treeNode = removed.get(i);
                treeNode.setParent(null);
            }

            listListeners.itemsRemoved(this, index, removed);
        }

        return removed;
    }

    @Override
    public void clear() {
        if (getLength() > 0) {
            for (int i = 0, n = treeNodes.getLength(); i < n; i++) {
                SSTreeNode treeNode = treeNodes.get(i);
                treeNode.setParent(null);
            }

            treeNodes.clear();
            listListeners.listCleared(this);
        }
    }

    @Override
    public SSTreeNode get(int index) {
        return treeNodes.get(index);
    }

    @Override
    public int indexOf(SSTreeNode treeNode) {
        return treeNodes.indexOf(treeNode);
    }

    @Override
    public boolean isEmpty() {
        return treeNodes.isEmpty();
    }

    @Override
    public int getLength() {
        return treeNodes.getLength();
    }

    @Override
    public Comparator<SSTreeNode> getComparator() {
        return treeNodes.getComparator();
    }

    @Override
    public void setComparator(Comparator<SSTreeNode> comparator) {
        Comparator<SSTreeNode> previousComparator = treeNodes.getComparator();

        if (previousComparator != comparator) {
            treeNodes.setComparator(comparator);

            // Recursively apply comparator change
            for (int i = 0, n = treeNodes.getLength(); i < n; i++) {
                SSTreeNode treeNode = treeNodes.get(i);

                if (treeNode instanceof SSTreeBranch) {
                    SSTreeBranch treeBranch = (SSTreeBranch) treeNode;
                    treeBranch.setComparator(comparator);
                }
            }

            listListeners.comparatorChanged(this, previousComparator);
        }
    }

    @Override
    public Iterator<SSTreeNode> iterator() {
        return new ImmutableIterator<SSTreeNode>(treeNodes.iterator());
    }

    @Override
    public ListenerList<ListListener<SSTreeNode>> getListListeners() {
        return listListeners;
    }
}
