/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.ss;

import jp.pegasus.pivot.factory.SSImageFactory;
import org.apache.pivot.util.concurrent.TaskExecutionException;
import org.apache.pivot.wtk.ApplicationContext;
import org.apache.pivot.wtk.Keyboard;
import org.apache.pivot.wtk.media.Image;

import java.net.URL;

public class SSTreeNode {
    private SSTreeBranch parent = null;

    private Image icon = null;
    private String text = null;
    private String shortCutText = null;
    private Object userData = null;

    private NodeType nodeType = null;
    private SSTabPane tabPane = null;

    /**
     * Node Type
     */
    public enum NodeType {
        MnuGrp,         // Menu  File Group
        MnuFilText,     // Menu  New Text File
        MnuFilGrid,     // Menu  New Grid File
        MnuFld,         // Menu  Folder
        MnuCpy,         // Menu  Create File/Folder Copy
        MnuDel,         // Menu  Delete File/Folder
        MnuExt,         // Menu  Exit
        MnuHlp,         // Menu  Help
        MnuSrh,         // Menu  Search
        ShtCpy,         // Sheet Copy
        ShtCut,         // Sheet Cut
        ShtDel,          // Sheet Delete
        ShtPst,         // Sheet Paste
        ShtSav,         // Sheet Save
        TabDel;         // Tab   Delete

        @Override
        public String toString() {
            return
            this == MnuGrp ? "File" : (
            this == MnuFilText ? "New Text File" : (
            this == MnuFilGrid ? "New Grid File" : (
            this == MnuFld ? "New Folder" : (
            this == MnuCpy ? "Copy" : (
            this == MnuDel ? "Delete" : (
            this == MnuExt ? "Exit" : (
            this == MnuHlp ? "Help" : (
            this == MnuSrh ? "Search" : (
            this == ShtCpy ? "Copy" : (
            this == ShtCut ? "Cut" : (
            this == ShtDel ? "Delete" : (
            this == ShtPst ? "Paste" : (
            this == ShtSav ? "Save" : (
            this == TabDel ? "Delete" : null))))))))))))));
        }

        public String getShortCutText() {
            return
            this == MnuCpy ? "CMD-C" : (
            this == MnuDel ? "Delete" : (
            this == MnuFilText ? "CMD-T" : (
            this == MnuFilGrid ? "CMD-G" : (
            this == MnuFld ? "CMD-I" : (
            this == MnuExt ? "CMD-D" : (
            this == MnuHlp ? "CMD-H" : (
            this == MnuSrh ? "CMD-F" : (
            this == ShtCpy ? "CMD-C" : (
            this == ShtCut ? "CMD-X" : (
            this == ShtDel ? "Delete" : (
            this == ShtPst ? "CMD-V" : (
            this == ShtSav ? "CMD-S" : (
            this == TabDel ? "Delete" :
            null)))))))))))));
        }

        public Image getIconImage() {
            return
            this == MnuGrp ? SSImageFactory.getIconImage(MnuGrp) : (
            this == MnuCpy ? SSImageFactory.getIconImage(MnuCpy) : (
            this == MnuDel ? SSImageFactory.getIconImage(MnuDel) : (
            this == MnuExt ? SSImageFactory.getIconImage(MnuExt) : (
            this == MnuFilText ? SSImageFactory.getIconImage(MnuFilText) : (
            this == MnuFilGrid ? SSImageFactory.getIconImage(MnuFilGrid) : (
            this == MnuFld ? SSImageFactory.getIconImage(MnuFld) : (
            this == MnuHlp ? SSImageFactory.getIconImage(MnuHlp) : (
            this == MnuSrh ? SSImageFactory.getIconImage(MnuSrh) : (
            this == ShtCpy ? SSImageFactory.getIconImage(ShtCpy) : (
            this == ShtCut ? SSImageFactory.getIconImage(ShtCut) : (
            this == ShtDel ? SSImageFactory.getIconImage(ShtDel) : (
            this == ShtPst ? SSImageFactory.getIconImage(ShtPst) : (
            this == ShtSav ? SSImageFactory.getIconImage(ShtSav) : (
            this == TabDel ? SSImageFactory.getIconImage(TabDel) : null))))))))))))));
        }

        public String getIconFileNamePath() {
            return
            this == MnuGrp ? "/image/document-new.png" : (
            this == MnuCpy ? "/image/edit-copy.png" : (
            this == MnuDel ? "/image/edit-delete.png" : (
            this == MnuExt ? "/image/anchor.png" : (
            this == MnuFilText ? "/image/page_white.png" : (
            this == MnuFilGrid ? "/image/x-office-document.png" : (
            this == MnuFld ? "/image/folder.png" : (
            this == MnuHlp ? "/image/help.png" : (
            this == MnuSrh ? "/image/magnifier.png" : (
            this == ShtCpy ? "/image/edit-copy.png" : (
            this == ShtCut ? "/image/edit-cut.png" : (
            this == ShtDel ? "/image/edit-delete.png" : (
            this == ShtPst ? "/image/edit-paste.png" : (
            this == ShtSav ? "/image/document-save.png" : (
            this == TabDel ? "/image/edit-delete.png" : null))))))))))))));
        }
    }

    public SSTreeNode() {
        this(null, null, null);
    }

    public SSTreeNode(Image icon) {
        this(icon, null, null);
    }

    public SSTreeNode(String text) {
        this(null, text, null);
    }

    public SSTreeNode(NodeType nodeType) {
        this.icon = nodeType.getIconImage();
        this.text = nodeType.toString();
        this.shortCutText = Keyboard.KeyStroke.decode(nodeType.getShortCutText()).toString();
        this.userData = null;
        this.nodeType = nodeType;
    }

    public SSTreeNode(Image icon, String text, NodeType nodeType) {
        this.icon = icon;
        this.text = text;
        this.shortCutText = null;
        this.userData = null;
        this.nodeType = nodeType;
    }

    public SSTreeBranch getParent() {
        return parent;
    }

    protected void setParent(SSTreeBranch parent) {
        this.parent = parent;
    }

    public Image getIcon() {
        return icon;
    }

    public void setIcon(Image icon) {
        this.icon = icon;
    }

    /**
     * Sets the tree node's icon by URL.
     * <p/>
     * If the icon already exists in the application context resource cache,
     * the cached value will be used. Otherwise, the icon will be loaded
     * synchronously and added to the cache.
     *
     * @param iconURL The location of the icon to set.
     */
    public void setIcon(URL iconURL) {
        if (iconURL == null) {
            throw new IllegalArgumentException("iconURL is null.");
        }

        Image icon = (Image) ApplicationContext.getResourceCache().get(iconURL);

        if (icon == null) {
            try {
                icon = Image.load(iconURL);
            } catch (TaskExecutionException exception) {
                throw new IllegalArgumentException(exception);
            }

            ApplicationContext.getResourceCache().put(iconURL, icon);
        }

        setIcon(icon);
    }

    /**
     * Sets the tree node's icon by {@linkplain ClassLoader#getResource(String)
     * resource name}.
     *
     * @param iconName The resource name of the icon to set.
     * @see #setIcon(URL)
     */
    public void setIcon(String iconName) {
        if (iconName == null) {
            throw new IllegalArgumentException("iconName is null.");
        }

        ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
        setIcon(classLoader.getResource(iconName.substring(1)));
    }

    public String getText() {
        return text;
    }

    public void setText(String text) {
        this.text = text;
    }

    public String getShortCutText() {
        return shortCutText;
    }

    public void setShortCutText(String shortCutText) {
        this.shortCutText = shortCutText;
    }

    public Object getUserData() {
        return userData;
    }

    public void setUserData(Object userData) {
        this.userData = userData;
    }

    public NodeType getNodeType() {
        return nodeType;
    }

    public void setNodeType(NodeType nodeType) {
        this.nodeType = nodeType;
    }

    public SSTabPane getTabPane() {
        return tabPane;
    }

    public void setTabPane(SSTabPane tabPane) {
        this.tabPane = tabPane;
    }
}
