/*
 * Copyright (C) 2010-2011 Toshiyuki Kitani
 *
 *  Pegasus is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Pegasus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package jp.pegasus.pivot.ss.tableview;

import jp.pegasus.pivot.listener.tableview.SSTableViewHeaderListener;
import jp.pegasus.pivot.listener.tableview.SSTableViewHeaderPressListener;
import org.apache.pivot.util.ListenerList;
import org.apache.pivot.wtk.Bounds;
import org.apache.pivot.wtk.Component;

public class SSTableViewHeader extends Component {

    private int[] selectRange = {-1, -1, -1, -1};
    private long selectIndex = 0x0000000000000000l;

    private SSTableView tableView;

    private boolean isHeaderPressed = false;
    private boolean isSelect = false;

    private SortMode sortMode = SortMode.NONE;

    private SSTableViewHeaderListenerList tableViewHeaderListeners = new SSTableViewHeaderListenerList();
    private SSTableViewHeaderPressListenerList tableViewHeaderPressListeners = new SSTableViewHeaderPressListenerList();

    /**
     * Enumeration representing a sort mode.
     */
    public enum SortMode {
        NONE,
        SINGLE_COLUMN,
        MULTI_COLUMN
    }

    public SSTableViewHeader() {
        this((SSTableView) null);
//        installSkin(SSTableViewHeader.class);
//        setTableView((SSTableView)null);
    }

    public SSTableViewHeader(SSTableView tableView) {
        installSkin(SSTableViewHeader.class);
        setTableView(tableView);
    }

    /**
     * Table view header skin interface. Table view header skins must
     * implement this.
     */
    public interface Skin {
        public int getHeaderAt(int x);

        public Bounds getHeaderBounds(int index);
    }

    @Override
    protected void setSkin(org.apache.pivot.wtk.Skin skin) {
        if (!(skin instanceof SSTableViewHeader.Skin)) {
            throw new IllegalArgumentException("Skin class must implement "
            + SSTableViewHeader.Skin.class.getName());
        }

        super.setSkin(skin);
    }

    public SSTableView getTableView() {
        return tableView;
    }

    public void setTableView(SSTableView tableView) {
        SSTableView previousTableView = this.tableView;

        if (previousTableView != tableView) {
            this.tableView = tableView;
            tableViewHeaderListeners.tableViewChanged(this, previousTableView);
        }
    }

    public int[] getSelectRange() {
        return selectRange;
    }

    public void setSelectRange(int[] selectRange) {
        this.selectRange = selectRange;
    }

    public long getSelectIndex() {
        return selectIndex;
    }

    public void setSelectIndex(long selectIndex) {
        this.selectIndex = selectIndex;
    }

    public boolean isHeaderPressed() {
        return isHeaderPressed;
    }

    public void setHeaderPressed(boolean headerPressed) {
        isHeaderPressed = headerPressed;
    }

    public boolean isSelect() {
        return isSelect;
    }

    public void setSelect(boolean select) {
        isSelect = select;
    }

    private static class SSTableViewHeaderListenerList extends ListenerList<SSTableViewHeaderListener>
    implements SSTableViewHeaderListener {
        @Override
        public void tableViewChanged(SSTableViewHeader tableViewHeader,
                                     SSTableView previousTableView) {
            for (SSTableViewHeaderListener listener : this) {
                listener.tableViewChanged(tableViewHeader, previousTableView);
            }
        }

        @Override
        public void sortModeChanged(SSTableViewHeader tableViewHeader, SortMode previousSortMode) {
            for (SSTableViewHeaderListener listener : this) {
                listener.sortModeChanged(tableViewHeader, previousSortMode);
            }
        }
    }

    private static class SSTableViewHeaderPressListenerList extends ListenerList<SSTableViewHeaderPressListener>
    implements SSTableViewHeaderPressListener {
        @Override
        public void headerPressed(SSTableViewHeader tableViewHeader, int index) {
            for (SSTableViewHeaderPressListener listener : this) {
                listener.headerPressed(tableViewHeader, index);
            }
        }
    }

    public SortMode getSortMode() {
        return sortMode;
    }

    public void setSortMode(SortMode sortMode) {
        if (sortMode == null) {
            throw new IllegalArgumentException();
        }

        SortMode previousSortMode = this.sortMode;
        if (previousSortMode != sortMode) {
            this.sortMode = sortMode;
            tableViewHeaderListeners.sortModeChanged(this, previousSortMode);
        }
    }

    public void pressHeader(int index) {
        tableViewHeaderPressListeners.headerPressed(this, index);
    }

    /**
     * Returns the index of the header at a given location.
     *
     * @param x The x-coordinate of the header to identify.
     * @return The column index, or <tt>-1</tt> if there is no column at the given
     *         x-coordinate.
     */
    public int getHeaderAt(int x) {
        SSTableViewHeader.Skin tableViewHeaderSkin = (SSTableViewHeader.Skin) getSkin();
        return tableViewHeaderSkin.getHeaderAt(x);
    }

    /**
     * Returns the bounding area of a given header.
     *
     * @param index The index of the header.
     * @return The bounding area of the header.
     */
    public Bounds getHeaderBounds(int index) {
        SSTableViewHeader.Skin tableViewHeaderSkin = (SSTableViewHeader.Skin) getSkin();
        return tableViewHeaderSkin.getHeaderBounds(index);
    }

    public ListenerList<SSTableViewHeaderListener> getTableViewHeaderListeners() {
        return tableViewHeaderListeners;
    }

    public ListenerList<SSTableViewHeaderPressListener> getTableViewHeaderPressListeners() {
        return tableViewHeaderPressListeners;
    }
}
